#include "pch.h"

#include "Logger.h"

#include "Mem.hpp"
#include "Common.h"
#include "CfgParse.hpp"
#include "Misc.hpp"
#include "Links.h"
#include "Areas.h"
#include "Msg.hpp"
#include "Tic.h"
#include "crc32.hpp"
#include "Temp.hpp"
#include "File.h"

#include "CfgLexems.hpp"
#include "WELexems.h"

#include "MsgAttrs.h"

#include "LineFormat.h"

#include "exceptions.h"
#include "singletons.h"

#include "actions/HelpAction.h"
#include "actions/ChangeGroupAction.h"

#include "AreasManager.h"
#include "LinksManager.h"

#include <MiniSAX.h>
#include "AdrenalinSAX.h"

#include "SuspendedSAX.h"

using	miniSAX::CDataSource;
using	miniSAX::CMiniSAXParser;
using	miniSAX::CMiniSAXException;

#include "Adrenalin.h"

// main class instance

CAdrenalin*	pAdrenalin;
// static members
TCHAR CAdrenalin::m_szStartupPath[MAX_PATH] = TEXT("");
TCHAR CAdrenalin::m_szStartupDrive[3] = TEXT("");		// <letter>':''\0'

class CSingletonDestroyer {
public:
	CSingletonDestroyer() {}
	~CSingletonDestroyer() {
		CDestructionManager::getInstance()->destroyAll();
	}
};

int _tmain(int argc, TCHAR **argv) {
	::SetFileApisToOEM();

	// on exit from main() destructor of the underlying class will be called
	CSingletonDestroyer	singletonDestroyer;

	int	result = 0;

	try {

		pAdrenalin = new CAdrenalin();
		pAdrenalin->init( argc, argv );

		result = pAdrenalin->Execute();

		delete	pAdrenalin;
		

	} catch (CRuntimeException& e) {
		cerr << "\n[!] RuntimeException occured:\n" << e.getMessage().c_str() << endl;
		result = 1;
	}

	return result;
}


CAdrenalin::CAdrenalin()
{
}


void CAdrenalin::init(int argc, TCHAR **argv)
{
    // keep start time
	::GetLocalTime( &startTime );

	tossingBadTics = false;

	// prepare file buffer
	m_fileBufferSize = 1024*1024;
	m_fileBuffer     = new BYTE[m_fileBufferSize];

    // setup log
	(*CLogger::getInstance()) << LOG_ERROR;

	int i, j;

	// prepare tearline
	_stprintf( g_vars.m_szTearLine, _T("\r--- %s"), CAdrenalin::getName().c_str() );

	// prepare forward flags
	m_fForwardsRead    = false;
	m_fForwardsChanged = false;

	// keep startup path
	GetModuleFileName( NULL, m_szStartupPath, MAX_PATH );

	for (i = 0, j = -1; m_szStartupPath[i] != _T('\0'); i++) {
	if (m_szStartupPath[i] == _T('\\'))
		j = i;
	}
	m_szStartupPath[++j] = _T('\0');
	// keep startup drive
	lstrcpyn( m_szStartupDrive, m_szStartupPath, 3 );

	m_iErrorCode = ERROR_NONE;

	// read configuration

	try {

		readConfiguration();

	} catch (CBadConfigException& cfgEx) {
		(*CLogger::getInstance()) << LOG_ERROR << cfgEx.getMessage() << TEXT("\n");
		// TODO: ALERT: fuckingbad code found!!! ;)
		m_iErrorCode = ERROR_NONE + 1;
		m_dwAction   = ACTION_ERROR;

		return;
	}

	try {
		// we initiate load of links and areas
		CLinksManager::getInstance();
		CAreasManager::getInstance();
	} catch (CBadConfigException& e) {
		(*CLogger::getInstance()) << LOG_ERROR << e.getMessage() << TEXT("\n");
		// TODO: ALERT: fuckingbad code found!!! ;)
		m_iErrorCode = ERROR_NONE + 1;
		m_dwAction   = ACTION_ERROR;

		return;
	}

	// possibly now log is working in StdErr+FileOut mode
	// "possibly" means that FileOut may be turned off if we failed to open log file

	// prepare language
	if (m_iErrorCode == ERROR_NONE) {
		const tstring&	lngFileName = CCommonData::getInstance()->getLanguageFileName();
		if (_taccess( lngFileName.c_str(), 0) == 0) {
			if (!m_language.Read( lngFileName.c_str() )) {
				(*CLogger::getInstance()) << TEXT("failed to read language file ")
					                      << lngFileName
                                          << TEXT("\n");
				m_iErrorCode = ERROR_LNG_READ;
				m_dwAction   = ACTION_ERROR;
			} else if (m_language.Count() != LNG_COUNT) {
				(*CLogger::getInstance()) << TEXT("wrong number of messages in language file:\n")
					         << TEXT("  ")
					         << (short int)m_language.Count()
					         << TEXT(" instead of ")
					         << LNG_COUNT << TEXT("\n");
				m_iErrorCode = ERROR_LNG_NUMBER;
				m_dwAction   = ACTION_ERROR;
			}
		}
		else {
			(*CLogger::getInstance()) << TEXT("language file ") << lngFileName
                         << TEXT(" not found.\n");
			m_iErrorCode = ERROR_LNG_FILE_NOT_FOUND;
			m_dwAction   = ACTION_ERROR;
		}
	}

	// no comments ;)

	if (m_iErrorCode == ERROR_NONE) {
		CheckDiskLimits();
	} else {
		TCHAR	szErrorMessage[256];
		GetErrorMessage( m_iErrorCode, szErrorMessage );
		(*CLogger::getInstance()) << szErrorMessage << TEXT("\n");
	}

	// can we continue?

	if (m_iErrorCode == ERROR_NONE) {
		ParseCmdLine( argc, argv );
	} else {
		m_dwAction = ACTION_ERROR;
	}

	if (m_iErrorCode == ERROR_NONE) {
		if (!AllocGlobalMsgBuf()) {
			m_iErrorCode = ERROR_MEMORY;
			m_dwAction   = ACTION_ERROR;
		}
	}
}


CAdrenalin::~CAdrenalin() {
	if (m_fForwardsChanged)
		SaveForwards();

	FreeGlobalMsgBuf();

	// free file buffer
	delete[] m_fileBuffer;
}


void CAdrenalin::readConfiguration() throw(CBadConfigException) {
	TCHAR	fullCfgName[MAX_PATH];

	lstrcpy( fullCfgName, CAdrenalin::getStartupPath() );
	lstrcat( fullCfgName, TEXT("adrenalin.xml") );

	try {
		try {
			auto_ptr<CMiniSAXParser>	parserPtr( CMiniSAXParser::create() );
			parserPtr->setHandlerEncoding( ENCODING_CP866 );
			CAdrenalinContentHandler	adrContentHandler( parserPtr.get() );

			CDataSource		dataSource( fullCfgName );

			parserPtr->setContentHandler( &adrContentHandler );

			parserPtr->parse( &dataSource );

		} catch (miniSAX::CIOException& ioEx) {
			throw CBadConfigException( tstring("Failed to read 'adrenalin.xml': ") + ioEx.getMessage() );
		} catch (CMiniSAXException& saxEx) {
			// prepare exception message
			tstring	msg("Failed to read 'adrenalin.xml':");
			if (saxEx.getColumn() >= 0 && saxEx.getRow() >= 0) {
				char	buffer[256];
				sprintf( buffer, " [col:%d,row:%d] ", saxEx.getColumn(), saxEx.getRow() );
				msg += buffer;
			} else {
				msg += " [no position info] ";
			}
			msg += saxEx.getMessage();

			throw CBadConfigException( msg );
		}
	} catch (CBadConfigException) {
		// ensure common log file name is valid
		tstring	fileName = CCommonData::getInstance()->getCommonLogFileName();
		CAdrenalin::adjustFileName( fileName );
		CCommonData::getInstance()->setCommonLogFileName( fileName );
		initCommonLog();
		throw;
	}
	initCommonLog();
}


void CAdrenalin::initCommonLog() {
	tstring	commonLogFileName = CCommonData::getInstance()->getCommonLogFileName();
	// get dir from name
	unsigned int	i;
	int	j;
	for (i = 0, j = -1; i < commonLogFileName.length(); i++) {
		if (commonLogFileName[i] == TEXT('\\'))
			j = i;
	}
	if (j >= 0) {
		// ensure dir exists
		tstring	path = commonLogFileName.substr( 0, j );
		MakeDirStructure( path.c_str() );
	}
	if (!CLogger::getInstance()->openLogFile( commonLogFileName.c_str() )) {
		(*CLogger::getInstance()) << TEXT("Failed to open log file: ") << commonLogFileName << TEXT("\n");
	}
	else {
		LoggerPtr	logger;
		// print start time only in log

		TCHAR	buffer[128];
		CLogger::formatDateTime( startTime, buffer );

		logger << LOG_START << buffer << TEXT("\n");
	}
}


void formatSize(DWORD size, char* str) {
	// 1234 -> 1,234

	int	i, j, count;

	count = 0;
	j = 0;
	while (TRUE) {
		str[j] = (char)(size % 10) + '0';
		size /= 10;
		j++;
		count++;
		if (count == 3 && size > 0) {
			str[j] = ',';
			j++;
			count = 0;
		}
		if (size == 0)
			break;
	}
	str[j] = '\0';

	for (i = 0; i < (j >> 1); i++) {
		char	c  = str[i];
		str[i]     = str[j - 1 - i];
		str[j - 1 - i] = c;
	}
}

struct DriveParams {
	tstring	path;
	__int64	freeBytes;
	__int64	limitBytes;
};


void formatSize(__int64 size, LPTSTR str) {
	_stprintf( str, TEXT("%I64d"), size );
	int	len = lstrlen( str );
	int	pos = len-1;
	while (pos > 0) {
		pos -= 3;
		if (pos >= 0) {
			// shift tail with zero char
			::MoveMemory( str + pos + 2, str + pos + 1, (len + 1 - pos)*sizeof(TCHAR) );
			++len;
			str[pos+1] = TEXT(',');
		}
	}
}


void CAdrenalin::CheckDiskLimits() {

	LoggerPtr	logger;

	// lets get OS version info to decide what function to use:
	//   GetDiskFreeSpace or GetDiskFreeSpaceEx
	bool		fUseExFunction = false;
	OSVERSIONINFO	versionInfo;

	versionInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if (GetVersionEx(&versionInfo)) {
		switch (versionInfo.dwPlatformId) {
			case VER_PLATFORM_WIN32_WINDOWS:
				if (LOWORD(versionInfo.dwBuildNumber) > 1000)
					fUseExFunction = true;
				break;
			case VER_PLATFORM_WIN32_NT:
				fUseExFunction = true;
				break;
		}
	}

	BOOL		fSuccess;
	DWORD		dwSectorsPerCluster;
	DWORD		dwBytesPerSector;
	DWORD		dwNumberOfFreeClusters;
	DWORD		dwTotalNumberOfClusters;
	ULARGE_INTEGER	freeBytesAvail;
	ULARGE_INTEGER	totalBytes;
	ULARGE_INTEGER	totalFreeBytes;
	ULARGE_INTEGER	limit;

	vector<DriveParams>	vFailedSizes;

	CommonDataPtr	cd = CCommonData::getInstance();

	for (unsigned int i = 0; i < cd->getDriveSpaceLimitCount(); i++) {
		if (fUseExFunction) {
			fSuccess = GetDiskFreeSpaceEx( cd->getDriveSpaceLimit(i).getPath(),
                        			       &freeBytesAvail,
                                           &totalBytes,
                                           &totalFreeBytes );
		}
		else {
			fSuccess = GetDiskFreeSpace( cd->getDriveSpaceLimit(i).getPath(),
                        			     &dwSectorsPerCluster,
                                         &dwBytesPerSector,
                                         &dwNumberOfFreeClusters,
                                         &dwTotalNumberOfClusters );
                        if (fSuccess) {
							freeBytesAvail.QuadPart = dwSectorsPerCluster*dwBytesPerSector*dwNumberOfFreeClusters;
                        }
		}
		if (fSuccess) {
			DWORD	freeKBytes = freeBytesAvail.LowPart;
			limit.QuadPart = (__int64)(cd->getDriveSpaceLimit(i).getLimit()) << 10;
			if (freeBytesAvail.QuadPart < limit.QuadPart) {
				DriveParams	params;
				params.path       = cd->getDriveSpaceLimit(i).getPath();
				params.freeBytes  = freeBytesAvail.QuadPart;
				params.limitBytes = limit.QuadPart;
				vFailedSizes.push_back( params );
			}
		} else {
            TCHAR* lpMsgBuf;
            
            FormatMessage(
                FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                NULL,
                GetLastError(),
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                (LPTSTR) &lpMsgBuf,
                0,
                NULL
            );
            
            // truncate tail CRs and LFs
            int	len = lstrlen(lpMsgBuf);
            for (int j = len-1; j > 0; j--) {
                if (lpMsgBuf[j] == _T('\r') || lpMsgBuf[i] == _T('\n'))
                    lpMsgBuf[j] = _T('\0');
                else
                    break;
            }

            // Display the string.						
            logger << LOG_ERROR << "failed to get drive "
					<< cd->getDriveSpaceLimit(i).getPath() << TEXT(" free space:\n")
					<< lpMsgBuf << TEXT("\n");						
            
            // Free the buffer.
            LocalFree( lpMsgBuf );
		}
	}

	if (!vFailedSizes.empty()) {
		logger << LOG_ERROR << TEXT("low disk space\n");

		m_iErrorCode = ERROR_LOW_DISK_SPACE;
		TCHAR		szErrorMsg[256];
		TCHAR		szBuf[256];

		TCHAR	freeBytesStr[64], limitBytesStr[64];

		LoggerPtr	logger;
		logger << LOG_ERROR;
		vector<DriveParams>::iterator	it = vFailedSizes.begin();
		while (it != vFailedSizes.end()) {
			formatSize( (*it).freeBytes,  freeBytesStr );
			formatSize( (*it).limitBytes, limitBytesStr );
			_stprintf( szErrorMsg, TEXT("  %s -  %s byte(s) with the limit of %s byte(s)\n"),
			           (*it).path.c_str(), freeBytesStr, limitBytesStr );
			logger << szErrorMsg;			
			it++;
		}

		// prepare data for sysop
		TCHAR	szLowSpaceFileName[MAX_PATH];
		TCHAR	szCurTime[256];
		bool	fAppend = false;

        time_t		timestamp = time(NULL);
        struct tm*	tmStruct = localtime(&timestamp);

        _stprintf( szCurTime, _T("%02d-%02d-%04d"), tmStruct->tm_mday, tmStruct->tm_mon + 1, (tmStruct->tm_year < 80 ? tmStruct->tm_year + 2000 : tmStruct->tm_year + 1900) );

		lstrcpy( szLowSpaceFileName, m_szStartupPath );
		lstrcat( szLowSpaceFileName, _T("LowSpace.dat") );

		ifstream	in(szLowSpaceFileName);
		if (in) {
			
			// read date
            in.getline( szBuf, 256 );

			if (in) {
				if (lstrcmp( szBuf, szCurTime ) == 0) {
					fAppend = true;
					// skip all low space data already found
					while (!in.eof() && !vFailedSizes.empty()) {
						in.getline( szBuf, 256 );
						if (!in)
							break;
						it = vFailedSizes.begin();
						while (it != vFailedSizes.end() && lstrcmpi( (*it).path.c_str(), szBuf ) != 0) {
							it++;
						}
						if (it != vFailedSizes.end()) {
							vFailedSizes.erase(it);
						}
					}
				}
			}
			in.close();
		}
		if (!vFailedSizes.empty()) {
			CFile	out;
			if (fAppend) {
        		out.Open( szLowSpaceFileName, GENERIC_WRITE, OPEN_ALWAYS );
				out.SeekToEnd();
			} else {
				out.Open( szLowSpaceFileName, GENERIC_WRITE, CREATE_ALWAYS );
			}
       		out.WriteString( szCurTime );
			out.WriteString( "\r\n" );

        	vector<tstring>	vBodyLines;

        	vBodyLines.push_back(_T(""));
        	vBodyLines.push_back(_T("Low disk space on drive(s):"));

        	TCHAR	szLine[256];

        	it = vFailedSizes.begin();
        	while (it != vFailedSizes.end()) {
				formatSize( (*it).freeBytes,  freeBytesStr );
				formatSize( (*it).limitBytes, limitBytesStr );
				_stprintf( szLine, TEXT("  %s -  %s byte(s) with the limit of %s byte(s)\n"),
							(*it).path.c_str(), freeBytesStr, limitBytesStr );
        		vBodyLines.push_back( szLine );
				out.WriteString( (*it).path.c_str() );
				out.WriteString( "\r\n" );
				it++;
        	}
        	out.Close();
			CommonDataPtr	cdata = CCommonData::getInstance();
			MakeDirStructure( cdata->getNetmailPath().c_str() );
        	CMsg	msg;
			msg.setFromAddress( cdata->getSysopAddress() );
        	msg.setToAddress( cdata->getSysopAddress() );
			msg.setSenderName( cdata->getSysopName() );
			msg.setReceiverName( "Adrenalin" );
            msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0, _T("Low disk space!"), vBodyLines );
		}
	}
}


void CAdrenalin::ParseCmdLine(int argc, TCHAR **argv) {
	// convert arguments from ANSI to OEM
	for (int index = 0; index < argc; index++) {
		CharToOem( argv[index], argv[index] );
	}

	LoggerPtr	logger;
	logger << LOG_ERROR;

    int i;

    m_dwAction = ACTION_NONE;

    if (argc <= 1) { // no parameters
		actions.push_back( new CHelpAction(CHelpAction::BRIEF_HELP) );
    }
    else {
		// parse actions
        for (i = 1; i < argc; i++) {
            if (lstrcmpi( argv[i], _T("HELP") ) == 0) {
				actions.push_back( new CHelpAction(CHelpAction::DETAILED_HELP) );
            }
            else if (lstrcmpi( argv[i], _T("FIX") ) == 0) {
                m_dwAction |= ACTION_FIX;
            }
            else if (lstrcmpi( argv[i], _T("TOSS") ) == 0) {
                m_dwAction |= ACTION_TOSS;
            }
			else if (lstrcmpi( argv[i], TEXT("TOSSBAD") ) == 0) {
				m_dwAction |= ACTION_TOSS_BAD;
			}
            else if (lstrcmpi( argv[i], _T("DEBUG") ) == 0) {
				m_dwAction |= ACTION_DEBUG;
            }
			else if (lstrcmpi( argv[i], _T("RELINK") ) == 0) {
				m_dwAction |= ACTION_RELINK;
				if (i == argc - 1) {
					logger << TEXT("RELINK address absent\n");
					m_dwAction |= ACTION_ERROR;
					m_iErrorCode = ERROR_UNDEFINED_ADDRESS;
					break;
				}
				i++;
				// get relink address
				relinkData.address = argv[i];
				i++;
				// get relink robot name
				if (i < argc) {
					if (lstrlen(argv[i]) > 35) {
						logger.rememberMode();
						logger << LOG_WARNING << TEXT("too long robot name, truncated to 35 symbols\n");
						logger.restoreMode();
					}
					TCHAR	buffer[36];
					lstrcpyn( buffer, argv[i], 36 );
					relinkData.robotName = buffer;
				}
			}
            else if (lstrcmpi( argv[i], _T("PURGE") ) == 0) {
                m_dwAction |= ACTION_PURGE;
                // check if we have areas to purge defined
                if (i < argc-1) {
                    for (++i; i < argc; i++) {
						m_vsPurgeAreas.push_back( argv[i] );
                    }
                }
            }
			/* 
			 * TimePurge 
			 *
			 */
			else if (lstrcmpi( argv[i], _T("TIMEPURGE") ) == 0) {
				m_dwAction |= ACTION_TIMEPURGE;
				i++;
				if (i == argc) {
					logger << TEXT("hours for TimePurge not defined\n");
					m_dwAction |= ACTION_ERROR;
					m_iErrorCode = ERROR_TIMEPURGE_NO_HOURS;
					break;
				}
				if (!StrToInt( argv[i], m_timePurge.m_iHours ) 
					|| m_timePurge.m_iHours <= 0)
				{
					logger << TEXT("invalid hours value\n");
					m_dwAction |= ACTION_ERROR;
					m_iErrorCode = ERROR_TIMEPURGE_INVALID_HOURS;
					break;
				}
				i++;
				while (i < argc) {
					m_timePurge.m_vsMasks.push_back( argv[i] );
					i++;
				}
				break;
			}
			/* 
			 * BasePurge 
			 *
			 */
			else if (lstrcmpi( argv[i], _T("BASEPURGE") ) == 0) {
				m_dwAction |= ACTION_BASEPURGE;
				i++;
				if (i == argc) {
					logger << TEXT("days for BasePurge not defined\n");
					m_dwAction |= ACTION_ERROR;
					m_iErrorCode = ERROR_BASEPURGE_NO_DAYS;
					break;
				}
				if (!StrToInt( argv[i], m_basePurge.m_iDays ) 
					|| m_basePurge.m_iDays <= 0)
				{
					logger << TEXT("invalid days value\n");
					m_dwAction |= ACTION_ERROR;
					m_iErrorCode = ERROR_BASEPURGE_INVALID_DAYS;
					break;
				}
				i++;
				while (i < argc) {
					m_basePurge.m_vsAreas.push_back( argv[i] );
					i++;
				}
				break;
			}

			/* 
			 * Hand
			 *
			 */

			else if (lstrcmpi( argv[i], _T("HAND") ) == 0) {
				m_dwAction |= ACTION_HAND;
				i++;
				if (i == argc) {
					logger << TEXT("link address absent\n");
					m_dwAction |= ACTION_ERROR;
					m_iErrorCode = ERROR_HAND_NO_LINK;
					break;
				}
				m_handCommands.m_sAddr = argv[i];
				i++;
				if (i == argc) {
					logger << TEXT("no commands were specified\n");
					m_dwAction |= ACTION_ERROR;
					m_iErrorCode = ERROR_HAND_NO_COMMANDS;
					break;
				}
				while (i < argc) {
					m_handCommands.m_vsCommands.push_back( argv[i] );
					i++;
				}
				break;
			}
			// change group
			else if (lstrcmpi( argv[i], TEXT("CHANGEGROUP") ) == 0) {
				i += 2;
				if (i >= argc) {
					logger << TEXT("not enough arguments\n");
					m_dwAction |= ACTION_ERROR;
					m_iErrorCode = ERROR_FEW_ARGUMENTS;
					break;
				}
				actions.push_back( new CChangeGroupAction( argv[i-1], argv[i] ) );
				++i;
				break;
			}
            else if (lstrcmpi( argv[i], _T("HATCH") ) == 0) {
                m_dwAction |= ACTION_HATCH;
                i++;
                if (i == argc) {
                    logger << TEXT("file to hatch not defined\n");
                    m_dwAction |= ACTION_ERROR;
                    m_iErrorCode = ERROR_HATCH_NO_FILE_DEF;
                    break;
                }
                lstrcpy( m_szHatchFile, argv[i] );
                i++;
                if (i == argc) {
                    logger << TEXT("area to hatch not defined\n");
                    m_dwAction |= ACTION_ERROR;
                    m_iErrorCode = ERROR_HATCH_NO_AREA_DEF;
                    break;
                }
                lstrcpy( m_szHatchArea, argv[i] );
                i++;
                if (i == argc) {
                    logger << TEXT("description not defined\n");
                    m_dwAction |= ACTION_ERROR;
                    m_iErrorCode = ERROR_HATCH_NO_DESC_DEF;
                    break;
                }
                lstrcpy( m_szHatchDesc, argv[i] );
                m_szHatchReplaces[0] = _T('\0');
                m_fLeaveHatchedFile  = false;
				m_suspendedHatch     = false;
				m_deleteAfterHatch   = false;
                i++;
                while (i < argc && argv[i][0] == _T('-')) {
                    if (lstrcmpi( argv[i], TEXT("-r") ) == 0) {
                        // replaces
                        i++;
                        if (i == argc) {
                            logger << TEXT("Replaces not defined\n");
                            m_dwAction |= ACTION_ERROR;
                            m_iErrorCode = ERROR_HATCH_NO_REPLACES;
                            break;
                        }
                        lstrcpy( m_szHatchReplaces, argv[i] );
                        i++;
                    }
                    else if (lstrcmpi( argv[i], TEXT("-l") ) == 0) {
                    	m_fLeaveHatchedFile = true;
                    	i++;
                    }
					else if (lstrcmpi( argv[i], TEXT("-release") ) == 0) {
						m_suspendedHatch = true;
						// get release date (and possibly time)
						++i;
                        if (i == argc) {
                            logger << TEXT("Release date absent\n");
                            m_dwAction |= ACTION_ERROR;
                            m_iErrorCode = ERROR_NONE + 1;
                            break;
                        }
						::ZeroMemory( &m_releaseDateTime, sizeof(m_releaseDateTime) );
						if (!parseReleaseDate( argv[i], m_releaseDateTime )) {
                            logger << TEXT("Invalid release date\n");
                            m_dwAction |= ACTION_ERROR;
                            m_iErrorCode = ERROR_NONE + 1;
                            break;
						}
						++i;
						if (i < argc && argv[i][0] != TEXT('-')) {
							if (!parseReleaseTime( argv[i], m_releaseDateTime )) {
								logger << TEXT("Invalid release time\n");
								m_dwAction |= ACTION_ERROR;
								m_iErrorCode = ERROR_NONE + 1;
								break;
							}
							++i;
						}
					} else if (lstrcmpi( argv[i], TEXT("-delete") ) == 0) {
						m_deleteAfterHatch = true;
						++i;
				    } else {
                        logger << TEXT("unknown command line option ") << argv[i]+1 << TEXT("\n");
                        m_dwAction |= ACTION_ERROR;
                        m_iErrorCode = ERROR_UNKNOWN_OPT;
                        break;
                    }
                }
                if (i < argc && ((m_dwAction & ACTION_ERROR) == 0)) {
                    logger << TEXT("extra parameter(s) for hatch\n");
                    m_dwAction |= ACTION_ERROR;
                    m_iErrorCode = ERROR_HATCH_EXTRA_ARG;
                }
				if (m_deleteAfterHatch && m_fLeaveHatchedFile) {
					logger << TEXT("Two incompatible options were used: delete after hatch and leave hatched file\n");
                    m_dwAction |= ACTION_ERROR;
                    m_iErrorCode = ERROR_NONE+1;
				}
                break;
            }
            else {
                logger << TEXT("unknown command line parameter ") << argv[i] << TEXT("\n");
                m_dwAction |= ACTION_ERROR;
                m_iErrorCode = ERROR_UNKNOWN_ARG;
                break;
            }
        }
    }
}


int CAdrenalin::Execute() {
	// process actions
	{
		LoggerPtr	logger;

		while (!actions.empty()) {
			ActionPtr	action = actions.front();
			actions.pop_front();

			try {
				action->invoke();	// can throw CActionException
			} catch(CActionException ae) {
				logger << LOG_ERROR << TEXT("exception occured: ") << ae.getMessage() << TEXT("\n");
			}

			delete action;
		}
	}

	LoggerPtr	logger;
	logger << LOG_LOUD;

	if (m_dwAction & ACTION_ERROR) {		// prevously failed
		return (m_iErrorCode);
	}
	if (m_dwAction & ACTION_FIX) {		// fix
		logger << TEXT("Executing FIX command\n");
		ProcessFixRequests();
	}
	if (m_dwAction & ACTION_TOSS) {		// toss
		logger << TEXT("Executing TOSS command\n");
		ProcessInboundTics( CCommonData::getInstance()->getInboundPath().c_str(),
							CCommonData::getInstance()->getInboundPath().c_str() );
	}
	if (m_dwAction & ACTION_TOSS_BAD) {		// toss
		logger << TEXT("Executing TOSSBAD command\n");
		tossingBadTics = true;
		ProcessInboundTics( CCommonData::getInstance()->getBadTicsPath().c_str(),
							CCommonData::getInstance()->getInboundPath().c_str() );
		tossingBadTics = false;
	}
	if (m_dwAction & ACTION_RELINK) {		// relink
		logger << TEXT("Executing RELINK command\n");
		return Relink();
	}
	if (m_dwAction & ACTION_PURGE) {		// purge
		logger << TEXT("Executing PURGE command\n");
		Purge();
	}
	if (m_dwAction & ACTION_HATCH) {		// hatch
		logger << TEXT("Executing HATCH command\n");
		return (hatch());
	}
	if (m_dwAction & ACTION_DEBUG) {		// debug
		logger << TEXT("Executing DEBUG command\n");
		PrintDebugInfo();
	}
	if (m_dwAction & ACTION_TIMEPURGE) {	// TimePurge
		logger << TEXT("Executing TIMEPURGE command\n");
		TimePurge();
	}
	if (m_dwAction & ACTION_BASEPURGE) {	// purge based on data from files.db
		logger << TEXT("Executing BASEPURGE command\n");
		BasePurge();
	}
	if (m_dwAction & ACTION_HAND) {	// Hand
		logger << TEXT("Executing HAND command\n");
		ProcessFixRequests(true);
	}

	return 0;
}

/*
bool CAdrenalin::ParseLinkAddr(LPCTSTR s, SFTNAddress &Addr) {
    TCHAR	szFullAddr[24+MAX_DOMAIN_SIZE];
    TCHAR	szTempStr[24+MAX_DOMAIN_SIZE];
    TCHAR	szDomain[MAX_DOMAIN_SIZE];
    bool	fColonPresent, fSlashPresent, fPointPresent, fDogPresent;
    int		DogPos;
    int		i;

    if (lstrlen(s) >= 24+MAX_DOMAIN_SIZE)
        return (false);
    
    fColonPresent = fSlashPresent = fPointPresent = fDogPresent = false;
    i = 0;
    while (s[i] != _T('\0')) {
        switch (s[i]) {
        case _T(':'):
            fColonPresent = true;
            break;
        case _T('/'):
            fSlashPresent = true;
            break;
        case _T('.'):
            fPointPresent = true;
            break;
        case _T('@'):
            if (!fDogPresent) {
                fDogPresent = true;
                DogPos      = i;
            }
            break;
        }
        i++;
    }
    lstrcpy( szTempStr, s );
    if (fDogPresent && !fPointPresent) {
        i = lstrlen( szTempStr + DogPos + 1);
        if (i > MAX_DOMAIN_SIZE-1)
            i = MAX_DOMAIN_SIZE-1;
        lstrcpyn( szDomain, szTempStr + DogPos + 1, i+1 );
        szTempStr[DogPos] = _T('\0');
    }
    if (!fSlashPresent) {
        _stprintf( szFullAddr, _T("%lu/%s"), g_vars.m_addrMain.Net, szTempStr );
        lstrcpy( szTempStr, szFullAddr );
    }
    if (!fColonPresent) {
        _stprintf( szFullAddr, _T("%lu:%s"), g_vars.m_addrMain.Zone, szTempStr );
    }
    else
        lstrcpy( szFullAddr, szTempStr );

    if (!ParseFTNAddress( szFullAddr, Addr ))
        return (false);
    
    if (!fDogPresent) {
        lstrcpy( Addr.Domain, g_vars.m_addrMain.Domain );
    }
    else if (!fPointPresent) {
        lstrcpy( Addr.Domain, szDomain );
    }

    return (*(Addr.Domain) != _T('\0'));
}
*/

void CAdrenalin::MoveTicToBads(LPCTSTR szFullTicName, int TicNameFrom) {
    TCHAR szDestTicName[MAX_PATH];
    TCHAR szLogLine[LINE_SIZE];
    int  i, j, Len;
    bool	fNameFound;
    bool	fNameChanged;

	lstrcpy( szDestTicName, CCommonData::getInstance()->getBadTicsPath().c_str() );
    lstrcat( szDestTicName, szFullTicName + TicNameFrom );

    Len = lstrlen(szDestTicName);

    fNameFound   = false;
    fNameChanged = false;
    
    do {

        // trying to check the file
        if (_taccess( szDestTicName, 0) != 0)
            fNameFound = true;
        else {
            
            // form next name
            i = Len-1;
            j = 0;
        
            while (j <= 2) {
                if (szDestTicName[i] < _T('0')) {
                    szDestTicName[i] = _T('0');
                    break;
                }
            
                if (szDestTicName[i] < _T('9')) {
                    szDestTicName[i]++;
                    break;
                }
            
                if (szDestTicName[i] < _T('A')) {
                    szDestTicName[i] = _T('a');
                    break;
                }
                
                if (szDestTicName[i] < _T('Z')) {
                    szDestTicName[i]++;
                    szDestTicName[i] += _T('a') - _T('A');
                    break;
                }
            
                if (szDestTicName[i] < _T('a')) {
                    szDestTicName[i] = _T('a');
                    break;
                }
            
                if (szDestTicName[i] < _T('z')) {
                    szDestTicName[i]++;
                    break;
                }
        
                szDestTicName[i] = _T('0');
                i--;
                j++;
            }

            // check if failed to form new name
            if (j > 2)
                break;
            fNameChanged = true;
        }
    } while (!fNameFound);

	if (!fNameFound || MyMoveFile( szFullTicName, szDestTicName ) != FO_SUCCEEDED) {
		LoggerPtr	logger;
		logger << LOG_ERROR << TEXT("failed to move TIC into BadTics\n");
	}
    else {
		// TODO: should we return some result to log it outside of this method?
        if (fNameChanged) {
            _stprintf( szLogLine, TEXT("                TIC was renamed to %s and moved into BadTics\n"),
                     szDestTicName + TicNameFrom + Len - lstrlen(szFullTicName) );
        }
        else
            lstrcpy( szLogLine, TEXT("                TIC was moved into BadTics\n") );
		(*CLogger::getInstance()) << szLogLine;
    }
}


int CAdrenalin::hatch() {
	if (m_suspendedHatch) {
		return (suspendFiles() ? ERROR_NONE : (ERROR_NONE+1));
	}

	LoggerPtr	logger;

	// check hatch file
	if (m_szHatchFile[0] != TEXT('@')) {
		// we have only one file to hatch
		logger << LOG_LOUD << TEXT("Hatching single file...\n");
		return hatchSingleFile( m_szHatchArea, m_szHatchFile, m_szHatchDesc, m_fLeaveHatchedFile, m_szHatchReplaces );
	} else {
		// we have filelist
		logger << LOG_LOUD << TEXT("Executing filelist hatch...\n");

		// remember filelist name
		tstring	filelistName( m_szHatchFile+1 );

		VTString	fileNames, descriptions;
		if (!readHatchList( filelistName, fileNames, descriptions )) {
			return ERROR_NONE + 1;
		}
		if (fileNames.empty()) {
			// warn user about empty file (and log this information)
			logger << LOG_WARNING << TEXT("there are no files found in filelist\n");
		} else {
			int		hatchedFileCount = 0;
			tstring	fileDesc;
			for (size_t i = 0; i < fileNames.size(); i++) {
				fileDesc = (descriptions[i].empty() ? m_szHatchDesc : descriptions[i]);
				// do hatch file!
				if (hatchSingleFile( m_szHatchArea, fileNames[i], fileDesc, m_fLeaveHatchedFile, m_szHatchReplaces ) == ERROR_NONE) {
					hatchedFileCount++;
				}
			}

			// log total and hatched files count (and show counts to sysop)
			logger << TEXT("Hatch result (hatched/total files): [")
				   << hatchedFileCount << TEXT("/") << (int)fileNames.size()
				   << TEXT("]\n");
		}
	}

	return ERROR_NONE;
}


bool CAdrenalin::suspendFiles() {
	LoggerPtr	logger;

	VTString	fileNames, descriptions;

	if (m_szHatchFile[0] != TEXT('@')) {
		fileNames.push_back( m_szHatchFile );
		descriptions.push_back( m_szHatchDesc );
	} else {
		// we have filelist
		logger << LOG_LOUD << TEXT("Executing filelist hatch...\n");
		// remember filelist name
		tstring	filelistName( m_szHatchFile+1 );

		if (!readHatchList( filelistName, fileNames, descriptions )) {
			return false;
		}
	}

	SuspendedFiles	suspendedFiles;
	tstring			suspendedFileName = SUSPENDED_FILES_FILE_NAME;
	CAdrenalin::adjustFileName( suspendedFileName );
	tstring			suspendedDocumentEncoding;
	size_t			originalCount = 0;

	logger << LOG_LOUD << TEXT("Performing suspended hatch.\n");
	if (SuspendedHandler::doesFileExist( suspendedFileName )) {
		if (!SuspendedHandler::readSuspendedFiles( suspendedFileName, &suspendedFiles, 
				                                    suspendedDocumentEncoding ))
		{
			logger << LOG_ERROR << TEXT("Failed to read information about previously suspended files, hatch cancelled.\n");
			return false;
		} else {
			originalCount = suspendedFiles.size();
		}
	}

	for (size_t i = 0; i < fileNames.size(); i++) {
		TCHAR	fullName[MAX_PATH];
		LPTSTR	filePart;
		::GetFullPathName( fileNames[i].c_str(), MAX_PATH, fullName, &filePart );
		SuspendedFilePtr	file = new SuspendedFile();
		file->setFileName( fullName );
		file->setAreaName( m_szHatchArea );
		file->setDescription( (descriptions[i].empty() ? m_szHatchDesc : descriptions[i]) );
		file->setNoCopy( m_fLeaveHatchedFile );
		file->setDeleteAfterHatch( m_deleteAfterHatch );
		file->setReplaces( m_szHatchReplaces );
		file->setReleaseDateTime( m_releaseDateTime );
		suspendedFiles.push_back( file );
	}

	if (originalCount < suspendedFiles.size()) {
		// overwrite suspended files file
		SuspendedHandler::writeSuspendedFiles( suspendedFileName, &suspendedFiles, 
												suspendedDocumentEncoding );
		// print all newly suspended files
		logger << LOG_LOUD;
		TCHAR	dateTimeBuffer[128];
		while (originalCount < suspendedFiles.size()) {
			SuspendedFilePtr	file     = suspendedFiles[originalCount];
			const SYSTEMTIME&	dateTime = file->getReleaseDateTime();
			_stprintf( dateTimeBuffer, TEXT("%02d.%02d.%04d at %02d:%02d:%02d"),
					    (int)dateTime.wDay, (int)dateTime.wMonth, (int)dateTime.wYear,
					    (int)dateTime.wHour, (int)dateTime.wMinute, (int)dateTime.wSecond );
			logger << TEXT("File suspended to be released on ") << dateTimeBuffer << TEXT(":\n")
				    << TEXT("      file: ") << file->getFileName() << TEXT("\n")
					<< TEXT("      area: ") << file->getAreaName() << TEXT("\n")
					<< TEXT("      desc: ") << file->getDescription() << TEXT("\n")
					<< TEXT("   no copy: ") << (file->isNoCopy() ? TEXT("yes") : TEXT("no")) << TEXT("\n")
					<< TEXT("    delete: ") << (file->doDeleteAfterHatch() ? TEXT("yes") : TEXT("no")) << TEXT("\n")
					<< TEXT("  replaces: ") << file->getReplaces() << TEXT("\n");
			++originalCount;
		}
	} else {
		logger << LOG_LOUD << TEXT("No files to hatch.\n");
	}

	// delete suspended files (from memory)
	while (!suspendedFiles.empty()) {
		SuspendedFilePtr	file = suspendedFiles[suspendedFiles.size()-1];
		suspendedFiles.pop_back();
		delete file;
	}

	return true;
}


bool CAdrenalin::readHatchList(const tstring& listFileName, VTString& fileNames, VTString& descriptions) {
	fileNames.clear();
	descriptions.clear();

	// prepare to read lines
	COemLineReader	lineReader(listFileName.c_str());

	tstring	fileName, fileDesc;

	if (lineReader.isOpened()) {
		// read filelist and hatch every file

		while (lineReader.next()) {
			// get next line
			LPCTSTR	line = lineReader.getLine();
			int		i, j;
			// extract file name and description
			// file name can be enclosed in double quotes
			TCHAR	separator1, separator2;
			if (line[0] == TEXT('"')) {
				// we suppose double quoted file name
				separator1 = TEXT('"');
				separator2 = ZERO_CHAR;
				// start loop from the second character
				i = 1;
			}
			else {
				// we suppose that file name is not double quoted
				separator1 = SPACE_CHAR;
				separator2 = TAB_CHAR;
				// start loop from the first character
				i = 0;
			}
			j = i;
			while (line[i] != separator1 && line[i] != separator2 && line[i] != ZERO_CHAR) {
				i++;
			}
			if (i > j) {

				// we have file name
				fileName.assign( line + j, i-j );
				fileDesc.clear();

				// try extract description

				if (line[i] != ZERO_CHAR) {
					// skip any separator
					i++;
				}
				// skip leading spaces and tabs
				while (line[i] == SPACE_CHAR || line[i] == TAB_CHAR) {
					i++;
				}
				j = i;
				// skip tailing spaces and tabs
				int	firstSpaceInSeq = -1;
				while (line[i] != ZERO_CHAR) {
					if (line[i] == SPACE_CHAR || line[i] == TAB_CHAR) {
						// we are on space or tab
						if (firstSpaceInSeq < 0) {
							// set space sequence start
							firstSpaceInSeq = i;
						}
					}
					else {
						// set space sequence start
						firstSpaceInSeq = -1;
					}
					i++;
				}
				if (firstSpaceInSeq >= 0)
					i = firstSpaceInSeq;
				if (i > j) {
					// we have description
					fileDesc = tstring( line + j, i-j );
				}

				// now we have file name and file description (description can be empty)
				// check description length (compare with max tic line size not including "DESC ")
				if (fileDesc.size() > (MAX_TIC_LINE_SIZE - 5)) {
					LoggerPtr	logger;
					logger << LOG_WARNING
						   << TEXT("description of \"") << fileName << TEXT("\" too long (longer than ")
						   << (MAX_TIC_LINE_SIZE - 5) << TEXT(" symbols), truncated\n");

					fileDesc.resize( MAX_TIC_LINE_SIZE - 5 );
				}

				fileNames.push_back( fileName );
				descriptions.push_back( fileDesc );
			}
		}
		return true;
	} else {
		// notify user about failure and log the failure
		LoggerPtr	logger;
		logger << LOG_ERROR << TEXT("failed to read ") << listFileName << TEXT("\n");
		return false;
	}
}


void CAdrenalin::AddStartupPath(LPTSTR szLocation) {
	TCHAR	szTempPath[MAX_PATH];

	if (szLocation[0] == _T('\0') || szLocation[1] != _T(':')) {
		lstrcpy( szTempPath, szLocation );
		ConcatStrings( szLocation, m_szStartupPath, szTempPath );
	}
}


CAdrenalin& CAdrenalin::getInstance() {
	return (*pAdrenalin);
}

struct	SAreaFiles {
	TCHAR		m_szAreaName[AREA_NAME_SIZE];
	VTString	m_vFiles;
};

void CAdrenalin::Purge() {
	WIN32_FIND_DATA		FindFileData;
	HANDLE			hFindFile;
	BOOL			fFileFound;
	TCHAR			szFindFileMask[MAX_PATH];
	TCHAR			szFullFileName[MAX_PATH];
	TCHAR			szFileName[MAX_PATH];
	int			iNamePos;
	CTic			tic;
	vector<SAreaFiles>	vAreaFiles;
	VTString		vCurFile;
	ifstream		in;
	ofstream		tempout;
	TCHAR			szTempFileName[MAX_PATH];
	TCHAR			szBuf[LINE_SIZE];
	int			i;
	int				nPurgedNumber, nTotalPurgedNumber;
	int				nPurgedSize,   nTotalPurgedSize;
	bool				fSkipLongDesc;
	VTString::iterator		itsCurPurgeArea;
	bool				fHaveToAddFile;
	bool				fAreaAsMask;
	vector<SAreaFiles>::iterator	itsArea;

	// TODO: check!!!

	LoggerPtr	logger;
	logger << LOG_LOUD;

	if (!m_vsPurgeAreas.empty()) {
		for (itsCurPurgeArea = m_vsPurgeAreas.begin(); itsCurPurgeArea != m_vsPurgeAreas.end();
		     itsCurPurgeArea++)
		{
			if ((*itsCurPurgeArea).c_str()[0] == _T('-')) { // exclude from arealist
				itsArea = vAreaFiles.begin();
				while (itsArea != vAreaFiles.end()) {
					if (CheckMask( (*itsArea).m_szAreaName,
							(TCHAR*)(*itsCurPurgeArea).c_str() + 1 ))
					{
						// remove area from list
						vAreaFiles.erase( itsArea );
						itsArea = vAreaFiles.begin();
					}
					else
						itsArea++;
				}
			}
			else { // add to arealist, if one or more areas meet mask
				i = 0; // number of added areas
				fAreaAsMask = IsMask( (*itsCurPurgeArea).c_str() );
				int	areasCount = CAreasManager::getInstance()->getAreasCount();
				for (i = 0; i < areasCount; i++) {
					PArea	area = CAreasManager::getInstance()->getArea( i );
					if (fAreaAsMask) {
						if (area->hasFlag( CArea::FLAG_VETO_MANUAL_PURGE ))
							continue;
						if (!CheckMask( area->getName().c_str(),
							(TCHAR*)(*itsCurPurgeArea).c_str()))
							continue;
					}
					else 
						if (stricmp( area->getName().c_str(),
							(*itsCurPurgeArea).c_str()) != 0)
							continue;
					// add only unique areas
					vector<SAreaFiles>::iterator	itsArea = vAreaFiles.begin();
					while (itsArea != vAreaFiles.end()) {
						if (stricmp( (*itsArea).m_szAreaName, area->getName().c_str() ) == 0)
							break;
						itsArea++;
					}
					if (itsArea == vAreaFiles.end()) { // not found
						SAreaFiles	newAreaFiles;
						strcpy( newAreaFiles.m_szAreaName,
								area->getName().c_str() );
						vAreaFiles.push_back( newAreaFiles );
						i++;
					}
				}
				if (i == 0) { // no areas added
					if (fAreaAsMask)
						_stprintf( szBuf,
							_T("! No purgable areas found for mask %s\n"),
							(*itsCurPurgeArea).c_str() );
					else
						_stprintf( szBuf, _T("! Area %s not found.\n"),
							(*itsCurPurgeArea).c_str() );
					logger.rememberMode();
					logger << szBuf;
					logger.restoreMode();
				}
			}
		}
	}
    
	lstrcpy( szFindFileMask, CCommonData::getInstance()->getNewTicsPath().c_str() );
	lstrcat( szFindFileMask, _T("*.ti?") );

	hFindFile = ::FindFirstFile( szFindFileMask, &FindFileData );
	fFileFound = (hFindFile != INVALID_HANDLE_VALUE);
	lstrcpy( szFullFileName, CCommonData::getInstance()->getNewTicsPath().c_str() );
	iNamePos = lstrlen( szFullFileName );

	while (fFileFound) {
		lstrcpy( szFullFileName + iNamePos, FindFileData.cFileName );
		if (tic.Read( szFullFileName )) {
			fHaveToAddFile = false;
			do {
				PArea	area = CAreasManager::getInstance()->getArea( tic.m_szArea );
				if (area == NULL)
					break;
				// try to find area
				itsArea = vAreaFiles.begin();
				while (itsArea != vAreaFiles.end()) {
					if (lstrcmpi( (*itsArea).m_szAreaName, tic.m_szArea) == 0)
						break;
					itsArea++;
				}
				if (itsArea != vAreaFiles.end()) { // area found
					fHaveToAddFile = true;
					break;
				}
				if (m_vsPurgeAreas.empty() && area->hasFlag( CArea::FLAG_PASSTHROUGH ))
				{
					// we will add area
					fHaveToAddFile = true;
					break;
				}
				break;
			} while (true);
			if (fHaveToAddFile) {
				// prepare file name
				if (*tic.m_szFullName != _T('\0'))
					lstrcpy( szFileName, tic.m_szFullName );
				else
					lstrcpy( szFileName, tic.m_szFile );
				// add file into area files list,
				// with this purpose find area or add it, if it's absent
				if (itsArea == vAreaFiles.end()) {
					// add area
					SAreaFiles	newAreaFiles;
					lstrcpy( newAreaFiles.m_szAreaName, tic.m_szArea );
					// add file
					newAreaFiles.m_vFiles.push_back( szFileName );
					vAreaFiles.push_back( newAreaFiles );
				}
				else {
					// try to find file to exclude duplicates
					VTString::iterator	itsCurFile = (*itsArea).m_vFiles.begin();
					while (itsCurFile != (*itsArea).m_vFiles.end()) {
						if (lstrcmpi( (*itsCurFile).c_str(), szFileName ) == 0)
							break;
						itsCurFile++;
					}
					if (itsCurFile == (*itsArea).m_vFiles.end()) {
						// add file
						(*itsArea).m_vFiles.push_back( szFileName );
					}
				}
			}
		}
		fFileFound = ::FindNextFile( hFindFile, &FindFileData );
	}
	if (hFindFile != INVALID_HANDLE_VALUE)
		::FindClose( hFindFile );

	// add passthrough areas if not directly specified
	if (m_vsPurgeAreas.empty()) {
		int	areasCount = CAreasManager::getInstance()->getAreasCount();
		for (int i = 0; i < areasCount; i++) {
			PArea	area = CAreasManager::getInstance()->getArea( i );
			if (area->hasFlag( CArea::FLAG_PASSTHROUGH )) {
				itsArea = vAreaFiles.begin();
				while (itsArea != vAreaFiles.end()) {
					if (lstrcmpi( (*itsArea).m_szAreaName, area->getName().c_str() ) == 0)
						break;
					itsArea++;
				}
				if (itsArea == vAreaFiles.end()) {
					// add area
					SAreaFiles	newAreaFiles;
					lstrcpy( newAreaFiles.m_szAreaName, area->getName().c_str() );
					vAreaFiles.push_back( newAreaFiles );
				}
			}
		}
	}

	// go through all selected areas
	nTotalPurgedSize   = 0;
	nTotalPurgedNumber = 0;
	itsArea      = vAreaFiles.begin();
	while (itsArea != vAreaFiles.end()) {
		_stprintf( szBuf, _T("* Area: %s\n"), (*itsArea).m_szAreaName );
		logger << szBuf;
		// select area
		PArea	area = CAreasManager::getInstance()->getArea( (*itsArea).m_szAreaName );
		// prepare directory
		lstrcpy( szFullFileName, area->getWorkingPath().c_str() );
		iNamePos = lstrlen( szFullFileName );
		// prepare mask to look files
		lstrcpy( szFindFileMask, area->getWorkingPath().c_str() );
		lstrcat( szFindFileMask, _T("*.*") );
		// process fecho files
		nPurgedNumber = 0; // number of removed files
		nPurgedSize   = 0; // size of purged files in this area
		hFindFile    = FindFirstFile( szFindFileMask, &FindFileData );
		fFileFound    = (hFindFile != INVALID_HANDLE_VALUE);
		while (fFileFound) {
			// skip directories, special files and "files.bbs"
			if (!(FindFileData.dwFileAttributes &
				(FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_HIDDEN |
				FILE_ATTRIBUTE_OFFLINE | FILE_ATTRIBUTE_READONLY |
				FILE_ATTRIBUTE_TEMPORARY))
				&&
				lstrcmpi( FindFileData.cFileName, CCommonData::getInstance()->getFileListName().c_str() ) != 0)
			{
				// let's check, if we have TIC for this file
				VTString::iterator	itsCurFile = (*itsArea).m_vFiles.begin();
				while (itsCurFile != (*itsArea).m_vFiles.end()) {
					if (lstrcmpi( FindFileData.cFileName, (*itsCurFile).c_str()) == 0)
						break;
					itsCurFile++;
				}
				// remove file, if it's not used
				if (itsCurFile == (*itsArea).m_vFiles.end()) {
					// form full name
					lstrcpy( szFullFileName + iNamePos, FindFileData.cFileName );
					if (_tunlink( szFullFileName ) == 0) {
						_stprintf( szBuf, _T("        Removed file %s\n"),
							FindFileData.cFileName );
						nPurgedNumber++;
						nPurgedSize += (FindFileData.nFileSizeHigh * MAXDWORD) +
							FindFileData.nFileSizeLow;
						logger << szBuf;
					}
					else {
						logger.rememberMode();
						logger << LOG_ERROR << TEXT("failed to remove file ")
							   << FindFileData.cFileName << TEXT("\n");
						logger.restoreMode();
					}
				}
			}
			fFileFound = FindNextFile( hFindFile, &FindFileData );
		}
		if (hFindFile != INVALID_HANDLE_VALUE)
			FindClose( hFindFile );

		_stprintf( szBuf, _T("        %ld file(s) of size "), nPurgedNumber );
		FormatSize( nPurgedSize, szBuf + lstrlen(szBuf) );
		lstrcat( szBuf, _T(" byte(s).\n") );
		logger << szBuf;
		// change "files.bbs"
		// set file names
		lstrcpy( szFullFileName + iNamePos, CCommonData::getInstance()->getFileListName().c_str() );
		lstrcpy( szTempFileName, area->getWorkingPath().c_str() );
		lstrcat( szTempFileName, _T("adrenalin.tmp") );
		// check names for spaces
		VTString::iterator	itsCurFile = (*itsArea).m_vFiles.begin();
		while (itsCurFile != (*itsArea).m_vFiles.end()) {
			for (unsigned int index = 0; index < (*itsCurFile).length(); index++) {
				if ((*itsCurFile).c_str()[index] == _T(' '))
					break;
			}
			if ((*itsCurFile).c_str()[i] != _T('\0')) {
				lstrcpy( szFileName, _T("\"") );
				lstrcat( szFileName, (*itsCurFile).c_str() );
				lstrcat( szFileName, _T("\"") );
				(*itsCurFile) = szFileName;
			}
			itsCurFile++;
		}
		
		// copy desc file into temp file, but skip lines with absent files
		bool	fOverwriteSuccess = true;
		do {
			// check temp file name
			if (!GetTmpFileName( szTempFileName ))
				break;
			in.open( szFullFileName );
			if (!in)
				break;
			tempout.open( szTempFileName );
			if (!tempout) {
				in.close();
				break;
			}
			fSkipLongDesc = false;
			while (!in.eof()) {
				in.getline( szBuf, LINE_SIZE );
				if (!in) {
					if (!in.eof())
						break;
					continue;
				}
				// check file name
				itsCurFile = (*itsArea).m_vFiles.begin();
				while (itsCurFile != (*itsArea).m_vFiles.end()) {
					if (_tcsnicmp( szBuf, (*itsCurFile).c_str(),
						(*itsCurFile).length()) == 0)
						break;
					itsCurFile++;
				}
				if (itsCurFile == (*itsArea).m_vFiles.end())
					fSkipLongDesc = true;
				else {
					if (fSkipLongDesc) {
						if (*szBuf != _T(' ') && *szBuf != _T('>') && *szBuf != _T('+')) {
							fSkipLongDesc = false;
							tempout << szBuf << endl;
							if (!tempout) {
								fOverwriteSuccess = false;
								break;
							}
						}
					}
					else {
						tempout << szBuf << endl;
						if (!tempout) {
							fOverwriteSuccess = false;
							break;
						}
					}
				}
			}
			in.close();
			tempout.close();
			if (fOverwriteSuccess) {
				_tunlink( szFullFileName );
				_trename( szTempFileName, szFullFileName );
			}
			else {
				_tunlink( szTempFileName );
				logger.rememberMode();
				logger << LOG_ERROR << TEXT("failed to update file ") << szFullFileName << TEXT("\n");
				logger.restoreMode();
			}
			break;
		} while (true);
		// remove "files.bbs", if it's empty
		hFindFile = FindFirstFile( szFullFileName, &FindFileData );
		fFileFound = (hFindFile != INVALID_HANDLE_VALUE);
		if (fFileFound) {
			FindClose( hFindFile );
			if (((FindFileData.nFileSizeHigh * MAXDWORD) +
				FindFileData.nFileSizeLow) == 0)
			_tunlink( szFullFileName);
		}
		nTotalPurgedNumber += nPurgedNumber;
		nTotalPurgedSize   += nPurgedSize;
		// free item memory
		(*itsArea).m_vFiles.clear();
		itsArea++;
	}
    
	_stprintf( szBuf, _T("Total: %ld file(s) of size "), nTotalPurgedNumber );
	FormatSize( nTotalPurgedSize, szBuf + lstrlen(szBuf) );
	lstrcat( szBuf, _T(" byte(s).\n") );
	logger << szBuf;
}

int CAdrenalin::Relink() {
	CMsg		msg;
	VTString	vBodyLines;
	TCHAR		szLine[256];
	int		nAreas;

	LoggerPtr	logger;

	// prepare link address;
	CAddress	relinkAddress;
	if (!relinkAddress.parseString( relinkData.address.c_str() )) {
		logger << LOG_ERROR
			   << TEXT("failed to parse RELINK address ")
			   << relinkData.address
			   << TEXT("\n");
		return ERROR_WRONG_ADDRESS;
	}
	relinkAddress.baseUpon( CCommonData::getInstance()->getSysopAddress() );
	lstrcpy( szLine, relinkAddress.toString().c_str() );

	// check link
	PLink	link = CLinksManager::getInstance()->getLink(relinkAddress);
	if (link == NULL) {
		logger << LOG_ERROR
			   << TEXT("Relink failed because of unknown link: ")
			   << szLine << TEXT("\n");
		return ERROR_WRONG_ADDRESS;
	}

	logger << LOG_LOUD;

	// print log message
	logger << TEXT("Relink: ") << szLine << TEXT("\n");

	nAreas = 0;
	int	areasCount = CAreasManager::getInstance()->getAreasCount();
	for (int i = 0; i < areasCount; i++) {
		PArea	area = CAreasManager::getInstance()->getArea( i );
		TiePtr	tie  = area->getTie( link );
		if (tie != NULL) {
			// add msg line
			_stprintf( szLine, TEXT("+%s"), area->getName().c_str() );
			vBodyLines.push_back( szLine );
			nAreas++;
			// print log message
			logger << TEXT("  ") << szLine << TEXT("\n");
		}
	}

	logger << nAreas << " area(s) added into the relink request.\n";

	if (nAreas > 0) {
		vBodyLines.push_back( TEXT("") );
		MakeDirStructure( CCommonData::getInstance()->getNetmailPath().c_str() );
		// from/to fields reversed because we will use reply method
		msg.setFromAddress( link->getAddress() );
		msg.setToAddress( link->getOurAddress() );
		// check distant robot name
		if (relinkData.robotName.size() == 0) {
			if (link->getRobotName().length() == 0) {
				msg.setSenderName( "AllFix" );
			} else {
				msg.setSenderName( link->getRobotName().c_str() );
			}
		} else {
			msg.setSenderName( relinkData.robotName.c_str() );
		}
		msg.setReceiverName( CCommonData::getInstance()->getSysopName() );
		msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0, link->getPassword().c_str(), vBodyLines );
		logger << TEXT("Relink request message created.\n");
	}
	else {
		logger << TEXT("There is no need to create relink request message because of no areas linked.\n");
	}

	return ERROR_NONE;
}

/*
SFTNAddress& CAdrenalin::GetOurAddr(int iOurAddr) {
	if (iOurAddr == MAIN_ADDRESS)
		return (g_vars.m_addrMain);
	return (g_vars.m_vAKAAddr[iOurAddr]);
}
*/

void CAdrenalin::LoadForwards() {
	int				i;
	TCHAR			szFileName[MAX_PATH];
	LPTSTR			szFrom;
	SForwardInfo	fwdInfo;
	int				nCurrentDays;

	// 0. Clear vector and prepare time
	m_vectForwardInfo.clear();
	SYSTEMTIME	systemTime;
	GetLocalTime( &systemTime );
	nCurrentDays = GetDays( systemTime.wDay, systemTime.wMonth, systemTime.wYear %= 100 );

	// 1. Prepare filename
	ConcatStrings( szFileName, m_szStartupPath, c_szForwardsFileName );
	// 2. Open file for reading
	ifstream	in(szFileName);
	if (!in.fail()) {
		// 3. consequently read lines
		TCHAR	szLine[LINE_SIZE];
		while (!in.eof()) {
			in.getline( szLine, LINE_SIZE );
			if (in.fail())
				break;
			// 3.1. Get date
			lstrcpyn( fwdInfo.m_szDate, szLine, 7 );
			// 3.1.1. Check date 
			if (CCommonData::getInstance()->getForwardExpirationTime() > 0) {
				WORD	wDay   = (fwdInfo.m_szDate[0] - _T('0'))*10 + (fwdInfo.m_szDate[1] - _T('0'));
				WORD	wMonth = (fwdInfo.m_szDate[2] - _T('0'))*10 + (fwdInfo.m_szDate[3] - _T('0'));
				WORD	wYear  = (fwdInfo.m_szDate[4] - _T('0'))*10 + (fwdInfo.m_szDate[5] - _T('0'));
				int	nDays = GetDays( wDay, wMonth, wYear );
				// skip expired request
				if (nCurrentDays - nDays > CCommonData::getInstance()->getForwardExpirationTime()) {
					m_fForwardsChanged = true;
					continue;
				}
			}
			// 3.2. Get link
			i = 0;
			szFrom = szLine + 7;
			while (*szFrom != _T(' ')) {
				fwdInfo.m_szLink[i] = *szFrom;
				i++;
				szFrom++;
			}
			fwdInfo.m_szLink[i] = _T('\0');
			// 3.3. Get area
			i = 0;
			szFrom++;
			while (*szFrom != _T('\0')) {
				fwdInfo.m_szArea[i] = *szFrom;
				i++;
				szFrom++;
			}
			fwdInfo.m_szArea[i] = _T('\0');
			m_vectForwardInfo.push_back(fwdInfo);
		}
		in.close();
	}

	// in spite of read file result we should set flag
	m_fForwardsRead = true;
}

void CAdrenalin::SaveForwards() {
	TCHAR	szForwardsFileName[MAX_PATH];
	CFile	forwardsFile;

	// 1. Prepare filename
	ConcatStrings( szForwardsFileName, m_szStartupPath, c_szForwardsFileName );
	// 2. Open file for writing
	if (forwardsFile.Open( szForwardsFileName, GENERIC_WRITE, CREATE_ALWAYS )) {
		CForwardInfoVector::iterator	i;
		for (i = m_vectForwardInfo.begin(); i != m_vectForwardInfo.end(); ++i) {
			forwardsFile.WriteString( (*i).m_szDate );
			forwardsFile.WriteString( _T(" ") );
			forwardsFile.WriteString( (*i).m_szLink );
			forwardsFile.WriteString( _T(" ") );
			forwardsFile.WriteString( (*i).m_szArea );
			forwardsFile.WriteString( _T("\r\n") );
		}
		forwardsFile.Close();
	} else {
		LoggerPtr	logger;
		logger << LOG_ERROR << TEXT("failed to write forwards file: ") << szForwardsFileName
			   << TEXT("\n");
	}
}

void CAdrenalin::GetErrorMessage(int nErrorCode, LPTSTR szMessage) {
	switch (nErrorCode) {
	case ERROR_UNKNOWN_ARG:
	    lstrcpy( szMessage, _T("Unknown argument") );
	    break;
	case ERROR_HATCH_NO_FILE_DEF:
	    lstrcpy( szMessage, _T("No file defined to hatch") );
	    break;		
	case ERROR_HATCH_NO_AREA_DEF:
	    lstrcpy( szMessage, _T("No area defined to hatch") );
	    break;
	case ERROR_HATCH_EXTRA_ARG:
	    lstrcpy( szMessage, _T("Extra arguments for hatch") );
	    break;
	case ERROR_SYNTAX:
	    lstrcpy( szMessage, _T("Syntax error") );
	    break;
	case ERROR_NUMBER:
	    lstrcpy( szMessage, _T("Wrong number of arguments") );
	    break;
	case ERROR_DUPLICATE:
	    lstrcpy( szMessage, _T("Duplicate usage of a keyword") );
	    break;
	case ERROR_FILE_ABSENT:
	    lstrcpy( szMessage, _T("File absent") );
	    break;
	case ERROR_UNKNOWN_KEYWORD:
	    lstrcpy( szMessage, _T("Unknown keyword") );
	    break;
	case ERROR_FORWARD_DEF:
	    lstrcpy( szMessage, _T("Forward definition") );
	    break;
	case ERROR_ABSENT_DEF:
	    lstrcpy( szMessage, _T("Definition is absent") );
	    break;
	case ERROR_MEMORY:
	    lstrcpy( szMessage, _T("Not enough memory") );
	    break;
	case ERROR_AREA_NOT_FOUND:
	    lstrcpy( szMessage, _T("Area not found") );
	    break;
	case ERROR_MOVE_FAILED:
	    lstrcpy( szMessage, _T("Failed to move file") );
	    break;
	case ERROR_CRC32_FAILED:
	    lstrcpy( szMessage, _T("Failed to calculate CRC32") );
	    break;
	case ERROR_UNKNOWN_AKA:
	    lstrcpy( szMessage, _T("Unknown AKA") );
	    break;
	case ERROR_LNG_READ:
	    lstrcpy( szMessage, _T("Error reading language file") );
	    break;
	case ERROR_LNG_FILE_NOT_FOUND:
	    lstrcpy( szMessage, _T("Language file not found") );
	    break;
	case ERROR_LNG_NUMBER:
	    lstrcpy( szMessage, _T("Wrong number of language messages") );
	    break;
	case ERROR_HATCH_NO_DESC_DEF:
	    lstrcpy( szMessage, _T("Description not defined for hatching") );
	    break;
	case ERROR_HATCH_NO_REPLACES:
	    lstrcpy( szMessage, _T("REPLACES value absent for hatching") );
	    break;
	case ERROR_UNKNOWN_OPT:
	    lstrcpy( szMessage, _T("Unknown option") );
	    break;
	case ERROR_LOW_DISK_SPACE:
	    lstrcpy( szMessage, _T("Low disk space") );
	    break;
	case ERROR_UNDEFINED_ADDRESS:
	    lstrcpy( szMessage, _T("Undefined address for relink") );
	    break;
	case ERROR_WRONG_ADDRESS:
	    lstrcpy( szMessage, _T("Wrong relink address") );
	    break;
	case ERROR_TIMEPURGE_NO_HOURS:
	    lstrcpy( szMessage, _T("Hours for TimePurge not defined") );
	    break;
	case ERROR_TIMEPURGE_INVALID_HOURS:
	    lstrcpy( szMessage, _T("Invalid hours value for TimePurge") );
	    break;
	case ERROR_BASEPURGE_NO_DAYS:
		lstrcpy( szMessage, _T("Days for BasePurge weren't defined") );
		break;
	case ERROR_BASEPURGE_INVALID_DAYS:
		lstrcpy( szMessage, _T("Invalid days value for BasePurge") );
		break;
	case ERROR_HAND_NO_LINK:
		lstrcpy( szMessage, _T("Link isn't specified for Hand command") );
		break;
	case ERROR_HAND_NO_COMMANDS:
		lstrcpy( szMessage, _T("No message commands for Hand command") );
		break;
	default:
		lstrcpy( szMessage, _T("Unknown error") );
	}
}


void CAdrenalin::TimePurge() {
	// if there in no masks, we shouldn't check links
	bool		fCheckByMask = !m_timePurge.m_vsMasks.empty();
	TCHAR		szAddr[128];
	CBSOManager	bsoManager;
	VTString	vsMsgBodyLines;
	VTString	vsbsoLines;
	vector<bool>	vbbsoLeft;
	TCHAR		szBuf[MAX_PATH];
	::ifstream	in;
	CTic		tic;
	int		iNewTicsPathLen;
	__int64		i64TresholdTime;
	TCHAR		szFileName[MAX_PATH];
	TCHAR		szShortFileName[MAX_PATH];
	TCHAR		szTicName[MAX_PATH];
	TCHAR		szShortNewTicsPath[MAX_PATH];

	bool			fbsoChanged;			// true, if bso was changed as a result of removing tics and files
	LARGE_INTEGER	li;
	unsigned int	i, j;
	TCHAR			szMsgHead[4096];
	TCHAR			szMsgInfo[4096];

	// precalc some vars
	{
		FILETIME	ftCurrent;
		GetSystemTimeAsFileTime( &ftCurrent );
		__int64	i64Hours = ((__int64)m_timePurge.m_iHours)*10000000*60*60;
		li.LowPart  = ftCurrent.dwLowDateTime;
		li.HighPart = ftCurrent.dwHighDateTime;
		i64TresholdTime = li.QuadPart - i64Hours;
	}
	ConvertNameToShort( CCommonData::getInstance()->getNewTicsPath().c_str(), szShortNewTicsPath );
	iNewTicsPathLen = lstrlen(szShortNewTicsPath);
	MakeDirStructure( CCommonData::getInstance()->getNetmailPath().c_str() );
	_stprintf( szMsgHead, m_language[LNG_TIMEPURGE_HEAD], m_timePurge.m_iHours );

	LoggerPtr	logger;

	for (CLinksManager::CIterator	itLink = CLinksManager::getInstance()->begin();
	itLink != CLinksManager::getInstance()->end(); ++itLink) {
		// get next link
		PLink	link = (*itLink);

		// check address by mask if have to
		if (fCheckByMask) {
			lstrcpy( szAddr, link->getAddress().toString().c_str() );
			VTString::iterator	vit = m_timePurge.m_vsMasks.begin();
			while (vit != m_timePurge.m_vsMasks.end()) {
				if (CheckMask( szAddr, (*vit).c_str() ))
					break;
				vit++;
			}
			// if no appropriate mask was found, make next loop
			if (vit == m_timePurge.m_vsMasks.end())
				continue;
		}

		// log link address
		logger << LOG_LOUD << TEXT("Link: ") << link->getAddress().toString() << TEXT("\n");

		// init bso manager
		bsoManager.setAddress( link->getAddress(), link->getFlavourType() );

		// set busy flag
		if (!bsoManager.setBusyFlag()) {
			logger << LOG_ERROR << TEXT("failed to create busy flag in BSO\n");
			continue;
		}

		// prepare vectors
		vsMsgBodyLines.clear();
		vsMsgBodyLines.push_back( szMsgHead );
		vsbsoLines.clear();
		vbbsoLeft.clear();

		// read bso file into the memory
		lstrcpy( szBuf, bsoManager.getName() );
		in.open( szBuf );
		if (in) {
			while (!in.eof()) {
				in.getline( szBuf, MAX_PATH );
				if (!in)
					break;
				vsbsoLines.push_back( szBuf );
				vbbsoLeft.push_back( true );
			}
			in.close();
		} else {
			logger << LOG_ERROR << TEXT("failed to open BSO file fo reading:\n") << szBuf << TEXT("\n");
		}

		// check presense of lines
		if (!vsbsoLines.empty()) {
			fbsoChanged = false;
			// go throught vsbsoLines
			for (i = 0; i < vsbsoLines.size(); i++) {
				// skip already erased lines
				if (!vbbsoLeft[i])
					continue;
				tstring&	s = vsbsoLines[i];
				// check 2 conditions:
				// 1) tic is in NEWTICS directory
				// 2) it's creation date is less than current-hours
				if (s.length() == 0 || s[0] != _T('^')) {
					continue;
				}
				lstrcpy( szTicName, s.c_str() + 1 );
				if (_tcsncmp( szTicName, 
						  szShortNewTicsPath,
						  iNewTicsPathLen ) == 0 &&
					!HaveSlashes(szTicName + iNewTicsPathLen))
				{
					// first condition is true
					// check second condition
					HANDLE	hTicFile = CreateFile( szTicName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
					if (hTicFile == INVALID_HANDLE_VALUE) {
						logger << LOG_ERROR << TEXT("failed to open TIC file for reading:\n")
							   << szTicName << TEXT("\n");
					}
					else {
						// check time
						FILETIME	ftTic;
						if (!GetFileTime( hTicFile, &ftTic, NULL, NULL )) {
							CloseHandle( hTicFile );
							logger << LOG_ERROR << TEXT("failed to get filetime of TIC file:\n")
								   << szTicName << TEXT("\n");
						} else {
							CloseHandle( hTicFile );
							li.LowPart  = ftTic.dwLowDateTime;
							li.HighPart = ftTic.dwHighDateTime;
							__int64	i64TicTime = li.QuadPart;
							if (i64TicTime <= i64TresholdTime) {
								// tic is obsolete
								logger << LOG_LOUD << TEXT("TIC file is obsolete: ") << szTicName << TEXT("\n");
								// read tic
								if (!tic.Read(szTicName)) {
									logger << LOG_ERROR << TEXT("failed to read TIC file:\n")
										   << szTicName << TEXT("\n");
								} else {
									// we doesn't need this tic anymore
									vbbsoLeft[i] = false;
									fbsoChanged = true;
									if (!DeleteFile( szTicName )) {
										logger << LOG_ERROR
											   << TEXT("failed to remove TIC file:\n")
											   << szTicName << TEXT("\n");
									}
									PArea	area = CAreasManager::getInstance()->getArea( tic.m_szArea );
									if (area == NULL) {
										logger << LOG_ERROR
											   << TEXT("area ") << tic.m_szArea << TEXT(" not found\n");
									} else {
										// form full file name
										lstrcpy( szFileName, area->getWorkingPath().c_str() );
										if (*tic.m_szFullName != _T('\0'))
											lstrcat( szFileName, tic.m_szFullName );
										else
											lstrcat( szFileName, tic.m_szFile );
										ConvertNameToShort( szFileName, szShortFileName );
										// erase line with filename from bso lines
										bool	fErased = false;
										for (j = 0; j < vsbsoLines.size(); j++) {
											if (lstrcmp( vsbsoLines[j].c_str(), szShortFileName ) == 0) {
												vbbsoLeft[j] = false;
												fErased = true;
												break;
											}
										}
										if (!fErased) {
											logger << LOG_ERROR
												   << TEXT("file not found in BSO:\n")
												   << szShortFileName << TEXT("\n");
										} else {
											logger << LOG_LOUD << TEXT("file removed from BSO:\n")
												   << szShortFileName << TEXT("\n");
											_stprintf( szMsgInfo, m_language[LNG_TIMEPURGE_INFO], 
												tic.m_szArea, 
												(tic.m_szFullName[0] == _T('\0')) ? tic.m_szFile : tic.m_szFullName );
											vsMsgBodyLines.push_back( szMsgInfo );
										}
									}
								}									
							}
						}							
					}
				}
			}
			// rewrite bso and create message to the link, if bso has been changed
			if (fbsoChanged) {
				CFile	bsoFile;
				if (!bsoFile.Open( bsoManager.getName(), GENERIC_WRITE, CREATE_ALWAYS )) {
					logger << LOG_ERROR
						   << TEXT("failed to open BSO file for writing:\n")
                           << bsoManager.getName() << TEXT("\n");
				}
				else {
					for (i = 0; i < vsbsoLines.size(); i++) {
						if (vbbsoLeft[i]) {
							bsoFile.WriteString( vsbsoLines[i].c_str() );
							bsoFile.WriteString( "\r\n" );
						}
					}
					bsoFile.Close();
				}
				// create message
				if (vsMsgBodyLines.size() > 1) {
        			CMsg	msg;
					// set receiver (we well reply)
        			msg.setFromAddress( link->getAddress() );
					// set sender
					msg.setToAddress( link->getOurAddress() );
					msg.setSenderName( "SysOp" );
					msg.setReceiverName( "Adrenalin" );
					msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0, m_language[LNG_TIMEPURGE_SUBJECT], vsMsgBodyLines );
				}
			}
			else {
				logger << LOG_LOUD << TEXT("nothing to purge\n");
			}
		}
		else {
			logger << LOG_LOUD << TEXT("nothing to purge\n");
		}

		// clear busy flag
		bsoManager.clearBusyFlag();
	}
}

const TCHAR*	szFilesDBFileName = TEXT("files.db");

void CAdrenalin::UpdateFilesDB() {
	CFile	dbFile;
	TCHAR	szFileName[MAX_PATH];

	// prepare full files db file name
	ConcatStrings( szFileName, m_szStartupPath, szFilesDBFileName );

	// open db in append mode
	if (!dbFile.Open( szFileName, GENERIC_WRITE, OPEN_ALWAYS )) {
		LoggerPtr	logger;
		logger << LOG_ERROR
			   << TEXT("Failed to open \"") << szFilesDBFileName << TEXT("\" for appending\n");
		return;
	}
	dbFile.SeekToEnd();

	// write current time
	SYSTEMTIME	stLocal;
	TCHAR		szDateTime[32];
	::GetLocalTime( &stLocal );
	wsprintf( szDateTime, _T("%04hu/%02hu/%02hu %02hu:%02hu:%02hu\n"),
		stLocal.wYear,
		stLocal.wMonth,
		stLocal.wDay,
		stLocal.wHour,
		stLocal.wMinute,
		stLocal.wSecond );

	dbFile.WriteString("Date ");
	dbFile.WriteString( szDateTime );

	// write new files names from the every area

	int	areasCount = CAreasManager::getInstance()->getAreasCount();
	for (int i = 0; i < areasCount; i++) {
		PArea	area = CAreasManager::getInstance()->getArea( i );
		int	filesCount = area->getNewFilesCount();
		if (filesCount == 0)
			continue;
		dbFile.WriteString("Area ");
		dbFile.WriteString( area->getName().c_str() );
		dbFile.WriteString("\n");
		for (int j = 0; j < filesCount; j++) {
			const tstring&	newFile = area->getNewFile( j );
			if (HasSpaces( newFile.c_str() )) {
				dbFile.WriteString("\"");
				dbFile.WriteString( newFile.c_str() );
				dbFile.WriteString("\"\n");
			}
			else {
				dbFile.WriteString( newFile.c_str() );
				dbFile.WriteString("\n");
			}
		}
	}
	
	// close db
	dbFile.Close();
}

void CAdrenalin::BasePurge() {
	CCfgParser	parser;
	TCHAR		szFileName[MAX_PATH];
	VTString	vsDBLines;		// we will keep here lines to overwrite db
	bool		fObsoleteFiles;		// flag is set when reading lines of obsolete
						// files
	__int64		i64TresholdTime;
	__int64		i64Date;
	SYSTEMTIME	stDate;
	FILETIME	ftDate;
	LARGE_INTEGER	li;
	TCHAR*		szNameFrom;

	LoggerPtr	logger;

	PArea		area = NULL;

	// prepare full files db file name
	ConcatStrings( szFileName, m_szStartupPath, szFilesDBFileName );

	// open files db file
	if (!parser.Open( szFileName )) {
		logger << LOG_ERROR
			   << TEXT("failed to open \"") << szFilesDBFileName << TEXT("\" for reading\n");
		return;
	}

	// prepare time
	{
		SYSTEMTIME	stCurrent;
		FILETIME	ftCurrent;
		GetLocalTime( &stCurrent );
		SystemTimeToFileTime( &stCurrent, &ftCurrent );
		__int64	i64Days = ((__int64)m_basePurge.m_iDays)*10000000*24*60*60;
		li.LowPart  = ftCurrent.dwLowDateTime;
		li.HighPart = ftCurrent.dwHighDateTime;
		i64TresholdTime = li.QuadPart - i64Days;
	}

	// clear areas
	// TODO: do we need it?
	{
		int	areasCount = CAreasManager::getInstance()->getAreasCount();
		for (int i = 0; i < areasCount; i++) {
			CAreasManager::getInstance()->getArea( i )->forgetNewFiles();
		}
	}

	// consequently read lines from files db file
	tstring	sDate;
	bool	fDateLineStored = false;
	fObsoleteFiles = false;	// we don't know yet what data we are reading
	while (!parser.Eof()) {
		// read line
		parser.ReadLine();
		// check errors
		if (parser.Failed()) {
			/*
			TODO: when file is done parser should not report failure
			logger << LOG_ERROR 
				   << TEXT("read error occured while reading \"")
				   << szFilesDBFileName << TEXT("\"\n");
		    */
			break;
		}
		// check for Date line
		if (parser.Count() == 3 && lstrcmpi( parser[0], _T("Date") ) == 0) {
			// keep date line
			sDate           = parser.String();
			fDateLineStored = false;
			// reset flag of obsolete files
			fObsoleteFiles  = false;
			memset( &stDate, 0, sizeof(stDate) );
			// parse date
			if (_stscanf( parser[1], _T("%hu/%hu/%hu"), 
				&stDate.wYear, &stDate.wMonth, &stDate.wDay ) == 3
			    &&
			    _stscanf( parser[2], _T("%hu:%hu:%hu"),
				&stDate.wHour, &stDate.wMinute, &stDate.wSecond ) == 3)
			{
				// convert date
				SystemTimeToFileTime( &stDate, &ftDate );
				li.LowPart  = ftDate.dwLowDateTime;
				li.HighPart = ftDate.dwHighDateTime;
				i64Date     = li.QuadPart;
				if (i64Date <= i64TresholdTime) {
					fObsoleteFiles = true;	// we are now reading obsolete files
					area           = NULL;
				}
			} else {
				logger << LOG_WARNING
					   << TEXT("failed to parse Date at line ") << parser.GetLine()
					   << TEXT(" in file \"") << szFilesDBFileName << TEXT("\"\n");
			}
			if (!fObsoleteFiles) {
				// date is young, keep line
				vsDBLines.push_back( parser.String() );
			}
		}
		else {
			if (fObsoleteFiles) {
				// skip empty lines
				if (parser.Count() == 0) {
					continue;
				}
				// keep file name
				if (parser.Count() == 1) {
					if (area != NULL) {
						lstrcpy( szFileName, parser[0] );
						if (szFileName[0] == _T('\"'))
							RemoveBoundaryQuotes( szFileName );
						if (szFileName[0] != _T('\0'))
							area->rememberNewFile( szFileName );
					}
					else {
						// keep date
						if (!fDateLineStored) {
							vsDBLines.push_back( sDate );
							fDateLineStored = true;
						}
						// keep line
						vsDBLines.push_back( parser.String() );
					}
				}
				else {
					// check area
					if (parser.Count() == 2 && lstrcmpi( parser[0], _T("Area") ) == 0) {
						area = CAreasManager::getInstance()->getArea( parser[1] );
						if (area != NULL) {
							// check veto
							if (area->hasFlag( CArea::FLAG_VETO_MANUAL_PURGE )) {
								area = NULL;
							}
							else if (!m_basePurge.m_vsAreas.empty()) {
								// check mask
								VTString::iterator	itAreaMask = m_basePurge.m_vsAreas.begin();
								for ( ; itAreaMask != m_basePurge.m_vsAreas.end(); itAreaMask++) {
									if (CheckMask( area->getName().c_str(), (*itAreaMask).c_str() ))
										break;
								}
								if (itAreaMask == m_basePurge.m_vsAreas.end()) {
									// satisfied mask was not found
									area = NULL;
								}
							}
						}
						if (area == NULL) {
							// keep date
							if (!fDateLineStored) {
								vsDBLines.push_back( sDate );
								fDateLineStored = true;
							}
							// keep line
							vsDBLines.push_back( parser.String() );
						}
					}
					else {
						logger << LOG_WARNING
							   << TEXT("incorrect number of lexems at line ") << parser.GetLine()
							   << TEXT(" in file \"") << szFilesDBFileName << TEXT("\"\n");							   
					}
				}
			}
			else {
				// keep line
				vsDBLines.push_back( parser.String() );
			}
		}
	}
	parser.Close();

	// remove obsolete files
	VTString	vsReport;
	VTString	vsRemovedFiles;

	int	areasCount = CAreasManager::getInstance()->getAreasCount();
	for (int i = 0; i < areasCount; i++) {
		// get area
		area = CAreasManager::getInstance()->getArea( i );
		// check number of files
		int	filesCount = area->getNewFilesCount();
		if (filesCount == 0)
			continue;
		vsReport.push_back( tstring(TEXT("Area: ")) + area->getName() );
		logger << LOG_LOUD << TEXT("Area: ") << area->getName() << TEXT("\n");
		// prepare file path
		lstrcpy( szFileName, area->getWorkingPath().c_str() );
		szNameFrom = szFileName + lstrlen( szFileName );
		vsRemovedFiles.clear();
		for (int j = 0; j < filesCount; j++) {
			const tstring&	newFile = area->getNewFile( j );
			// prepare full file name
			lstrcpy( szNameFrom, newFile.c_str() );
			// remove file
			if (DeleteFile( szFileName )) {
				vsReport.push_back( tstring(TEXT("         ")) + newFile );
				logger << LOG_LOUD << TEXT("  ") << szNameFrom << TEXT("\n");
				vsRemovedFiles.push_back( newFile );
			}
			else {
				vsReport.push_back( tstring(TEXT("  failed ")) + newFile );
				logger << LOG_ERROR << TEXT("failed to remove \"") << szFileName << TEXT("\"\n");
			}
		}
		// clear descriptions
		if (!vsRemovedFiles.empty())
			area->clearDescFile( vsRemovedFiles );
	}

	if (!vsReport.empty() && !(CCommonData::getInstance()->isNoMailNotifyForBasePurge())) {
		CMsg	msg;
		// set receiver (we will reply)
		msg.setFromAddress( CCommonData::getInstance()->getSysopAddress() );
		msg.setSenderName( CCommonData::getInstance()->getSysopName() );
		// set sender (we will reply)
		msg.setToAddress( CCommonData::getInstance()->getSysopAddress() );
		msg.setReceiverName( "Adrenalin" );
		msg.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0, _T("List of files removed while purge"), vsReport );
		// create flag crying about new message (if have to)
		createFileFlagsOnEvent( EE_FIX );
	}

	// overwrite files DB
	CFile	dbFile;
	// prepare full files db file name
	ConcatStrings( szFileName, m_szStartupPath, szFilesDBFileName );
	if (!dbFile.Open( szFileName, GENERIC_WRITE, CREATE_ALWAYS )) {
		logger << LOG_ERROR
			   << TEXT("Failed to open \"") << szFilesDBFileName << TEXT("\" for writing\n");
		return;
	}
	VTString::iterator	itLine = vsDBLines.begin();
	for ( ; itLine != vsDBLines.end(); itLine++) {
		if ((*itLine).length() > 0) {
			dbFile.WriteString( (*itLine).c_str() );
		}
		dbFile.WriteString("\n");
	}
	dbFile.Close();
}

bool CAdrenalin::CorrectPath(LPTSTR szPath) {
	TCHAR	szTempPath[MAX_PATH];

	// 1. Check quotes
	if (szPath[0] == _T('"'))
            RemoveBoundaryQuotes( szPath );
	// 2. Check empty path
        if (szPath[0] == _T('\0'))
		return (false);

	// keep path in a temporary buffer
	lstrcpy( szTempPath, szPath );

	if (szPath[0] == _T('\\')) {
		// append only startup drive
		ConcatStrings( szPath, m_szStartupDrive, szTempPath );
	} else if (szPath[1] != _T(':')) {
		// append only startup path
		ConcatStrings( szPath, m_szStartupPath, szTempPath );
	}

	return (true);
}

void CAdrenalin::adjustFilePath(tstring& path) {
	CAdrenalin::adjustFileName( path );

	if (!path.empty()) {
		if (path[path.length()-1] != PATH_SEPARATOR) {
			path += PATH_SEPARATOR_STR;
		}
	}
}


void CAdrenalin::adjustFileName(tstring& name) {
	if (!name.empty()) {
		if (name[0] == PATH_SEPARATOR) {
			name = tstring(CAdrenalin::getStartupDrive()) + name;
		} else {
			if (name.length() < 2 || name[1] != TEXT(':')) {
				name = tstring(CAdrenalin::getStartupPath()) + name;
			}
		}
	} else {
		name = CAdrenalin::getStartupPath();
	}
}


/**
 * Remove double quotes, check empty names and names with path separator
 */
bool CAdrenalin::CorrectName(LPTSTR szName) {
	// 1. Check quotes
	if (szName[0] == TEXT('"'))
            RemoveBoundaryQuotes( szName );
	// 2. Check empty name
        if (szName[0] == TEXT('\0'))
		return false;

	// check path separators
	LPCTSTR	charPointer = szName;
	while (*charPointer != TEXT('\0')) {
		if (*charPointer == TEXT('\\') || *charPointer == TEXT('/')) {
			// name is not a single name :(
			return false;
		}
		charPointer++;
	}
	// name is correct;
	return true;
}


void CAdrenalin::ProcessResendCommand(PLink link, SMsgData& msgData, vector<tstring>& bodyLines)
{
	
	// Proposed steps in this function are:
	// 
	// 1) prepare list of files to resend
	// 2) prepare links to be used as SEEN-BY
	// 3) create tics and fill outbound

	// parse message line tail; it should be in format <area name> [<file mask> ...]
	tstring			areaName;
	vector<tstring>	fileMasks;

	LoggerPtr	logger;

	if (!ParseResendCommandTail( msgData.m_sTail, areaName, fileMasks )) {
		// failed to parse command tail
		TCHAR	notifyLine[LINE_SIZE];
		_stprintf( notifyLine, m_language[LNG_RESEND_TAIL_PARSE_ERROR], msgData.m_sTail.c_str() );
		bodyLines.push_back( notifyLine );
		logger << LOG_ERROR << TEXT("failed to parse commaind tail\n");
		return;
	}

	// select area
	PArea	area = CAreasManager::getInstance()->getArea( areaName );
	if (area == NULL)
	{
		// failed to select area
		TCHAR	notifyLine[LINE_SIZE];
		_stprintf( notifyLine, m_language[LNG_AREA_NOT_FOUND], areaName.c_str() );
		bodyLines.push_back( notifyLine );
		logger << LOG_ERROR << TEXT("area not found: ") << areaName << TEXT("\n");
		return;
	}

	if (!area->hasFlag( CArea::FLAG_OPEN_FOR_RESEND )) {
		if ((area->getTie( link ) == NULL) || !area->canReceive( link ))
		{
			// user have no access to the specified area
			TCHAR	notifyLine[LINE_SIZE];
			_stprintf( notifyLine, m_language[LNG_AREA_NO_ACCESS], areaName.c_str() );
			bodyLines.push_back( notifyLine );
			logger << LOG_ERROR << TEXT("link ") << link->getAddress().toString()
				   << TEXT(" cannot receive files from area ") << areaName << TEXT("\n");
			return;
		}
	}

	if (fileMasks.size() == 0) {
		// file masks not specified, let's define one
		fileMasks.push_back( TEXT("*") );
	}

	// get list of files to resend
	vector<tstring>	files;
	vector<tstring> descriptions;
	ReadFilesList( area->getWorkingPath().c_str(), fileMasks, files, descriptions );
	if (files.empty())
	{
		// there is nothing to resend
		// inform user
		TCHAR	notifyLine[LINE_SIZE];
		_stprintf( notifyLine, m_language[LNG_RESEND_NO_FILES], areaName.c_str() );
		bodyLines.push_back( notifyLine );
		return;
	}

	// prepare notification title
	bodyLines.push_back( m_language[LNG_RESEND_FILELIST] );

	// prepare sender address
    CAddress	senderAddress = link->getOurAddress();

	// prepare ticProcessor

	CTic	ticProcessor;
    // set area name
    lstrcpy( ticProcessor.m_szArea, area->getName().c_str() );

	// set Replaces
    ticProcessor.m_szReplaces[0] = TEXT('\0');

    // add all us to seenby list, if have to
	if (CCommonData::getInstance()->areAllAKASeen()) {
		ticProcessor.getSeenByList().push_back( CCommonData::getInstance()->getSysopAddress() );
        for (unsigned int i = 0; i < CCommonData::getInstance()->getSysopAKACount(); i++) {
			ticProcessor.getSeenByList().push_back( CCommonData::getInstance()->getSysopAKA(i) );
        }
    }

	// prepare seen-by list
	{
		unsigned int	count = area->getTiesCount();
		for (unsigned int i = 0; i < count; i++) {
			ConstTiePtr	tie = area->getTie( i );
			// add all linked links; we cannot determine who was in original seen-by list
			if (link->getAddress() != tie->getLink()->getAddress()) {
				ticProcessor.getSeenByList().push_back( tie->getLink()->getAddress() );
			}
		}
	}

	// TODO: don't know for what :) the code below
	vector<PLink>	forwards;
	if (area->getTie( link ) != NULL) {
		forwards.push_back( link );
	}

	ticProcessor.setOriginAddress( senderAddress );

	for (unsigned int i = 0; i < files.size(); i++) {
		LPCTSTR	fullFileName = files[i].c_str();

		// adjust ticProcessor

		// calc crc & size
		ticProcessor.m_iSize = CalcCRC32( fullFileName, ticProcessor.m_uiCRC32 );
		if (ticProcessor.m_iSize == -1) {
			// TODO: process error
		}
		else {

			// check and set file name
			TCHAR	shortFileName[MAX_PATH];
			ConvertNameToShort( fullFileName, shortFileName );
			int	index, shortSlashPos, fullSlashPos;
			for (index = 0, shortSlashPos = -1; shortFileName[index] != TEXT('\0'); index++) {
				if (shortFileName[index] == TEXT('\\'))
					shortSlashPos = index;
			}
			shortSlashPos++;
			lstrcpy( ticProcessor.m_szFile, shortFileName + shortSlashPos );
			for (index = 0, fullSlashPos = -1; fullFileName[index] != TEXT('\0'); index++) {
				if (fullFileName[index] == TEXT('\\'))
					fullSlashPos = index;
			}
			fullSlashPos++;
			if (lstrcmpi( fullFileName + fullSlashPos, shortFileName + shortSlashPos ) != 0)
				lstrcpy( ticProcessor.m_szFullName, fullFileName + fullSlashPos );
			else
				ticProcessor.m_szFullName[0] = TEXT('\0');

			// set description
			lstrcpy( ticProcessor.m_szDesc, descriptions[i].c_str() );

			ticProcessor.Forward( link, area, area->getWorkingPath().c_str(), forwards );

			// inform user what file we send
			bodyLines.push_back( tstring(TEXT("    ")) + (fullFileName + fullSlashPos) );
		}
	}
}


bool CAdrenalin::ParseResendCommandTail(const tstring& tail, tstring& areaName, vector<tstring>& fileMasks)
{
	unsigned int	i, j;

	// skip leading spaces
	i = 0;
	while ((i < tail.size()) && (tail.at(i) == TEXT(' '))) {
		i++;
	}
	// check for empty tail
	if (i == tail.size()) {
		// there is no data
		return false;
	}

	// get area name
	j = i+1;
	while ((j < tail.size()) && (tail.at(j) != TEXT(' '))) {
		j++;
	}
	areaName = tail.substr( i, j-i );

	// now let's get file masks if they are defined
	fileMasks.clear();

	i = j;
	while (i < tail.size()) {
		// skip spaces
		while ((i < tail.size()) && (tail.at(i) == TEXT(' '))) {
			i++;
		}
		if (i < tail.size()) {
			// suppose we are near next file mask
			TCHAR	stopChar;
			if (tail.at(i) == TEXT('"')) {
				stopChar = TEXT('"');
				i++;
			}
			else {
				stopChar = TEXT(' ');
			}
			// we don't know whether file mask is empty, so let's start from i, not from i+1
			j = i;
			while ((j < tail.size()) && (tail.at(j) != stopChar)) {
				j++;
			}
			if (j == i) {
				// file mask is empty or invalid
				return false;
			}
			tstring	fileMask = tail.substr( i, j-i );
			// skip closing doublequote if it is present
			if ((j < tail.size()) && (tail.at(j) == TEXT('"'))) {
				j++;
			}
			if ((j < tail.size()) && (tail.at(j) != TEXT(' '))) {
				// file mask is incorrect, e.g. tail was something like ' "somearc.rar"somefile1.txt'
				return false;
			}
			i = j;
			fileMasks.push_back( fileMask );
		}
	}
	return true;
}


void CAdrenalin::ReadFilesList(LPCTSTR areaPath, const vector<tstring>& fileMasks, vector<tstring>& files, vector<tstring>& descriptions)
{
    WIN32_FIND_DATA findFileData;
	HANDLE          findFileHandle;
	BOOL            fileFound;

	// prepare file mask to find all files in area directory
	TCHAR           findFileMask[MAX_PATH];
	lstrcpy( findFileMask, areaPath );
	lstrcat( findFileMask, TEXT("*.*") );

	// do search
	findFileHandle = FindFirstFile( findFileMask, &findFileData );
	fileFound      = (findFileHandle != INVALID_HANDLE_VALUE);
	while (fileFound) {
		// skip hidden, temporary files and directories
		if ((findFileData.dwFileAttributes & (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_TEMPORARY)) == 0) {
			// skip filelist
			if (_tcscmp( findFileData.cFileName, CCommonData::getInstance()->getFileListName().c_str() ) != 0) {
				// check file for masks
				for (unsigned int i = 0; i < fileMasks.size(); i++) {
					if (CheckMask( findFileData.cFileName, fileMasks[i].c_str() )) {
						files.push_back( findFileData.cFileName );
						// store empty description
						descriptions.push_back( tstring() );
						break;
					}
				}
			}
		}
		fileFound = FindNextFile( findFileHandle, &findFileData );
	}
	if (findFileHandle != INVALID_HANDLE_VALUE)
		FindClose( findFileHandle );

	// read descriptions

	TCHAR	descFileName[MAX_PATH];
	lstrcpy( descFileName, areaPath );
	lstrcat( descFileName, CCommonData::getInstance()->getFileListName().c_str() );

	COemLineReader	lineReader = COemLineReader( descFileName );
	while (lineReader.next()) {

		// get file name from the head of the line

		LPCSTR	line = lineReader.getLine();
		char	stopChar = (line[0] == '"') ? '"' : ' ';		
		int		stopPos  = (stopChar == '"') ? 1 : 0;
		int		len = 0;
		while ((line[stopPos] != '\0') && (line[stopPos] != stopChar) && (len < MAX_PATH)) {
			stopPos++;
			len++;
		}
		if ((line[stopPos] == stopChar) && (len > 0)) {
			// remember file name
			char	fileName[MAX_PATH];
			::CopyMemory( fileName, (stopChar == '"' ? line+1 : line), len );
			fileName[len] = '\0';

			// check file name
			for (unsigned int i = 0; i < files.size(); i++) {
				if (descriptions[i].length() == 0) {
					// there is no description yet
					if (lstrcmpi( files[i].c_str(), fileName) == 0) {
						// file name found, let's make description
						if (line[stopPos] == '"') {
							stopPos++;
						}
						while (line[stopPos] == ' ') {
							stopPos++;
						}
						int	startPos = stopPos;
						while (line[stopPos] != '\0') {
							stopPos++;
						}
						if (stopPos > startPos) {
							descriptions[i] = tstring( line + startPos, stopPos - startPos );
						}
					}
				}
			}
		}
	}

	// append file path to every file found
	for (unsigned int i = 0; i < files.size(); i++) {
		files[i] = tstring(areaPath) + files[i];
	}
}


/**
 * @return	true if tic was older than max tic age (in hours)
 */
bool CAdrenalin::isTicOld(LPCTSTR ticFileName, int maxTicAge)
{
	LoggerPtr	logger;
	logger << LOG_ERROR;

	HANDLE	ticHandle = CreateFile( ticFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
	if (ticHandle == INVALID_HANDLE_VALUE) {
		logger << TEXT("failed to open TIC file for reading:\n")
			   << ticFileName << TEXT("\n");
		return false;
	}
	// check time
	FILETIME	ticTime;
	// get creation time
	if (!GetFileTime( ticHandle, &ticTime, NULL, NULL )) {
		logger  << TEXT("failed to determine TIC file creation time:\n")
			    << ticFileName << TEXT("\n");
		CloseHandle( ticHandle );
		return false;
	}
	CloseHandle( ticHandle );

	// get system time
	FILETIME	sysTime;
	GetSystemTimeAsFileTime( &sysTime );
	ULARGE_INTEGER	convert;
	convert.LowPart  = sysTime.dwLowDateTime;
	convert.HighPart = sysTime.dwHighDateTime;
	__int64	nanotime = convert.QuadPart;

	convert.LowPart  = ticTime.dwLowDateTime;
	convert.HighPart = ticTime.dwHighDateTime;
	nanotime -= convert.QuadPart;

	__int64	limit = ((__int64)maxTicAge)*10000000*60*60;
	return (nanotime >= limit);
}


int	CAdrenalin::hatchSingleFile(const tstring& areaName, const tstring& fileName, const tstring& fileDesc,
								bool noCopy, const tstring& replaces)
{
	LoggerPtr	logger;

    // check area
	PArea	area = CAreasManager::getInstance()->getArea( areaName );
    if (area == NULL) {
		logger << LOG_ERROR << TEXT("area ") << areaName << TEXT(" doesn't exist\n");
        return (ERROR_AREA_NOT_FOUND);
    }

    // check file
    if (_taccess( fileName.c_str(), 0) != 0) {
		logger << LOG_ERROR << TEXT("file ") << fileName << TEXT(" not found\n");
        return (ERROR_FILE_ABSENT);
    }

    TCHAR       szDestFileName[MAX_PATH];
    TCHAR		szDestFilePath[MAX_PATH];
    TCHAR       szShortFileName[MAX_PATH];
    TCHAR       szFullFileName[MAX_PATH];
    TCHAR       szLine[LINE_SIZE];
    TCHAR       szTimeLine[LINE_SIZE];
    int         i , j, nNameFrom;

    CTic        tic;

    ofstream    out;

    ifstream    in;

    CCfgParser *trafficParser;

    int         nTraffic;
    bool        fDistributedForLinks;

    logger << LOG_LOUD << TEXT("Hatch:\n");
    
    // make all fecho directories and set areas traffic to zero
	MakeDirStructure( area->getWorkingPath().c_str() );
	// TODO: should we reset area traffic?
	area->setTraffic( 0 );

    // make tics directories
	MakeDirStructure( CCommonData::getInstance()->getBadTicsPath().c_str() );
    MakeDirStructure( CCommonData::getInstance()->getNewTicsPath().c_str() );

    // calc crc & size
    tic.m_iSize = CalcCRC32( fileName.c_str(), tic.m_uiCRC32 );
    if (tic.m_iSize == -1) {
		logger << LOG_ERROR << TEXT("failed to calculate CRC32 of ") << fileName << TEXT("\n");
        return (ERROR_CRC32_FAILED);
    }

    // move file into fecho directory if have to
    // first let's find name without path
    for (i = 0, nNameFrom = -1; fileName[i] != ZERO_CHAR; i++) {
        if (fileName[i] == TEXT('\\'))
            nNameFrom = i;
    }
    nNameFrom++;

    if (noCopy) {
        lstrcpy( szDestFileName, fileName.c_str() );
		CorrectPath( szDestFileName );
        for (i = 0, j = -1; szDestFileName[i] != TEXT('\0'); i++) {
			if (szDestFileName[i] == PATH_SEPARATOR)
				j = i;
        }
        j++;
        lstrcpy( szDestFilePath, szDestFileName );
        szDestFilePath[j] = ZERO_CHAR;
    }
    else {
		lstrcpy( szDestFilePath, area->getWorkingPath().c_str() );
		lstrcpy( szDestFileName, szDestFilePath );
        lstrcat( szDestFileName, fileName.c_str() + nNameFrom );

		// avoid overriting of a file with long name by file with the same short name
		LongNameKeeper	longNameKeeper;
		if (longNameKeeper.keep( szDestFileName ) == LongNameKeeper::ONLY_ONE_NAME) {
			// file has only one name, not two names (short and long)
			RenameFilePlan	renamePlan = RFP_NONE;
			if (area->doesRenameFilePlanSet()) {
				renamePlan = area->getRenameFilePlan();
			} else {
				renamePlan = CCommonData::getInstance()->getRenameFilePlan();
			}
			if (renamePlan != RFP_NONE) {
				area->renameFile( szDestFileName, renamePlan );
			}
		}

		FileOpResult	foResult = MyCopyFile( fileName.c_str(), szDestFileName );
		if (foResult != FO_SUCCEEDED) {
			logger << LOG_ERROR
				   << TEXT("failed to copy file ") << fileName << TEXT(" into ")
				   << szDestFileName << TEXT("\n");
			switch (foResult) {
				case FO_FAILED:
					logger << TEXT("  unknown failure reason\n");
					break;
				case FO_SOURCE_UNAVAILABLE:
					logger << TEXT("  source unavailable\n");
					break;
				case FO_DEST_UNAVAILABLE:
					logger << TEXT("  destination unavailable\n");
					break;
				case FO_DEST_INVALID:
					logger << TEXT("  destination file has invalid name\n");
					break;
				case FO_READ_FAILED:
					logger << TEXT("  failed to read source file\n");
					break;
				case FO_WRITE_FAILED:
					logger << TEXT("  failed to write destination file\n");
					break;
			}
			return (ERROR_MOVE_FAILED);
        }
    }

    logger << LOG_LOUD << TEXT("  file: ") << fileName << TEXT("\n  area: ")
           << area->getName() << TEXT("\n");
	if (!fileDesc.empty()) {
        logger << TEXT("  desc: \"") << fileDesc << TEXT("\"\n");
	} else {
        logger << TEXT("  no description\n");
	}
	if (!replaces.empty()) {
        logger << TEXT("  repl: ") << replaces << TEXT("\n");
	} else {
        logger << TEXT("  no Replaces\n");
	}

    if (!noCopy) {
        // update descriptions
        logger << TEXT("  updating description file...\n");

		filesize_t	fileSize;
		fileSize.LowPart  = tic.m_iSize;
		fileSize.HighPart = 0;
        area->updateDescFile( fileName.c_str() + nNameFrom, fileDesc.c_str(), fileSize, VTString() );

        logger << TEXT("  done\n");
    }

    // prepare list of links

	vector<PLink>	forwards;
	
	int	count = area->getTiesCount();
	for (i = 0; i < count; i++) {
		TiePtr	tie  = area->getTie(i);
		PLink	link = tie->getLink();
        // skip passive or unavailable seenby
		if (link->getActivity() != ACTIVITY_USUAL)
            continue;
        // skip links that cannot receive
		if (!area->canReceive( link )) {
			// area disallow to receive files
			continue;
		}
        // add link to list
		forwards.push_back( link );
    }

    CAddress	addr;

    // if only one link is connected to area, set hatch address to its AKA
	if (forwards.size() == 1) {
		addr = forwards.back()->getOurAddress();
	} else {
		addr = CCommonData::getInstance()->getSysopAddress();
	}

	{
		logger << TEXT("  updating newfiles.txt file...\n");
		ConcatStrings( szShortFileName, m_szStartupPath, _T("newfiles.txt") );
		CFile	newFilesFile;
		// open newfiles list
		newFilesFile.Open( szShortFileName, GENERIC_WRITE );
		newFilesFile.SeekToEnd();
		// update newfiles list
		TCHAR	szBuf[1024];
		_stprintf( szBuf, 
				   _T("Area %s\r\nFile %s\r\nSize %ld\r\nDesc %s"),
				   area->getName().c_str(),
				   fileName.c_str() + nNameFrom,
				   tic.m_iSize,
				   fileDesc.c_str() );
		tstring	addrStr = addr.toStringWithoutDomain();
		lstrcat( szBuf, _T("\r\nOrigin ") );
		lstrcat( szBuf, addrStr.c_str() );
		lstrcat( szBuf, _T("\r\nUplink ") );
		lstrcat( szBuf, addrStr.c_str() );
		lstrcat( szBuf, _T("\r\n") );
		newFilesFile.WriteString( szBuf );
		newFilesFile.Close();
		logger << TEXT("  done\n");
	}
    
    // update traffic base
    
    // read areas traffic
    
    ConcatStrings( szFullFileName, m_szStartupPath, TEXT("traffic.dat") );

    *szTimeLine = TEXT('\0');
    trafficParser = new CCfgParser( szFullFileName );
    if (!trafficParser->Failed()) {
        // read date line
        trafficParser->ReadLine();
        if (!trafficParser->Failed())
            lstrcpy( szTimeLine, trafficParser->String() );
        // read data lines
        while (!trafficParser->Eof()) {
            trafficParser->ReadLine();
            if (trafficParser->Failed())
                break;
            if (trafficParser->Count() != 2)
                break;
			PArea	trafficArea = CAreasManager::getInstance()->getArea( (*trafficParser)[0] );
            if (trafficArea == NULL)
                continue;
			if (!StrToInt( (*trafficParser)[1], nTraffic ))
				break;
            trafficArea->setTraffic( nTraffic );
        }
    }
    delete trafficParser;

    area->setTraffic( area->getTraffic() + tic.m_iSize );
    
    {
	    CFile	trafficFile;
	    if (!trafficFile.Open( szFullFileName, GENERIC_WRITE, CREATE_ALWAYS )) {
			logger << LOG_ERROR
			       << TEXT("failed to write traffic base file ") << szFullFileName << TEXT("\n");
	    }
	    else {
			if (IsEmptyStr( szTimeLine )) {
				time_t curTime = time(NULL);
				trafficFile.WriteString( ctime( &curTime ) );
			}
			else {
				trafficFile.WriteString( szTimeLine );
				trafficFile.WriteString( "\r\n" );
			}

			int	areasCount = CAreasManager::getInstance()->getAreasCount();
			for (int i = 0; i < areasCount; i++) {
				PArea	trafficArea = CAreasManager::getInstance()->getArea( i );
				if (trafficArea->getTraffic() > 0) {
					_stprintf( szLine, _T("%s %lu\r\n"), 
						   trafficArea->getName().c_str(), trafficArea->getTraffic() );
					trafficFile.WriteString( szLine );
				}
			}
			trafficFile.Close();
	    }
    }

    logger << LOG_LOUD << TEXT("  creating TICs for connected links...\n");
    // set area name
    lstrcpy( tic.m_szArea, area->getName().c_str() );
    // check and set file name
    ConvertNameToShort( szDestFileName, szShortFileName );
    for (i = 0, j = -1; szShortFileName[i] != _T('\0'); i++) {
        if (szShortFileName[i] == _T('\\'))
            j = i;
    }
    j++;
    lstrcpy( tic.m_szFile, szShortFileName + j );
    if (lstrcmpi( fileName.c_str() + nNameFrom, szShortFileName + j ) != 0)
        lstrcpy( tic.m_szFullName, fileName.c_str() + nNameFrom );
    else
	tic.m_szFullName[0] = _T('\0');
    // set desc
    if (!fileDesc.empty())
        lstrcpy( tic.m_szDesc, fileDesc.c_str() );
    // set Replaces
	if (!replaces.empty()) {
        lstrcpy( tic.m_szReplaces, replaces.c_str() );
	}
    // set Origin to main address, if have to
	if (CCommonData::getInstance()->isHatchOriginAddressMain() || CCommonData::getInstance()->getSysopAKACount() == 0) {
        tic.setOriginAddress( CCommonData::getInstance()->getSysopAddress() );
	}
    // add all us to seenby list, if have to
	if (CCommonData::getInstance()->areAllAKASeen()) {
	    tic.getSeenByList().push_back( CCommonData::getInstance()->getSysopAddress() );
        for (unsigned int index = 0; index < CCommonData::getInstance()->getSysopAKACount(); index++) {
			tic.getSeenByList().push_back( CCommonData::getInstance()->getSysopAKA(index) );
        }
    }
    // do tics for all connected links
    // now do tics
    fDistributedForLinks = (!forwards.empty());
    vector<PLink>::iterator	itLinked = forwards.begin();
    while (itLinked != forwards.end()) {
        // set Origin to AKA, if have to
		if (CCommonData::getInstance()->getSysopAKACount() > 0 && !CCommonData::getInstance()->isHatchOriginAddressMain()) {
			tic.setOriginAddress( (*itLinked)->getOurAddress() );
        }
		tic.ensureValidity();
       	tic.Forward( (*itLinked), area, szDestFilePath, forwards );
        ++itLinked;
    }

    // create flag
	if (fDistributedForLinks) {
		createFileFlagsOnEvent( EE_HATCH );
	}

    logger << TEXT("  done\n")
		   << TEXT("File ") << fileName << TEXT(" was successfully hatched into area ")
		   << area->getName() << TEXT("\n");
    return (ERROR_NONE);
}


void CAdrenalin::checkTempBSO() {
	// collect temp entries
	CBSOManager	bsoManager;
	VTString	entries;
	bsoManager.collectTempEntries( entries );

	if (entries.size() > 0) {

		// prepare buffer for file copying
		const	BUFFER_SIZE = 32*1024;
		LPBYTE	buffer = new BYTE[BUFFER_SIZE];

		// prepare vector of BSO route types
		const int	routeTypeCount             = 5;
		FlavourType	routeTypes[routeTypeCount] = 
			{ FT_NORMAL, FT_CRASH, FT_DIRECT, FT_HOLD, FT_IMMEDIATE };

		tstring	tempEntryName;
		tstring	fullEntryName;

		// go through all links
		for (CLinksManager::CIterator itLink = CLinksManager::getInstance()->begin();
			 itLink != CLinksManager::getInstance()->end(); ++itLink)
		{
			PLink	link = (*itLink);
			for (int i = 0; i < routeTypeCount; i++) {
				bsoManager.getTempEntryName( link->getAddress(), routeTypes[i],
											 tempEntryName );
				// go through all entries
				for (VTString::iterator	it = entries.begin(); it != entries.end(); it++) {
					if (lstrcmpi( (*it).c_str(), tempEntryName.c_str() ) == 0) {

						// corresponding entry found

						// setup BSO manager for the link
						bsoManager.setAddress( link->getAddress(), routeTypes[i] );

						if (bsoManager.setBusyFlag()) {
							// prepare full entry name
							bsoManager.getTempEntryPath( fullEntryName );
							fullEntryName += tempEntryName;

							// try to append temp file to the ordinary outbound file

							HANDLE	srcFile = CreateFile( fullEntryName.c_str(),
														  GENERIC_READ, FILE_SHARE_READ,
														  NULL, OPEN_EXISTING,
														  FILE_ATTRIBUTE_NORMAL, NULL );

							if (srcFile != INVALID_HANDLE_VALUE) {

								bool	canBeDeleted = false;

								HANDLE	destFile = CreateFile( bsoManager.getName(), 
												 			   GENERIC_WRITE, 0, NULL, OPEN_ALWAYS,
															   FILE_ATTRIBUTE_NORMAL, NULL);
								if (destFile != INVALID_HANDLE_VALUE) {
									SetFilePointer( destFile, 0, NULL, FILE_END );

									BOOL	goodRead = TRUE, goodWrite = TRUE;
									DWORD	bytesRead, bytesWritten;
									for (;;) {
										goodRead = ReadFile( srcFile, buffer, BUFFER_SIZE, &bytesRead, NULL );
										if (!goodRead || bytesRead <= 0) {
											break;
										}
										goodWrite = WriteFile( destFile, buffer, bytesRead, &bytesWritten, NULL );
										if (!goodWrite || bytesWritten != bytesRead) {
											break;
										}
									}

									CloseHandle( destFile );

									canBeDeleted = (goodRead && goodWrite);
								}

								CloseHandle( srcFile );

								if (canBeDeleted) {
									if (!DeleteFile( fullEntryName.c_str() )) {
										LoggerPtr	logger;
										logger << LOG_ERROR
											   << TEXT("failed to delete file ")
											   << fullEntryName << TEXT("\n");
									}
								}
							}

							bsoManager.clearBusyFlag();
						}

						break;
					}
				}
			}
		}

		// free buffer
		delete [] buffer;
	}
}


tstring CAdrenalin::getName() {
	return TEXT("Adrenalin v0.72");
}


void CAdrenalin::createFileFlagsOnEvent(ExecuteEvent event) {
	LoggerPtr		logger;
	CommonDataPtr	cd = CCommonData::getInstance();
	size_t			flagCount = cd->getEventFileFlagCount();
	for (size_t i = 0; i < flagCount; i++) {
		const CEventFileFlag&	flag = cd->getEventFileFlag( i );
		if (flag.getEvent() == event) {
			ensurePathExists( flag.getFileName() );
			if (CreateFileFlag( flag.getFileName().c_str() )) {
				logger << LOG_SILENT << TEXT("file flag created: ") << flag.getFileName() << TEXT("\n");
			} else {
				logger << LOG_WARNING << TEXT("failed to create file flag: ") << flag.getFileName() << TEXT("\n");
			}
		}
	}
}


void CAdrenalin::createFileFlagsOnFile(LPCTSTR fileName) {
	LoggerPtr		logger;
	CommonDataPtr	cd = CCommonData::getInstance();
	size_t			flagCount = cd->getMaskFileFlagCount();
	for (size_t i = 0; i < flagCount; i++) {
		CMaskFileFlag&	flag = cd->getMaskFileFlag( i );
		if (!flag.isUsed() && CheckMask( fileName, flag.getMask().c_str() )) {
			ensurePathExists( flag.getFileName() );
			if (CreateFileFlag( flag.getFileName().c_str() )) {
				logger << LOG_SILENT << TEXT("file flag created: ") << flag.getFileName() << TEXT("\n");
				flag.setUsed( true );
			} else {
				logger << LOG_WARNING << TEXT("failed to create file flag: ") << flag.getFileName() << TEXT("\n");
			}
		}
	}
}


void CAdrenalin::ensurePathExists(const tstring& fileName) {
	size_t	len = fileName.length();
	int		lastSeparatorPos = -1;
	for (size_t i = 0; i < fileName.length(); i++) {
		if (fileName[i] == PATH_SEPARATOR) {
			lastSeparatorPos = i;
		}
	}
	if (lastSeparatorPos > 0) {
		tstring	path = fileName.substr( 0, lastSeparatorPos );
		MakeDirStructure( path.c_str() );
	}
}


bool CAdrenalin::parseReleaseDate(LPCTSTR s, SYSTEMTIME& dateTime) {
	// allowed formats are:
	// DD.MM and DD.MM.YYYY

	TCHAR	tpl[] = TEXT("##.##.####");
	int		number    = 0;

	int	index = 0;
	while (tpl[index] != TEXT('\0') && s[index] != TEXT('\0')) {
		if (tpl[index] == TEXT('#')) {
				if (_istdigit( s[index] )) {
					number = number*10 + (s[index] - TEXT('0'));
				} else {
					return false;
				}
		} else {
			if (s[index] == tpl[index]) {
				if (index == 2) {
					if (number < 1 || number > 31) {
						return false;
					}
					dateTime.wDay = number;
				} else {
					if (number < 1 || number > 12) {
						return false;
					}
					dateTime.wMonth = number;
				}
				number = 0;
			} else {
				return false;
			}
		}
		++index;
	}
	if (s[index] != TEXT('\0')) {
		return false;
	}
	if (index == 5) {
		// short format
		if (number < 1 || number > 12) {
			return false;
		}
		dateTime.wMonth = number;
		SYSTEMTIME	currentTime;
		::GetLocalTime( &currentTime );
		dateTime.wYear = currentTime.wYear;
	} else if (index == 10) {
		// long format
		dateTime.wYear = number;
	} else {
		return false;
	}

	return true;
}


bool CAdrenalin::parseReleaseTime(LPCTSTR s, SYSTEMTIME& dateTime) {
	// allowed formats are:
	// HH:MM and HH:MM:SS

	TCHAR	tpl[] = TEXT("##:##:##");
	int		number    = 0;

	int	index = 0;
	while (tpl[index] != TEXT('\0') && s[index] != TEXT('\0')) {
		if (tpl[index] == TEXT('#')) {
				if (_istdigit( s[index] )) {
					number = number*10 + (s[index] - TEXT('0'));
				} else {
					return false;
				}
		} else {
			if (s[index] == tpl[index]) {
				if (index == 2) {
					if (number > 23) {
						return false;
					}
					dateTime.wHour = number;
				} else {
					if (number > 59) {
						return false;
					}
					dateTime.wMinute = number;
				}
				number = 0;
			} else {
				return false;
			}
		}
		++index;
	}
	if (s[index] != TEXT('\0')) {
		return false;
	}
	if (index == 5) {
		// short format
		if (number > 59) {
			return false;
		}
		dateTime.wMinute = number;
		dateTime.wSecond = 0;
	} else if (index == 8) {
		// long format
		if (number > 59) {
			return false;
		}
		dateTime.wSecond = number;
	} else {
		return false;
	}

	return true;
}

