#include "pch.h"
#include "AdrenalinSAX.h"
#include "Adrenalin.h"

#include "MsgAttrs.h"
#include "Misc.hpp"


CAdrenalinContentHandler::CAdrenalinContentHandler(CMiniSAXParser* parser)
	: m_parser(parser), m_state(ROOT_STATE), m_entriesHolder()
{
}


CAdrenalinContentHandler::~CAdrenalinContentHandler() {
}

LPCTSTR	YES = TEXT("yes");
LPCTSTR	NO  = TEXT("no");

LPCTSTR	ADRENALIN     = TEXT("adrenalin");
	LPCTSTR	SYSOP         = TEXT("sysop");
		LPCTSTR NAME          = TEXT("name");
		LPCTSTR ADDRESS       = TEXT("address");
		LPCTSTR AKA           = TEXT("aka");
	LPCTSTR PATHS         = TEXT("paths");
		LPCTSTR INBOUND       = TEXT("inbound");
		LPCTSTR BINKOUTBOUND  = TEXT("binkOutbound");
		LPCTSTR FILEBOXES     = TEXT("fileBoxes");
		LPCTSTR LONGFILEBOXES = TEXT("longFileBoxes");
		LPCTSTR NETMAIL       = TEXT("netmail");
		LPCTSTR	NEWTICS       = TEXT("newTics");
		LPCTSTR	BADTICS       = TEXT("badTics");
		LPCTSTR	PROCESSEDTICS = TEXT("processedTics");
		LPCTSTR	AUTOCREATEDAREAS = TEXT("autocreatedAreas");
	LPCTSTR	ALIAS         = TEXT("alias");
	LPCTSTR	LOG           = TEXT("log");
		LPCTSTR	COMMONFILE    = TEXT("commonFile");
		LPCTSTR	REQUESTSFILE  = TEXT("requestsFile");
		LPCTSTR	SHOWMUSTGOON  = TEXT("showMustGoOn");
		LPCTSTR	LOGMSGNAMES   = TEXT("logMsgNames");
	LPCTSTR	DESCRIPTIONS  = TEXT("descriptions");
		LPCTSTR	AUTODESCFILE   = TEXT("autoDescFile");
	LPCTSTR	FILES         = TEXT("files");
		LPCTSTR	RENAMEPLAN     = TEXT("renamePlan");
	LPCTSTR	FILELIST      = TEXT("fileList");
		// name constant is already defined
		LPCTSTR	FORMAT        = TEXT("format");
		LPCTSTR	SORT          = TEXT("sort");
	LPCTSTR	MAIL          = TEXT("mail");
		LPCTSTR	KILLPROCESSED = TEXT("killProcessed");
		LPCTSTR	REPLYATTRIBUTES = TEXT("replyAttributes");
			LPCTSTR	ATTR          = TEXT("attr");
		LPCTSTR	CREATIONLIMIT = TEXT("creationLimit");
			LPCTSTR	KBYTES        = TEXT("kBytes");
			LPCTSTR	LINES         = TEXT("lines");
		LPCTSTR	NONOTIFY         = TEXT("noNotify");
		LPCTSTR	NOTIFYONERRORS   = TEXT("notifyOnErrors");
		LPCTSTR	NOTIFYONWARNINGS = TEXT("notifyOnWarnings");
	LPCTSTR	LNGFILE       = TEXT("lngFile");
	LPCTSTR	FORWARDEXPIRATIONTIME = TEXT("forwardExpirationTime");
	LPCTSTR	MAXTICAGE             = TEXT("maxTicAge");
	LPCTSTR	WARNINGINSTEADOFERROR = TEXT("warningInsteadOfError");
	LPCTSTR	HATCHORIGINADDRESS    = TEXT("hatchOriginAddress");
	LPCTSTR	ALLAKASEEN            = TEXT("allAKASeen");
	LPCTSTR	HANDLEREPLACES        = TEXT("handleReplaces");
	LPCTSTR	FREEDRIVESPACELIMIT   = TEXT("freeDriveSpaceLimit");
	LPCTSTR	FILEFLAGS             = TEXT("fileFlags");
		LPCTSTR	ONEVENT               = TEXT("onEvent");
		LPCTSTR	ONFILE                = TEXT("onFile");


void CAdrenalinContentHandler::startElement(const char* name, CAttributeList& attributes)
	throw (CMiniSAXException)
{
	// the main idea is:
	//   we try to change state according to the element name;
	//   at the end of the method we check whether state changed or not;
	//	 if not then the element is unexpected

	// remember state
	HandleState	oldState = m_state;

	switch (m_state) {
	case ROOT_STATE:
		if (strcmp( name, ADRENALIN ) == 0) {
			checkDup( ADRENALIN );
			m_state = ADRENALIN_STATE;
			break;
		}
		break;
	case ADRENALIN_STATE:
		if (strcmp( name, SYSOP ) == 0) {
			checkDup( SYSOP );
			m_state = SYSOP_STATE;
			break;
		}
		if (strcmp( name, PATHS ) == 0) {
			checkDup( PATHS );
			m_state = PATHS_STATE;
			break;
		}
		if (strcmp( name, ALIAS ) == 0) {
			m_state = ALIAS_STATE;
			break;
		}
		if (strcmp( name, LOG ) == 0) {
			checkDup( LOG );
			m_state = LOG_STATE;
			break;
		}
		if (strcmp( name, DESCRIPTIONS ) == 0) {
			checkDup( DESCRIPTIONS );
			m_state = DESC_STATE;
			break;
		}
		if (strcmp( name, FILES ) == 0) {
			checkDup( FILES );
			m_state = FILES_STATE;
			break;
		}
		if (strcmp( name, FILELIST ) == 0) {
			checkDup( FILELIST );
			m_state = FILELIST_STATE;
			break;
		}
		if (strcmp( name, MAIL ) == 0) {
			checkDup( MAIL );
			m_state = MAIL_STATE;
			break;
		}
		if (strcmp( name, LNGFILE ) == 0) {
			checkDup( LNGFILE );
			m_state = LNGFILE_STATE;
			break;
		}
		if (strcmp( name, FORWARDEXPIRATIONTIME ) == 0) {
			checkDup( FORWARDEXPIRATIONTIME );
			m_state = FORWARDEXPIRATIONTIME_STATE;
			break;
		}
		if (strcmp( name, MAXTICAGE ) == 0) {
			checkDup( MAXTICAGE );
			m_state = MAXTICAGE_STATE;
			break;
		}
		if (strcmp( name, WARNINGINSTEADOFERROR ) == 0) {
			checkDup( WARNINGINSTEADOFERROR );
			m_state = WARNINGINSTEADOFERROR_STATE;
			break;
		}
		if (strcmp( name, HATCHORIGINADDRESS ) == 0) {
			checkDup( HATCHORIGINADDRESS );
			m_state = HATCHORIGINADDRESS_STATE;
			break;
		}
		if (strcmp( name, ALLAKASEEN ) == 0) {
			checkDup( ALLAKASEEN );
			m_state = ALLAKASEEN_STATE;

			// element exists => turn on flag
			CCommonData::getInstance()->setAllAKASeen( true );

			break;
		}
		if (strcmp( name, HANDLEREPLACES ) == 0) {
			checkDup( HANDLEREPLACES );
			m_state = HANDLEREPLACES_STATE;

			// element exists => turn on flag
			CCommonData::getInstance()->setHandleReplaces( true );

			break;
		}
		if (strcmp( name, FREEDRIVESPACELIMIT ) == 0) {
			const CAttribute*	pathAttr = attributes.getAttribute( "path" );
			if (pathAttr == NULL) {
				m_parser->raiseEx( "\"path\" attribute required" );
			}
			m_drivePath = pathAttr->getValue();			

			m_state = FREEDRIVESPACELIMIT_STATE;
			break;
		}
		if (strcmp( name, FILEFLAGS ) == 0) {
			checkDup( FILEFLAGS );
			m_state = FILEFLAGS_STATE;
			break;
		}
		break;
	case SYSOP_STATE:
		if (strcmp( name, NAME ) == 0) {
			checkDup( NAME );
			m_state = SYSOP_NAME_STATE;
			break;
		}
		if (strcmp( name, ADDRESS ) == 0) {
			checkDup( ADDRESS );
			m_state = SYSOP_ADDRESS_STATE;
			break;
		}
		if (strcmp( name, AKA ) == 0) {
			if (! CCommonData::getInstance()->getSysopAddress().isDefined()) {
				// main address is not defined
				m_parser->raiseEx( "Please define sysop address before defining any AKA addresses" );
			}
			m_state = SYSOP_AKA_STATE;
			break;
		}
		break;
	case PATHS_STATE:
		if (strcmp( name, INBOUND ) == 0) {
			checkDup( INBOUND );
			m_state = PATHS_INBOUND_STATE;
			break;
		}
		if (strcmp( name, BINKOUTBOUND ) == 0) {
			checkDup( BINKOUTBOUND );
			m_state = PATHS_BINKOUTBOUND_STATE;
			const CAttribute*	useDomainAttr = attributes.getAttribute( "useDomain" );
			if (useDomainAttr != NULL) {
				CCommonData::getInstance()->setUseDomainInBSO( parseYesNo( useDomainAttr->getValue() ) );
			}
			break;
		}
		if (strcmp( name, FILEBOXES ) == 0) {
			checkDup( FILEBOXES );
			checkDup( LONGFILEBOXES, false );
			m_state = PATHS_FILEBOXES_STATE;
			break;
		}
		if (strcmp( name, LONGFILEBOXES ) == 0) {
			checkDup( LONGFILEBOXES );
			checkDup( FILEBOXES, false );
			m_state = PATHS_LONGFILEBOXES_STATE;
			break;
		}
		if (strcmp( name, NETMAIL ) == 0) {
			checkDup( NETMAIL );
			m_state = PATHS_NETMAIL_STATE;
			break;
		}
		if (strcmp( name, BADTICS ) == 0) {
			checkDup( BADTICS );
			m_state = PATHS_BADTICS_STATE;
			break;
		}
		if (strcmp( name, NEWTICS ) == 0) {
			checkDup( NEWTICS );
			m_state = PATHS_NEWTICS_STATE;
			break;
		}
		if (strcmp( name, PROCESSEDTICS ) == 0) {
			checkDup( PROCESSEDTICS );
			m_state = PATHS_PROCESSEDTICS_STATE;
			break;
		}
		if (strcmp( name, AUTOCREATEDAREAS ) == 0) {
			checkDup( AUTOCREATEDAREAS );
			m_state = PATHS_AUTOCREATEDAREAS_STATE;
			const CAttribute*	longPathsAttr = attributes.getAttribute( "longPaths" );
			if (longPathsAttr != NULL) {
				CCommonData::getInstance()->setLongAutocreatedPaths( parseYesNo( longPathsAttr->getValue() ) );
			}
			break;
		}
		break;
	case LOG_STATE:
		if (strcmp( name, COMMONFILE ) == 0) {
			checkDup( COMMONFILE );
			m_state = LOG_COMMONFILE_STATE;
			break;
		}
		if (strcmp( name, REQUESTSFILE ) == 0) {
			checkDup( REQUESTSFILE );
			m_state = LOG_REQUESTSFILE_STATE;
			break;
		}
		if (strcmp( name, SHOWMUSTGOON ) == 0) {
			checkDup( SHOWMUSTGOON );
			m_state = LOG_SHOWMUSTGOON_STATE;

			// element exists => turn on flag
			CCommonData::getInstance()->setShowMustGoOn( true );

			break;
		}
		if (strcmp( name, LOGMSGNAMES ) == 0) {
			checkDup( LOGMSGNAMES );
			m_state = LOG_LOGMSGNAMES_STATE;

			// element exists => turn on flag
			CCommonData::getInstance()->setLogMsgNames( true );

			break;
		}
		break;
	case DESC_STATE:
		if (strcmp( name, AUTODESCFILE ) == 0) {
			m_state = DESC_AUTODESCFILE_STATE;
			break;
		}
		break;
	case FILES_STATE:
		if (strcmp( name, RENAMEPLAN ) == 0) {

			const CAttribute*	typeAttr = attributes.getAttribute( "type" );
			if (typeAttr == NULL) {
				m_parser->raiseEx( "type attribute of renamePlan absent" );
			}

			const char*		value = typeAttr->getValue();
			RenameFilePlan	plan = RFP_ABSENT;
			if (strcmp( value, "increaseExt" ) == 0) {
				plan = RFP_INCREASE_EXT;
			} else
			if (strcmp( value, "addExtPrefix" ) == 0) {
				plan = RFP_ADD_EXT_PREFIX;
			} else
			if (strcmp( value, "addExtSuffix" ) == 0) {
				plan = RFP_ADD_EXT_SUFFIX;
			} else {
				m_parser->raiseEx( (std::string("unknown rename file plan: ") + value).c_str() );
			}

			CCommonData::getInstance()->setRenameFilePlan( plan );

			m_state = FILES_RENAMEPLAN_STATE;
			break;
		}
		break;
	case FILELIST_STATE:
		if (strcmp( name, NAME ) == 0) {
			checkDup( NAME );
			m_state = FILELIST_NAME_STATE;
			break;
		}
		if (strcmp( name, FORMAT ) == 0) {
			checkDup( FORMAT );
			m_state = FILELIST_FORMAT_STATE;
			break;
		}
		if (strcmp( name, SORT ) == 0) {
			checkDup( SORT );
			CCommonData::getInstance()->setSortFileList( true );
			m_state = FILELIST_SORT_STATE;
			break;
		}
		break;
	case MAIL_STATE:
		if (strcmp( name, KILLPROCESSED ) == 0) {
			checkDup( KILLPROCESSED );
			m_state = MAIL_KILLPROCESSED_STATE;

			// element exists => turn on flag
			CCommonData::getInstance()->setKillProcessedMail( true );

			break;
		}
		if (strcmp( name, REPLYATTRIBUTES ) == 0) {
			checkDup( REPLYATTRIBUTES );
			m_state = MAIL_REPLYATTRIBUTES_STATE;
			break;
		}
		if (strcmp( name, CREATIONLIMIT ) == 0) {
			checkDup( CREATIONLIMIT );
			m_state = MAIL_CREATIONLIMIT_STATE;
			break;
		}
		if (strcmp( name, NONOTIFY ) == 0) {
			checkDup( NONOTIFY );
			m_state = MAIL_NONOTIFY_STATE;
			break;
		}
		if (strcmp( name, NOTIFYONERRORS ) == 0) {
			checkDup( NOTIFYONERRORS );
			m_state = MAIL_NOTIFYONERRORS_STATE;
			CCommonData::getInstance()->setNotifyOnErrors( true );
			break;
		}
		if (strcmp( name, NOTIFYONWARNINGS ) == 0) {
			checkDup( NOTIFYONWARNINGS );
			m_state = MAIL_NOTIFYONWARNINGS_STATE;
			CCommonData::getInstance()->setNotifyOnWarnings( true );
			break;
		}
		break;
	case MAIL_REPLYATTRIBUTES_STATE:
		if (strcmp( name, ATTR ) == 0) {
			m_state = MAIL_REPLYATTRIBUTES_ATTR_STATE;
			break;
		}
		break;
	case MAIL_CREATIONLIMIT_STATE:
		if (strcmp( name, KBYTES ) == 0) {
			checkDup( KBYTES );
			m_state = MAIL_CREATIONLIMIT_KBYTES_STATE;
			break;
		}
		if (strcmp( name, LINES ) == 0) {
			checkDup( LINES );
			m_state = MAIL_CREATIONLIMIT_LINES_STATE;
			break;
		}
		break;
	case FILEFLAGS_STATE:
		if (strcmp( name, ONEVENT ) == 0) {
			const CAttribute*	typeAttr = attributes.getAttribute( "type" );
			if (typeAttr == NULL) {
				m_parser->raiseEx( "\"type\" attribute required" );
			}
			if (strcmp( typeAttr->getValue(), "fix" ) == 0) {
				m_executeEvent = EE_FIX;
			} else 
			if (strcmp( typeAttr->getValue(), "toss" ) == 0) {
				m_executeEvent = EE_TOSS;
			} else
			if (strcmp( typeAttr->getValue(), "hatch" ) == 0) {
				m_executeEvent = EE_HATCH;
			} else
			if (strcmp( typeAttr->getValue(), "any" ) == 0) {
				m_executeEvent = EE_ANY;
			} else {
				m_parser->raiseEx( "Unknown event type (we expecting only \"fix\", \"toss\", \"hatch\" or \"any\")" );
			}

			m_state = FILEFLAGS_ONEVENT_STATE;
			break;
		}
		if (strcmp( name, ONFILE ) == 0) {
			// read and remember mask
			const CAttribute*	maskAttr = attributes.getAttribute( "mask" );
			if (maskAttr == NULL) {
				m_parser->raiseEx( "\"mask\" attribute required" );
			}
			m_fileMask = maskAttr->getValue();

			m_state = FILEFLAGS_ONFILE_STATE;
			break;
		}
		break;
	}
	if (oldState == m_state) {
		// we didn't change the state => we have unexpected element
		m_parser->raiseEx( "Unexpected element" );
	}
	m_parser->checkUnexpected( attributes );
}


void CAdrenalinContentHandler::endElement(const char* name)
	throw (CMiniSAXException)
{
	HandleState	newState = m_state;

	switch (m_state) {
	case ADRENALIN_STATE:
		if (strcmp( name, ADRENALIN ) == 0) {
			// check bound elements
			if (!hasEntry( SYSOP )) {
				m_parser->raiseEx( "Sysop undefined" );
			}
			if (!hasEntry( PATHS )) {
				m_parser->raiseEx( "Paths undefined" );
			}

			// do some post-processing

			CommonDataPtr	cd = CCommonData::getInstance();
			tstring			fileName;
			// ensure that common log file name is OK
			fileName = cd->getCommonLogFileName();
			CAdrenalin::adjustFileName( fileName );
			cd->setCommonLogFileName( fileName );
			// ensure that language file name is OK
			fileName = cd->getLanguageFileName();
			CAdrenalin::adjustFileName( fileName );
			cd->setLanguageFileName( fileName );
			// ensure that reply atributes are set
			if (cd->getReplyMessageAttributes() == 0) {
				cd->setReplyMessageAttributes( ATTR_PRIVATE | ATTR_LOCAL );
			}

			newState = ROOT_STATE;
			break;
		}
		break;
	case SYSOP_STATE:
		if (strcmp( name, SYSOP ) == 0) {
			// check bound elements
			if (!hasEntry( NAME )) {
				m_parser->raiseEx( "Sysop name undefined" );
			}
			if (!hasEntry( ADDRESS )) {
				m_parser->raiseEx( "Sysop address undefined" );
			}

			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case SYSOP_NAME_STATE:
		if (strcmp( name, NAME ) == 0) {
			newState = SYSOP_STATE;
			break;
		}
		break;
	case SYSOP_ADDRESS_STATE:
		if (strcmp( name, ADDRESS ) == 0) {
			newState = SYSOP_STATE;
			break;
		}
		break;
	case SYSOP_AKA_STATE:
		if (strcmp( name, AKA ) == 0) {
			newState = SYSOP_STATE;
			break;
		}
		break;
	case PATHS_STATE:
		if (strcmp( name, PATHS ) == 0) {
			// check bound elements
			if (!hasEntry( INBOUND )) {
				m_parser->raiseEx( "Inbound path undefined" );
			}
			if (!hasEntry( NETMAIL )) {
				m_parser->raiseEx( "Netmail path undefined" );
			}
			if (!hasEntry( BINKOUTBOUND ) || (!hasEntry( FILEBOXES ) && !hasEntry( LONGFILEBOXES ))) {
				m_parser->raiseEx( "Neither bink outbound nor fileboxes defined" );
			}
			if (!hasEntry( BADTICS )) {
				m_parser->raiseEx( "Path for bad tics undefined" );
			}
			if (hasEntry( BINKOUTBOUND ) && !hasEntry( NEWTICS )) {
				m_parser->raiseEx( "Path for new tics undefined" );
			}
			if (!hasEntry( AUTOCREATEDAREAS )) {
				m_parser->raiseEx( "Path for autocreated areas undefined" );
			}

			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case PATHS_INBOUND_STATE:
		if (strcmp( name, INBOUND ) == 0) {
			newState = PATHS_STATE;
			break;
		}
		break;
	case PATHS_BINKOUTBOUND_STATE:
		if (strcmp( name, BINKOUTBOUND ) == 0) {
			newState = PATHS_STATE;
			break;
		}
		break;
	case PATHS_FILEBOXES_STATE:
		if (strcmp( name, FILEBOXES ) == 0) {
			newState = PATHS_STATE;
			break;
		}
		break;
	case PATHS_LONGFILEBOXES_STATE:
		if (strcmp( name, LONGFILEBOXES ) == 0) {
			newState = PATHS_STATE;
			break;
		}
		break;
	case PATHS_NETMAIL_STATE:
		if (strcmp( name, NETMAIL ) == 0) {
			newState = PATHS_STATE;
			break;
		}
		break;
	case PATHS_BADTICS_STATE:
		if (strcmp( name, BADTICS ) == 0) {
			newState = PATHS_STATE;
			break;
		}
		break;
	case PATHS_NEWTICS_STATE:
		if (strcmp( name, NEWTICS ) == 0) {
			newState = PATHS_STATE;
			break;
		}
		break;
	case PATHS_PROCESSEDTICS_STATE:
		if (strcmp( name, PROCESSEDTICS ) == 0) {
			newState = PATHS_STATE;
			break;
		}
		break;
	case PATHS_AUTOCREATEDAREAS_STATE:
		if (strcmp( name, AUTOCREATEDAREAS ) == 0) {
			newState = PATHS_STATE;
			break;
		}
		break;
	case ALIAS_STATE:
		if (strcmp( name, ALIAS ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case LOG_STATE:
		if (strcmp( name, LOG ) == 0) {
			// check bound elements
			if (!hasEntry( COMMONFILE )) {
				m_parser->raiseEx( "Common log file name undefined" );
			}
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case LOG_COMMONFILE_STATE:
		if (strcmp( name, COMMONFILE ) == 0) {
			newState = LOG_STATE;
			break;
		}
		break;
	case LOG_REQUESTSFILE_STATE:
		if (strcmp( name, REQUESTSFILE ) == 0) {
			newState = LOG_STATE;
			break;
		}
		break;
	case LOG_SHOWMUSTGOON_STATE:
		if (strcmp( name, SHOWMUSTGOON ) == 0) {
			newState = LOG_STATE;
			break;
		}
		break;
	case LOG_LOGMSGNAMES_STATE:
		if (strcmp( name, LOGMSGNAMES ) == 0) {
			newState = LOG_STATE;
			break;
		}
		break;
	case FILES_STATE:
		if (strcmp( name, FILES ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case FILES_RENAMEPLAN_STATE:
		if (strcmp( name, RENAMEPLAN ) == 0) {
			newState = FILES_STATE;
			break;
		}
		break;
	case DESC_STATE:
		if (strcmp( name, DESCRIPTIONS ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case DESC_AUTODESCFILE_STATE:
		if (strcmp( name, AUTODESCFILE ) == 0) {
			newState = DESC_STATE;
			break;
		}
		break;
	case FILELIST_STATE:
		if (strcmp( name, FILELIST ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case FILELIST_NAME_STATE:
		if (strcmp( name, NAME ) == 0) {
			newState = FILELIST_STATE;
			break;
		}
		break;
	case FILELIST_FORMAT_STATE:
		if (strcmp( name, FORMAT ) == 0) {
			newState = FILELIST_STATE;
			break;
		}
		break;
	case FILELIST_SORT_STATE:
		if (strcmp( name, SORT ) == 0) {
			newState = FILELIST_STATE;
			break;
		}
		break;
	case MAIL_STATE:
		if (strcmp( name, MAIL ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case MAIL_KILLPROCESSED_STATE:
		if (strcmp( name, KILLPROCESSED ) == 0) {
			newState = MAIL_STATE;
			break;
		}
		break;
	case MAIL_REPLYATTRIBUTES_STATE:
		if (strcmp( name, REPLYATTRIBUTES ) == 0) {
			newState = MAIL_STATE;
			break;
		}
		break;
	case MAIL_REPLYATTRIBUTES_ATTR_STATE:
		if (strcmp( name, ATTR ) == 0) {
			newState = MAIL_REPLYATTRIBUTES_STATE;
			break;
		}
		break;
	case MAIL_CREATIONLIMIT_STATE:
		if (strcmp( name, CREATIONLIMIT ) == 0) {
			newState = MAIL_STATE;
			break;
		}
		break;
	case MAIL_CREATIONLIMIT_KBYTES_STATE:
		if (strcmp( name, KBYTES ) == 0) {
			newState = MAIL_CREATIONLIMIT_STATE;
			break;
		}
		break;
	case MAIL_CREATIONLIMIT_LINES_STATE:
		if (strcmp( name, LINES ) == 0) {
			newState = MAIL_CREATIONLIMIT_STATE;
			break;
		}
		break;
	case MAIL_NONOTIFY_STATE:
		if (strcmp( name, NONOTIFY ) == 0) {
			newState = MAIL_STATE;
			break;
		}
		break;
	case MAIL_NOTIFYONERRORS_STATE:
		if (strcmp( name, NOTIFYONERRORS ) == 0) {
			newState = MAIL_STATE;
			break;
		}
		break;
	case MAIL_NOTIFYONWARNINGS_STATE:
		if (strcmp( name, NOTIFYONWARNINGS ) == 0) {
			newState = MAIL_STATE;
			break;
		}
		break;
	case LNGFILE_STATE:
		if (strcmp( name, LNGFILE ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case FORWARDEXPIRATIONTIME_STATE:
		if (strcmp( name, FORWARDEXPIRATIONTIME ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case MAXTICAGE_STATE:
		if (strcmp( name, MAXTICAGE ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case WARNINGINSTEADOFERROR_STATE:
		if (strcmp( name, WARNINGINSTEADOFERROR ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case HATCHORIGINADDRESS_STATE:
		if (strcmp( name, HATCHORIGINADDRESS ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case ALLAKASEEN_STATE:
		if (strcmp( name, ALLAKASEEN ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case HANDLEREPLACES_STATE:
		if (strcmp( name, HANDLEREPLACES ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case FREEDRIVESPACELIMIT_STATE:
		if (strcmp( name, FREEDRIVESPACELIMIT ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case FILEFLAGS_STATE:
		if (strcmp( name, FILEFLAGS ) == 0) {
			newState = ADRENALIN_STATE;
			break;
		}
		break;
	case FILEFLAGS_ONEVENT_STATE:
		if (strcmp( name, ONEVENT ) == 0) {
			newState = FILEFLAGS_STATE;
			break;
		}
		break;
	case FILEFLAGS_ONFILE_STATE:
		if (strcmp( name, ONFILE ) == 0) {
			newState = FILEFLAGS_STATE;
			break;
		}
		break;
	}
	if (newState == m_state) {
		// state not changed => we have unexpected element
		m_parser->raiseEx( "Unexpected end of element" );
	} else {
		removeEntries();
		m_state = newState;
	}
}


void CAdrenalinContentHandler::handleCharacters(const char* buffer, size_t len)
	throw (CMiniSAXException)
{
	switch (m_state) {
		case SYSOP_NAME_STATE:
			CCommonData::getInstance()->setSysopName( tstring( buffer, len ) );
			return;
		case SYSOP_ADDRESS_STATE:
			{
				CAddress	address;
				// if address is wrong or some part undefined (except domain) we throw exception
				if (!address.parseString( tstring( buffer, len ).c_str() )) {
					m_parser->raiseEx( "Failed to parse address" );
				}
				if (address.getZone() == -1 || address.getNet() == -1 || address.getNode() == -1 || address.getPoint() == -1) {
					m_parser->raiseEx( "You should specify address in form zone:net/node or zone:net/node.point with optional domain" );
				}
				if (address.getDomain().length() == 0) {
					// default domain is fidonet.org
					address.setDomain( TEXT("fidonet.org") );
				}

				CCommonData::getInstance()->setSysopAddress( address );
			}
			return;
		case SYSOP_AKA_STATE:
			{
				CAddress	address;
				if (!address.parseString( tstring( buffer, len ).c_str() ))
				{
					m_parser->raiseEx( "Invalid AKA address" );
				}
				CommonDataPtr	commonData = CCommonData::getInstance();
				address.baseUpon( commonData->getSysopAddress() );
				if (address == commonData->getSysopAddress()) {
					m_parser->raiseEx( "AKA address is the same as base address" );
				}
				if (commonData->hasSysopAKA( address )) {
					m_parser->raiseEx( "AKA address is already defined" );
				}
				commonData->addSysopAKA( address );
			}
			return;
		case PATHS_INBOUND_STATE:
			{
				tstring	path( buffer, len );
				CAdrenalin::adjustFilePath( path );
				CCommonData::getInstance()->setInboundPath( path );
			}
			return;
		case PATHS_BINKOUTBOUND_STATE:
			{
				tstring	path( buffer, len );
				CAdrenalin::adjustFilePath( path );
				CCommonData::getInstance()->setBinkOutboundPath( path );
			}
			return;
		case PATHS_FILEBOXES_STATE:
			{
				tstring	path( buffer, len );
				CAdrenalin::adjustFilePath( path );
				CCommonData::getInstance()->setFileBoxesPath( path );
				CCommonData::getInstance()->setLongFileBoxes( false );
			}
			return;
		case PATHS_LONGFILEBOXES_STATE:
			{
				tstring	path( buffer, len );
				CAdrenalin::adjustFilePath( path );
				CCommonData::getInstance()->setFileBoxesPath( path );
				CCommonData::getInstance()->setLongFileBoxes( true );
			}
			return;
		case PATHS_NETMAIL_STATE:
			{
				tstring	path( buffer, len );
				CAdrenalin::adjustFilePath( path );
				CCommonData::getInstance()->setNetmailPath( path );
			}
			return;
		case PATHS_NEWTICS_STATE:
			{
				tstring	path( buffer, len );
				CAdrenalin::adjustFilePath( path );
				CCommonData::getInstance()->setNewTicsPath( path );
			}
			return;
		case PATHS_BADTICS_STATE:
			{
				tstring	path( buffer, len );
				CAdrenalin::adjustFilePath( path );
				CCommonData::getInstance()->setBadTicsPath( path );
			}
			return;
		case PATHS_PROCESSEDTICS_STATE:
			{
				tstring	path( buffer, len );
				CAdrenalin::adjustFilePath( path );
				CCommonData::getInstance()->setProcessedTicsPath( path );
			}
			return;
		case PATHS_AUTOCREATEDAREAS_STATE:
			{
				tstring	path( buffer, len );
				// append path separator if absent
				if (!path.empty()) {
					if (path[path.length()-1] != PATH_SEPARATOR) {
						path += PATH_SEPARATOR_STR;
					}
				}

				CCommonData::getInstance()->setAutocreatedAreasPath( path );
			}
			return;
		case ALIAS_STATE:
			{
				tstring	alias( buffer, len );
				if (CCommonData::getInstance()->hasAlias( alias )) {
					m_parser->raiseEx( "The same alias is already defined" );
				}
				CCommonData::getInstance()->addAlias( alias );
			}
			return;
		case LOG_COMMONFILE_STATE:
			{
				tstring	name( buffer, len );
				CAdrenalin::adjustFileName( name );
				CCommonData::getInstance()->setCommonLogFileName( name );
			}
			return;
		case LOG_REQUESTSFILE_STATE:
			{
				tstring	name( buffer, len );
				CAdrenalin::adjustFileName( name );
				CCommonData::getInstance()->setRequestsLogFileName( name );
			}
			return;
		case DESC_AUTODESCFILE_STATE:
			{
				tstring	name( buffer, len );
				CAdrenalin::adjustFileName( name );
				CCommonData::getInstance()->addAutoDescFileName( name );
			}
			return;
		case FILELIST_NAME_STATE:
			CCommonData::getInstance()->setFileListName( tstring( buffer, len ) );
			return;
		case FILELIST_FORMAT_STATE:
			CCommonData::getInstance()->setFileListFormat( tstring( buffer, len ) );
			return;
		case MAIL_REPLYATTRIBUTES_ATTR_STATE:
			{
				uint16	attr = 0, flag = 0;
				if (!GetAttrOrFlag( tstring( buffer, len ).c_str(), attr, flag )) {
					m_parser->raiseEx( "Unknown attribute" );
				}
				CommonDataPtr	commonData = CCommonData::getInstance();
				commonData->setReplyMessageAttributes( commonData->getReplyMessageAttributes() | attr );
				commonData->setReplyMessageFlags( commonData->getReplyMessageFlags() | flag );
			}
			return;
		case MAIL_CREATIONLIMIT_KBYTES_STATE:
			{
				tstring	limitStr( buffer, len );
				int		limit;
				int	scanResult = sscanf( limitStr.c_str(), "%d", &limit );
				if (scanResult == 0 || scanResult == EOF || limit <= 0) {
					m_parser->raiseEx( "Invalid creation limit (it should be integer above zero)" );
				}
				CCommonData::getInstance()->setMailCreationLimitInKBytes( limit );
			}
			return;
		case MAIL_CREATIONLIMIT_LINES_STATE:
			{
				tstring	limitStr( buffer, len );
				int		limit;
				int	scanResult = sscanf( limitStr.c_str(), "%d", &limit );
				if (scanResult == 0 || scanResult == EOF || limit <= 0) {
					m_parser->raiseEx( "Invalid creation limit (it should be integer above zero)" );
				}
				CCommonData::getInstance()->setMailCreationLimitInLines( limit );
			}
			return;
		case MAIL_NONOTIFY_STATE:
			{
				if (lstrcmp( tstring( buffer, len ).c_str(), "basePurge" ) == 0) {
					CCommonData::getInstance()->setNoMailNotifyForBasePurge( true );
				} else {
					m_parser->raiseEx( "Unexpected data (we support only basePurge)" );
				}				
			}
			return;
		case LNGFILE_STATE:
			{
				tstring	name( buffer, len );
				CAdrenalin::adjustFileName( name );
				CCommonData::getInstance()->setLanguageFileName( name );
			}
			return;
		case FORWARDEXPIRATIONTIME_STATE:
			{
				tstring	timeStr( buffer, len );
				int		time;
				int	scanResult = sscanf( timeStr.c_str(), "%d", &time );
				if (scanResult == 0 || scanResult == EOF || time <= 0) {
					m_parser->raiseEx( "Invalid forward expiration time (it should be integer above zero)" );
				}
				CCommonData::getInstance()->setForwardExpirationTime( time );
			}
			return;
		case MAXTICAGE_STATE:
			{
				tstring	ageStr( buffer, len );
				int		age;
				int	scanResult = sscanf( ageStr.c_str(), "%d", &age );
				if (scanResult == 0 || scanResult == EOF || age <= 0) {
					m_parser->raiseEx( "Invalid max tic age (it should be integer above zero)" );
				}
				CCommonData::getInstance()->setMaxTicAge( age );
			}
			return;
		case WARNINGINSTEADOFERROR_STATE:
			{
				if (lstrcmp( tstring( buffer, len ).c_str(), "incorrectCRC32" ) == 0) {
					CCommonData::getInstance()->setWarnOnIncorrectCRC32( true );
				} else {
					m_parser->raiseEx( "Unexpected data (we support only incorrectCRC32)" );
				}				
			}
			return;
		case HATCHORIGINADDRESS_STATE:
			{
				tstring	option( buffer, len );
				if (lstrcmp( option.c_str(), "main" ) == 0) {
					CCommonData::getInstance()->setHatchOriginAddressMain( true );
				} else 
				if (lstrcmp( option.c_str(), "aka" ) == 0) {
					CCommonData::getInstance()->setHatchOriginAddressMain( false );
				} else {
					m_parser->raiseEx( "Unexpected data (we support only \"main\" or \"aka\")" );
				}			
			}
			return;
		case FREEDRIVESPACELIMIT_STATE:
			{
				int	kBytes;
				if (!parseIntWithCommas( tstring( buffer, len ).c_str(), kBytes )) {
					m_parser->raiseEx( "Invalid drive space limit (it should be integer above zero, possibly comma-separated)" );
				}
				CDriveSpaceLimit	limit;
				limit.setPath( m_drivePath.c_str() );
				limit.setLimit( kBytes );
				CCommonData::getInstance()->addDriveSpaceLimit( limit );
			}
			return;
		case FILEFLAGS_ONEVENT_STATE:
			{
				// we don't need to adjust file path evidently because CEventFileFlag does it itself
				tstring	fileFlag( buffer, len );
				if (m_executeEvent == EE_ANY) {
					CCommonData::getInstance()->addEventFileFlag( CEventFileFlag( EE_FIX,   fileFlag ) );
					CCommonData::getInstance()->addEventFileFlag( CEventFileFlag( EE_TOSS,  fileFlag ) );
					CCommonData::getInstance()->addEventFileFlag( CEventFileFlag( EE_HATCH, fileFlag ) );
				} else {
					CCommonData::getInstance()->addEventFileFlag( CEventFileFlag( m_executeEvent, fileFlag ) );
				}
			}
			return;
		case FILEFLAGS_ONFILE_STATE:
			{
				// we don't need to adjust file path evidently because CMaskFileFlag does it itself
				CCommonData::getInstance()->addMaskFileFlag( CMaskFileFlag( m_fileMask, tstring( buffer, len ) ) );
			}
			return;
	}
	if (!hasOnlySpaces( buffer, len )) {
		m_parser->raiseEx( "Unexpected symbols" );
	}
}


bool CAdrenalinContentHandler::parseYesNo(LPCTSTR flag) throw(CMiniSAXException) {
	if (strcmp( flag, YES ) == 0) {
		return true;
	} else
	if (strcmp( flag, NO ) == 0) {
		return false;
	} else {
		m_parser->raiseEx( "Unexpected data (we are waiting for \"yes\" or \"no\")" );
		return false;
	}
}


void CAdrenalinContentHandler::checkDup(LPCTSTR name, bool remember /* = true */) throw(CMiniSAXException) {
	if (hasEntry( name )) {
		m_parser->raiseEx( (tstring("Element '") + tstring(name) + tstring("' was already defined")).c_str() );
	} else {
		if (remember) {
			rememberEntry( name );
		}
	}
}


void CAdrenalinContentHandler::rememberEntry(LPCTSTR name) {
	EntryDescriptor	desc;
	desc.m_name  = name;
	desc.m_state = m_state;
	m_entriesHolder.push_back( desc );
}


void CAdrenalinContentHandler::removeEntry(LPCTSTR name) {
	EntriesHolder::iterator	it = getEntry(name);
	if (it != m_entriesHolder.end()) {
		m_entriesHolder.erase( it );
	}
}

CAdrenalinContentHandler::EntriesHolder::iterator CAdrenalinContentHandler::getEntry(LPCTSTR name) {
	for (EntriesHolder::iterator it = m_entriesHolder.begin();
		 it != m_entriesHolder.end(); ++it)
	{
		if ((*it).m_name == name && (*it).m_state == m_state) {
			return it;
		}
	}
	return m_entriesHolder.end();
}


bool CAdrenalinContentHandler::hasEntry(LPCTSTR name) {
	return getEntry( name ) != m_entriesHolder.end();
}


void CAdrenalinContentHandler::removeEntries() {
	EntriesHolder::iterator it = m_entriesHolder.begin();
	while (it != m_entriesHolder.end())	{
		if ((*it).m_state == m_state) {
			it = m_entriesHolder.erase( it );
		} else {
			++it;
		}
	}
}
