#pragma once

#include "FTNStruc.hpp"
#include "Flag.h"
#include "LineFormat.h"

#include "Singletons.h"
#include "Address.h"

enum { ROUTE_NONE = 0, ROUTE_SEND = 1, ROUTE_RECEIVE = 2, ROUTE_BIDIRECT = 3 };

enum { ACTIVITY_USUAL, ACTIVITY_PASSIVE, ACTIVITY_UNAVAILABLE };

enum { MSGDATA_NONE, MSGDATA_COMMAND, MSGDATA_SUBSCRIBE, MSGDATA_UNSUBSCRIBE };

enum OutboundType {
	OT_BINKLEY,
	OT_FILE_BOXES
};

enum FlavourType {
    FT_NORMAL    = 0,
    FT_CRASH     = 1,
    FT_DIRECT    = 2,
    FT_HOLD      = 4,
    FT_IMMEDIATE = 8
};

//const	uint16	WARN_SYSOP_ABOUT_WARNINGS = 1,
//		WARN_SYSOP_ABOUT_ERRORS   = 2;
/*
#if defined(UNICODE)
typedef	wofstream	tofstream;
#define	COUT	wcout
#define	CERR	wcerr
#else
typedef ofstream	tofstream;
#define	COUT	cout
#define	CERR	cerr
#endif
*/

const LPCTSTR	TEXT_EOL = TEXT("\n");	// end of line string
const TCHAR		LF       = TEXT('\n');

const TCHAR		PATH_SEPARATOR     = TEXT('\\');
const LPCTSTR	PATH_SEPARATOR_STR = TEXT("\\");

const TCHAR	SPACE_CHAR = TEXT(' ');
const TCHAR	TAB_CHAR   = TEXT('\011');
const TCHAR	ZERO_CHAR  = TEXT('\000');

typedef	vector<SFTNAddress>	VAddr;

struct SMsgData {
	int		m_iType;
	tstring	m_sLine;
	tstring	m_sTail;
};

// common constants

const int	LONG_DESC_PREFIX_SIZE = 128;
const int	TEARLINE_SIZE         = 80;

extern	LPCTSTR	c_szForwardsFileName;

struct SGlobalVars {
//	SFTNAddress		m_addrMain;
//	VAddr			m_vAKAAddr;
//	TCHAR			m_szSysOp[36];
//	TCHAR			m_szInboundPath[MAX_PATH];
//	TCHAR			m_szOutboundPath[MAX_PATH];
//	TCHAR			m_szBinkOutboundPath[MAX_PATH];
//	TCHAR			m_szNetmailPath[MAX_PATH];
//	TCHAR			m_szBadTicsPath[MAX_PATH];
//	TCHAR			m_szNewTicsPath[MAX_PATH];
//	TCHAR			m_szProcessedTicsPath[MAX_PATH];
//	TCHAR			m_szAutoCreatedPath[MAX_PATH];
//	TCHAR			m_szAutoDescFileName[MAX_PATH];
//	TCHAR			m_szLongDescPrefix[LONG_DESC_PREFIX_SIZE+1];
//	TCHAR			m_szFilelistFileName[MAX_PATH];
//	CLineFormat		m_fileListFormat;
//	CFlag			m_fLongPaths;
//	vector<tstring>		m_vsAliases;
	vector<SMsgData>	m_vMsgData;
//	CFlag			m_fHandleReplaces;
//	CFlag			m_fHatchOriginMain;
//	CFlag			m_fAllAKASeen;
//	CFlag			m_fKillProcessedMail;
//	CFlag			m_fShowMustGoOn;
//	CFlag			m_fLogMsgNames;
//	TCHAR			m_szFixFlagName[MAX_PATH];
//	TCHAR			m_szTossFlagName[MAX_PATH];
//	TCHAR			m_szHatchFlagName[MAX_PATH];
//	CFlag			m_fCheckDomainInBSO;
//	int				m_nForwardExpirationTime;	// days
	void			FreeMsgData();
//	CFlag			m_flagConcatDescriptions;	// concat Desc & LDesc lines
								// while creating TIC
//	int			m_nMessageKBytesLimit;	// in _bytes_
//	int			m_nMessageLinesLimit;

	TCHAR			m_szTearLine[TEARLINE_SIZE];

//	uint16			m_nWarnSysopByMail;	// bitfield
										// bits to check:
										// WARN_SYSOP_ABOUT_WARNINGS
										// WARN_SYSOP_ABOUT_ERRORS

//	uint16			m_nNoMailNotify;	// bitfield
										// bits to check
										// 1 - no notify about base purge

//	uint16			m_nReplyMessageAttrs;
//	uint16			m_nReplyMessageFlags;
//	int				maxTicAge;
};

extern SGlobalVars g_vars;


//struct SWarnings {
//	CFlag	m_flagIncorrectCRC32;
//};

//extern SWarnings g_warnings;


class CDriveSpaceLimit {
public:
	CDriveSpaceLimit() {}
	CDriveSpaceLimit(const CDriveSpaceLimit& driveLimit) : m_limit(driveLimit.m_limit)
	{
		lstrcpy( m_drivePath, driveLimit.m_drivePath );
	}

	LPCTSTR getPath() const { return m_drivePath; }
	void    setPath(LPCTSTR newPath) { lstrcpy( m_drivePath, newPath ); }

	DWORD   getLimit() const { return m_limit; }
	void    setLimit(DWORD newLimit) { m_limit = newLimit; }

protected:
	TCHAR	m_drivePath[MAX_PATH];
	DWORD	m_limit;
};

// events that can happen while executing adrenalin
// these events can be used to create file flags
enum ExecuteEvent {
	EE_FIX,
	EE_TOSS,
	EE_HATCH,
	EE_ANY		// this type of event is used only when parsing configuration
				// it is substituted by a set of all events when storing file flags in a vector
};

// Base class for file flags, allows to store file name of the flag.
// When file name supplied we store both original file name and adjusted file name
// (adjusted means that file path have been changed relatively to the executable path).
// Having original file name allows us to rewrite configuration correctly (with the same
// name that was specified).
class CFileFlagBase {
public:
	CFileFlagBase() : m_fileName(), m_used(false) {}
	CFileFlagBase(const tstring& fileName, bool used = false);

	const tstring& getFileName() const { return m_fileName; }
	void setFileName(const tstring& newName);

	const tstring& getOriginalFileName() const { return m_originalFileName; }

	bool isUsed() const     { return m_used; }
	void setUsed(bool flag) { m_used = flag; }

protected:
	tstring	m_fileName;
	tstring	m_originalFileName;
	bool	m_used;
};

// file flag that is created on event
class CEventFileFlag : public CFileFlagBase {
public:
	CEventFileFlag() : m_event() {}
	CEventFileFlag(ExecuteEvent event, const tstring& fileName, bool used = false) : CFileFlagBase(fileName, used), m_event(event) {}

	ExecuteEvent getEvent() const { return m_event; }
	void setEvent(ExecuteEvent event) { m_event = event; }

protected:
	ExecuteEvent	m_event;
};

// file flag that is created on file processed (tossed or hatched) with name that satisfies the mask
class CMaskFileFlag : public CFileFlagBase {
public:
	CMaskFileFlag() : m_mask() {}
	CMaskFileFlag(const tstring& mask, const tstring& fileName, bool used = false) : CFileFlagBase(fileName, used), m_mask(mask) {}

	const tstring& getMask() const { return m_mask; }
	void setMask(const tstring& newMask) { m_mask = newMask; }

protected:
	tstring	m_mask;
};


/**
 * Plans of file renaming; used when file arrived into some fecho but there is a file with the same name;
 * this case we rename old file.
 */
enum RenameFilePlan {
	RFP_ABSENT,				// was not specified for entity (used in CArea)
	RFP_NONE,				// no plan => replace file (default common value)
	RFP_INCREASE_EXT,		// aaa.bbb => aaa.bbc => aaa.bbd => ...
	RFP_ADD_EXT_PREFIX,		// aaa.bbb => aaa.1.bbb => aaa.2.bbb => ... 
	RFP_ADD_EXT_SUFFIX		// aaa.bbb => aaa.bbb.1 => aaa.bbb.2 => ...
};


class CCommonData : public CSingleton {
public:
	// sysop name
	void setSysopName(const tstring& sysopName) { m_sysopName = sysopName; }
	const tstring& getSysopName() const { return m_sysopName; }

	// sysop address
	void setSysopAddress(const CAddress& address) { m_sysopAddress = address; }
	const CAddress& getSysopAddress() const { return m_sysopAddress; }

	// methods for sysop AKAs
	size_t getSysopAKACount() const { return m_sysopAKAs.size(); }
	const CAddress& getSysopAKA(int index) const { return m_sysopAKAs[index]; }
	void addSysopAKA(const CAddress& address) { m_sysopAKAs.push_back( address ); }
	void replaceSysopAKA(int index, const CAddress& address) { m_sysopAKAs[index] = address; }
	void removeSysopAKA(int index) { m_sysopAKAs.erase( m_sysopAKAs.begin() + index );  }
	bool hasSysopAKA(const CAddress& address) const;

	// inbound
	void setInboundPath(const tstring& path) { m_inboundPath = path; }
	const tstring& getInboundPath() const { return m_inboundPath; }

	// bink outbound
	void setBinkOutboundPath(const tstring& path) { m_binkOutboundPath = path; }
	const tstring& getBinkOutboundPath() const { return m_binkOutboundPath; }

	// file boxes

	void setFileBoxesPath(const tstring& newPath) { m_fileBoxesPath = newPath; }
	const tstring& getFileBoxesPath() const { return m_fileBoxesPath; }
	void setLongFileBoxes(bool newFlag) { m_useLongFileBoxes = newFlag; }
	bool areLongFileBoxes() const { return m_useLongFileBoxes; }

	// netmail path
	void setNetmailPath(const tstring& path) { m_netmailPath = path; }
	const tstring& getNetmailPath() const { return m_netmailPath; }

	// new tics path
	void setNewTicsPath(const tstring& path) { m_newTicsPath = path; }
	const tstring& getNewTicsPath() const { return m_newTicsPath; }

	// bad tics path
	void setBadTicsPath(const tstring& path) { m_badTicsPath = path; }
	const tstring& getBadTicsPath() const { return m_badTicsPath; }

	// processed tics path
	void setProcessedTicsPath(const tstring& path) { m_processedTicsPath = path; }
	const tstring& getProcessedTicsPath() const { return m_processedTicsPath; }

	// autocreated areas path
	void setAutocreatedAreasPath(const tstring& path) { m_autocreatedAreasPath = path; }
	const tstring& getAutocreatedAreasPath() const { return m_autocreatedAreasPath; }

	bool areLongAutocreatedPaths() const { return m_useLongAutocreatedPaths; }
	void setLongAutocreatedPaths(bool flag) { m_useLongAutocreatedPaths = flag; }

	// aliases
	int getAliasCount() const { return m_aliases.size(); }
	const tstring& getAlias(int index) const { return m_aliases[index]; }
	void addAlias(const tstring& alias) { m_aliases.push_back( alias ); }
	void replaceAlias(int index, const tstring& alias) { m_aliases[index] = alias; }
	void removeAlias(int index) { m_aliases.erase( m_aliases.begin() + index ); }
	bool hasAlias(const tstring& alias) const;

	// log

	void setCommonLogFileName(const tstring& fileName) { m_commonLogFileName = fileName; }
	const tstring& getCommonLogFileName() const { return m_commonLogFileName; }
	void setRequestsLogFileName(const tstring& fileName) { m_requestsLogFileName = fileName; }
	const tstring& getRequestsLogFileName() const { return m_requestsLogFileName; }

	void setShowMustGoOn(bool flag) { m_showMustGoOn = flag; }
	bool mustShowGoOn() const { return m_showMustGoOn; }
	void setLogMsgNames(bool flag) { m_logMsgNames = flag; }
	bool doLogMsgNames() const { return m_logMsgNames; }

	// descriptions

	size_t getAutoDescFileNameCount() const { return m_autoDescFileNames.size(); }
	void addAutoDescFileName(const tstring& fileName) { m_autoDescFileNames.push_back( fileName ); }
	const tstring& getAutoDescFileName(size_t index) const { return m_autoDescFileNames[index]; }

	// file list

	void setFileListName(const tstring& name) { m_fileListName = name; }
	const tstring& getFileListName() const { return m_fileListName; }

	void setFileListFormat(const tstring& format) { m_fileListFormat.setFormat( format ); }
	CLineFormat& getFileListFormat() { return m_fileListFormat; }

	void setSortFileList(bool sort) { m_sortFileList = sort; }
	bool doSortFileList() const { return m_sortFileList; }

	// mail

	void setKillProcessedMail(bool flag) { m_killProcessedMail = flag; }
	bool doKillProcessedMail() const { return m_killProcessedMail; }

	void setReplyMessageAttributes(uint16 attrs) { m_replyMessageAttributes = attrs; }
	uint16 getReplyMessageAttributes() const { return m_replyMessageAttributes; }
	void setReplyMessageFlags(uint16 flags) { m_replyMessageFlags = flags; }
	uint16 getReplyMessageFlags() const { return m_replyMessageFlags; }

	void setMailCreationLimitInKBytes(int kBytes) { m_mailCreationLimitInKBytes = kBytes; }
	int  getMailCreationLimitInKBytes() const { return m_mailCreationLimitInKBytes; }

	void setMailCreationLimitInLines(int lines) { m_mailCreationLimitInLines = lines; }
	int  getMailCreationLimitInLines() const { return m_mailCreationLimitInLines; }

	void setNoMailNotifyForBasePurge(bool flag) { m_noMailNotifyForBasePurge = flag; }
	bool isNoMailNotifyForBasePurge() const { return m_noMailNotifyForBasePurge; }

	// lng file

	void setLanguageFileName(const tstring& fileName) { m_languageFileName = fileName; }
	const tstring& getLanguageFileName() const { return m_languageFileName; }

	// forward expiration time

	void setForwardExpirationTime(int time) { m_forwardExpirationTime = time; }
	int  getForwardExpirationTime() const   { return m_forwardExpirationTime; }

	// max tic age

	void setMaxTicAge(int age) { m_maxTicAge = age;  }
	int  getMaxTicAge() const  { return m_maxTicAge; }

	// warnings instead of errors

	void setWarnOnIncorrectCRC32(bool flag) { m_warnOnIncorrectCRC32 = flag; }
	bool doWarnOnIncorrectCRC32() const { return m_warnOnIncorrectCRC32; }

	// origin address for hatch

	void setHatchOriginAddressMain(bool flag) { m_hatchOriginAddressMain = flag; }
	bool isHatchOriginAddressMain() const { return m_hatchOriginAddressMain; }

	// all aka seen

	void setAllAKASeen(bool flag) { m_allAKASeen = flag; }
	bool areAllAKASeen() const    { return m_allAKASeen; }

	// handle replaces

	void setHandleReplaces(bool flag) { m_handleReplaces = flag; }
	bool doHandleReplaces() const     { return m_handleReplaces; }

	// drive space limits

	size_t getDriveSpaceLimitCount() const { return m_driveSpaceLimits.size(); }
	const CDriveSpaceLimit& getDriveSpaceLimit(int index) const { return m_driveSpaceLimits[index]; }
	void addDriveSpaceLimit(const CDriveSpaceLimit& limit) { m_driveSpaceLimits.push_back( limit ); }

	// event file flags

	size_t getEventFileFlagCount() const { return m_eventFileFlags.size(); }
	const CEventFileFlag& getEventFileFlag(int index) const { return m_eventFileFlags[index]; }
	void addEventFileFlag(const CEventFileFlag& flag) { m_eventFileFlags.push_back(flag); }

	// mask file flags

	size_t getMaskFileFlagCount() const { return m_maskFileFlags.size(); }
	// two getters: const and usual
	const CMaskFileFlag& getMaskFileFlag(int index) const { return m_maskFileFlags[index]; }
	CMaskFileFlag& getMaskFileFlag(int index) { return m_maskFileFlags[index]; }
	void addMaskFileFlag(const CMaskFileFlag& flag) { m_maskFileFlags.push_back(flag); }

	// use domain in bso

	void setUseDomainInBSO(bool flag) { m_useDomainInBSO = flag; }
	bool doUseDomainInBSO() const { return m_useDomainInBSO; }

	// notification

	void setNotifyOnErrors(bool flag) { m_notifyOnErrors = flag; }
	bool doNotifyOnErrors() const { return m_notifyOnErrors; }

	void setNotifyOnWarnings(bool flag) { m_notifyOnWarnings = flag; }
	bool doNotifyOnWarnings() const { return m_notifyOnWarnings; }

	void setRenameFilePlan(RenameFilePlan plan) { m_renameFilePlan = plan; }
	RenameFilePlan getRenameFilePlan() const { return m_renameFilePlan; }
	bool doesRenameFilePlanSet() const { return (m_renameFilePlan != RFP_NONE); }

	// singleton implementation

	virtual void destroy() {}
	virtual CSingletonVector getDependencies() { return CSingletonVector(); }

	static CCommonData* getInstance() {
		static CCommonData	commonData;

		return &commonData;
	}

protected:
	CCommonData();

	tstring		m_sysopName;
	CAddress	m_sysopAddress;

	typedef std::vector<CAddress>	Addresses;
	Addresses	m_sysopAKAs;

	tstring	m_inboundPath;

	tstring	m_binkOutboundPath;
	bool	m_useDomainInBSO;

	tstring	m_fileBoxesPath;
	bool	m_useLongFileBoxes;

	tstring	m_netmailPath;
	tstring	m_newTicsPath;
	tstring	m_badTicsPath;
	tstring	m_processedTicsPath;
	tstring	m_autocreatedAreasPath;
	bool	m_useLongAutocreatedPaths;

	typedef std::vector<tstring>	Aliases;
	Aliases	m_aliases;

	tstring	m_commonLogFileName;
	tstring	m_requestsLogFileName;
	bool	m_showMustGoOn;
	bool	m_logMsgNames;

	std::vector<tstring>	m_autoDescFileNames;

	tstring		m_fileListName;
	CLineFormat	m_fileListFormat;
	bool		m_sortFileList;

	bool	m_killProcessedMail;
	uint16	m_replyMessageAttributes;
	uint16	m_replyMessageFlags;
	int		m_mailCreationLimitInKBytes;
	int		m_mailCreationLimitInLines;
	bool	m_noMailNotifyForBasePurge;

	tstring	m_languageFileName;

	int		m_forwardExpirationTime;
	int		m_maxTicAge;

	bool	m_warnOnIncorrectCRC32;
	bool	m_hatchOriginAddressMain;
	bool	m_allAKASeen;
	bool	m_handleReplaces;

	typedef std::vector<CDriveSpaceLimit>	DriveSpaceLimits;
	DriveSpaceLimits						m_driveSpaceLimits;

	typedef std::vector<CEventFileFlag>	EventFileFlags;
	EventFileFlags						m_eventFileFlags;

	typedef std::vector<CMaskFileFlag>	MaskFileFlags;
	MaskFileFlags						m_maskFileFlags;

	bool	m_notifyOnErrors;
	bool	m_notifyOnWarnings;

	RenameFilePlan	m_renameFilePlan;
};
typedef CCommonData*	CommonDataPtr;

const LPCTSTR	SUSPENDED_FILES_FILE_NAME = TEXT("suspended.xml");
