#include "pch.h"
#include "Misc.hpp"
#include "LinksManager.h"
#include "LinksSAX.h"

// TODO: check existance of obligatory elements (e.g. defaultGroup)

/**********************
  CLinksContentHandler
 **********************/


CLinksContentHandler::CLinksContentHandler(CMiniSAXParser* aParser) {
	parser = aParser;
	state  = ROOT_STATE;
	link   = NULL;
}


CLinksContentHandler::~CLinksContentHandler() {
	if (link != NULL) {
		delete link;
	}
}


void CLinksContentHandler::startElement(const char* name, CAttributeList& attributes)
        throw (CMiniSAXException)
{
	State	oldState = state;

	switch (state) {
	case ROOT_STATE:
		if (strcmp( name, "links" ) == 0) {
			state = LINKS_STATE;
			break;
		}
		break;
	case LINKS_STATE:
		if (strcmp( name, "link" ) == 0) {
			link = new CLink();
			state = LINK_STATE;
			break;
		}
		break;
	case LINK_STATE:
		if (strcmp( name, "name" ) == 0) {
			state = NAME_STATE;
			break;
		}
		if (strcmp( name, "address" ) == 0) {
			state = ADDRESS_STATE;
			break;
		}
		if (strcmp( name, "outbound" ) == 0) {
			// check type
			const CAttribute*	outboundType = attributes.getAttribute( "type" );
			if (outboundType != NULL) {
				if (strcmp( outboundType->getValue(), "binkley" ) == 0) {
					link->setOutboundType( OT_BINKLEY );
				}
				else 
				if (strcmp( outboundType->getValue(), "fileBoxes" ) == 0) {
					link->setOutboundType( OT_FILE_BOXES );
				}
				else {
					throw CMiniSAXException( "Unknown outbound type", 
											parser->getCurrentColumn(), parser->getCurrentRow() );
				}
			} else {
				// outbound type not specified, default is binkley
				link->setOutboundType( OT_BINKLEY );
			}

			const CAttribute*	flavour = attributes.getAttribute( "flavour" );
			if (flavour == NULL) {
				// default flavour is hold
				link->setFlavourType( FT_HOLD );
			} else {
				FlavourType	flavourType;
				if (strcmp( flavour->getValue(), "normal" ) == 0) {
					flavourType = FT_NORMAL;
				}
				else
				if (strcmp( flavour->getValue(), "direct" ) == 0) {
					flavourType = FT_DIRECT;
				}
				else
				if (strcmp( flavour->getValue(), "hold" ) == 0) {
					flavourType = FT_HOLD;
				}
				else
				if (strcmp( flavour->getValue(), "crash" ) == 0) {
					flavourType = FT_CRASH;
				}
				else
				if (strcmp( flavour->getValue(), "immediate" ) == 0) {
					flavourType = FT_IMMEDIATE;
				}
				else {
					throw CMiniSAXException( "Unknown outbound flavour", 
												parser->getCurrentColumn(), parser->getCurrentRow() );
				}

				link->setFlavourType( flavourType );
			}

			// check file boxes flavour
			if (link->getOutboundType() == OT_FILE_BOXES) {
				if (link->getFlavourType() != FT_NORMAL && link->getFlavourType() != FT_HOLD) {
					throw CMiniSAXException( "FileBoxes outbound can have only either normal or hold flavour", 
												parser->getCurrentColumn(), parser->getCurrentRow() );
				}
			}

			state = OUTBOUND_STATE;
			break;
		}
		if (strcmp( name, "allowAreaCreation" ) == 0) {
			state = ALLOW_AREA_CREATION_STATE;
			link->setAllowAreaCreation( true );
			break;
		}
		if (strcmp( name, "status" ) == 0) {
			state = STATUS_STATE;
			break;
		}
		if (strcmp( name, "robotName" ) == 0) {
			state = ROBOT_NAME_STATE;
			break;
		}
		if (strcmp( name, "availAreas" ) == 0) {
			state = AVAIL_AREAS_STATE;
			break;
		}
		if (strcmp( name, "useAKA" ) == 0) {
			state = USE_AKA_STATE;
			break;
		}
		if (strcmp( name, "security" ) == 0) {
			state = SECURITY_STATE;
			break;
		}
		break;
	case ALLOW_AREA_CREATION_STATE:
		if (strcmp( name, "path" ) == 0) {
			state = PATH_STATE;
			break;
		}
		if (strcmp( name, "defaultGroup" ) == 0) {
			state = DEFAULT_GROUP_STATE;
			break;
		}
		if (strcmp( name, "sendPriority" ) == 0) {
			state = AUTOCREATED_SEND_PRIORITY_STATE;
			break;
		}
		if (strcmp( name, "receivePriority" ) == 0) {
			state = AUTOCREATED_RECEIVE_PRIORITY_STATE;
			break;
		}
		if (strcmp( name, "passthrough" ) == 0) {
			link->setAutocreatedPassthrough( true );
			state = AUTOCREATED_PASSTHROUGH_STATE;
			break;
		}
		if (strcmp( name, "openForResend" ) == 0) {
			link->setAutocreatedOpenForResend( true );
			state = AUTOCREATED_OPEN_FOR_RESEND_STATE;
			break;
		}
		break;
	case SECURITY_STATE:
		if (strcmp( name, "password" ) == 0) {
			state = PASSWORD_STATE;
			break;
		}
		if (strcmp( name, "sendPriority" ) == 0) {
			state = SEND_PRIORITY_STATE;
			break;
		}
		if (strcmp( name, "receivePriority" ) == 0) {
			state = RECEIVE_PRIORITY_STATE;
			break;
		}
		if (strcmp( name, "group" ) == 0) {
			state = GROUP_STATE;
			break;
		}
		break;
	}
	if (state == oldState) {
		// if we are here then we have unexpected element
		parser->raiseEx( "Unexpected element" );
	}
	parser->checkUnexpected( attributes );
}


void CLinksContentHandler::endElement(const char* name)
        throw (CMiniSAXException)
{
	switch (state) {
	case LINKS_STATE:
		if (strcmp( name, "links" ) == 0) {
			state = ROOT_STATE;
			return;
		}
		break;
	case LINK_STATE:
		if (strcmp( name, "link" ) == 0) {
			if (!link->isAKASpecified()) {
				CLinksManager::chooseAKA( link );
			}
			CLinksManager::getInstance()->addLink( link );
			link = NULL;
			state = LINKS_STATE;
			return;
		}
		break;
	case NAME_STATE:
		if (strcmp( name, "name" ) == 0) {
			state = LINK_STATE;
			return;
		}
		break;
	case ADDRESS_STATE:
		if (strcmp( name, "address" ) == 0) {
			state = LINK_STATE;
			return;
		}
		break;
	case OUTBOUND_STATE:
		if (strcmp( name, "outbound" ) == 0) {
			state = LINK_STATE;
			return;
		}
		break;
	case STATUS_STATE:
		if (strcmp( name, "status" ) == 0) {
			state = LINK_STATE;
			return;
		}
		break;
	case ROBOT_NAME_STATE:
		if (strcmp( name, "robotName" ) == 0) {
			state = LINK_STATE;
			return;
		}
		break;
	case AVAIL_AREAS_STATE:
		if (strcmp( name, "availAreas" ) == 0) {
			state = LINK_STATE;
			return;
		}
		break;
	case USE_AKA_STATE:
		if (strcmp( name, "useAKA" ) == 0) {
			state = LINK_STATE;
			return;
		}
		break;
	case PATH_STATE:
		if (strcmp( name, "path" ) == 0) {
			state = ALLOW_AREA_CREATION_STATE;
			return;
		}
		break;
	case DEFAULT_GROUP_STATE:
		if (strcmp( name, "defaultGroup" ) == 0) {
			state = ALLOW_AREA_CREATION_STATE;
			return;
		}
		break;
	case AUTOCREATED_SEND_PRIORITY_STATE:
		if (strcmp( name, "sendPriority" ) == 0) {
			state = ALLOW_AREA_CREATION_STATE;
			return;
		}
		break;
	case AUTOCREATED_RECEIVE_PRIORITY_STATE:
		if (strcmp( name, "receivePriority" ) == 0) {
			state = ALLOW_AREA_CREATION_STATE;
			return;
		}
		break;
	case AUTOCREATED_PASSTHROUGH_STATE:
		if (strcmp( name, "passthrough" ) == 0) {
			state = ALLOW_AREA_CREATION_STATE;
			return;
		}
		break;
	case AUTOCREATED_OPEN_FOR_RESEND_STATE:
		if (strcmp( name, "openForResend" ) == 0) {
			state = ALLOW_AREA_CREATION_STATE;
			return;
		}
		break;
	case ALLOW_AREA_CREATION_STATE:
		if (strcmp( name, "allowAreaCreation" ) == 0) {
			state = LINK_STATE;
			return;
		}
		break;
	case PASSWORD_STATE:
		if (strcmp( name, "password" ) == 0) {
			state = SECURITY_STATE;
			return;
		}
		break;
	case SEND_PRIORITY_STATE:
		if (strcmp( name, "sendPriority" ) == 0) {
			state = SECURITY_STATE;
			return;
		}
		break;
	case RECEIVE_PRIORITY_STATE:
		if (strcmp( name, "receivePriority" ) == 0) {
			state = SECURITY_STATE;
			return;
		}
		break;
	case GROUP_STATE:
		if (strcmp( name, "group" ) == 0) {
			state = SECURITY_STATE;
			return;
		}
		break;
	case SECURITY_STATE:
		if (strcmp( name, "security" ) == 0) {
			state = LINK_STATE;
			return;
		}
		break;
	}
	// if we are here then we have unexpected element
	throw CMiniSAXException( "Unexpected element", parser->getCurrentColumn(), parser->getCurrentRow() );
}


void CLinksContentHandler::handleCharacters(const char* buffer, size_t len)
        throw (CMiniSAXException)
{
	switch (state) {
	case NAME_STATE:
		{
			link->setName( tstring( buffer, len ) );
		}
		return;
	case ADDRESS_STATE:
		{
			CAddress	address;
			if (!address.parseString( tstring( buffer, len ).c_str() ))
			{
				parser->raiseEx( "Invalid link address" );
			}
			address.baseUpon( CCommonData::getInstance()->getSysopAddress() );
			link->setAddress( address );
		}
		return;
	case STATUS_STATE:
		{
			tstring	status( buffer, len );
			int		activity = ACTIVITY_USUAL;
			
			if (strcmp( status.c_str(), "passive" ) == 0) {
				activity = ACTIVITY_PASSIVE;
			}
			else 
			if (strcmp( status.c_str(), "unavailable" ) == 0) {
				activity = ACTIVITY_UNAVAILABLE;
			}

			if (activity == ACTIVITY_USUAL) {
				throw CMiniSAXException( "Unknown activity", 
										 parser->getCurrentColumn(), parser->getCurrentRow() );
			}

			link->setActivity( activity );
		}
		return;
	case ROBOT_NAME_STATE:
		{
			link->setRobotName( tstring( buffer, len ) );
		}
		return;
	case AVAIL_AREAS_STATE:
		{
			link->setAvailAreasFileName( tstring( buffer, len ) );
		}
		return;
	case USE_AKA_STATE:
		{
			CAddress	address;
			if (!address.parseString( tstring( buffer, len ).c_str() ))
			{
				parser->raiseEx( "Invalid AKA address" );
			}
			CommonDataPtr	commonData = CCommonData::getInstance();
			address.baseUpon( commonData->getSysopAddress() );
			// check AKA (it should be main address or in AKA list
			if (address != commonData->getSysopAddress()) {
				if (!commonData->hasSysopAKA( address )) {
					parser->raiseEx( "Unknown AKA address" );
				}
			}
			link->setOurAddress( address );
			link->setAKASpecified( true );
		}
		return;
	case PATH_STATE:
		{
			link->setAutoCreatedPathName( tstring( buffer, len ) );
		}
		return;
	case DEFAULT_GROUP_STATE:
		{
			link->setDefaultGroup( tstring( buffer, len ) );
		}
		return;
	case PASSWORD_STATE:
		{
			link->setPassword( tstring( buffer, len ) );
		}
		return;
	case SEND_PRIORITY_STATE:
		{
			tstring	priorityStr( buffer, len );
			int		priority;
			int		scanResult = sscanf( priorityStr.c_str(), "%d", &priority );
			if (scanResult == 0 || scanResult == EOF || priority < 0) {
				parser->raiseEx( "Invalid send priority" );
			}
			link->setSendPriority( priority );
		}
		return;
	case RECEIVE_PRIORITY_STATE:
		{
			tstring	priorityStr( buffer, len );
			int		priority;
			int		scanResult = sscanf( priorityStr.c_str(), "%d", &priority );
			if (scanResult == 0 || scanResult == EOF || priority < 0) {
				parser->raiseEx( "Invalid receive priority" );
			}
			link->setReceivePriority( priority );
		}
		return;
	case AUTOCREATED_SEND_PRIORITY_STATE:
		{
			tstring	priorityStr( buffer, len );
			int		priority;
			int		scanResult = sscanf( priorityStr.c_str(), "%d", &priority );
			if (scanResult == 0 || scanResult == EOF || priority < 0) {
				parser->raiseEx( "Invalid autocreated send priority" );
			}
			link->setAutocreatedSendPriority( priority );
		}
		return;
	case AUTOCREATED_RECEIVE_PRIORITY_STATE:
		{
			tstring	priorityStr( buffer, len );
			int		priority;
			int		scanResult = sscanf( priorityStr.c_str(), "%d", &priority );
			if (scanResult == 0 || scanResult == EOF || priority < 0) {
				parser->raiseEx( "Invalid autocreated receive priority" );
			}
			link->setAutocreatedReceivePriority( priority );
		}
		return;
	case GROUP_STATE:
		{
			tstring	group( buffer, len );
			if (link->canAccessGroup( group )) {
				parser->raiseEx( "Duplicate group" );
			}
			link->addGroup( group );
		}
		return;
	}
	if (!hasOnlySpaces( buffer, len )) {
		parser->raiseEx( "Unexpected symbols" );
	}
}
