#pragma once

#include "Singletons.h"
#include "Address.h"

enum LogMode {
	LM_START,
	LM_STOP,
	LM_SILENT,
	LM_LOUD,
	LM_WARNING,
	LM_ERROR
};

class ModeHolder {
public:
	ModeHolder(LogMode mode) : m_mode(mode) {}

	LogMode getMode() const { return m_mode; }
protected:
	LogMode	m_mode;
};

extern ModeHolder	LOG_START;
extern ModeHolder	LOG_STOP;
extern ModeHolder	LOG_SILENT;
extern ModeHolder	LOG_LOUD;
extern ModeHolder	LOG_WARNING;
extern ModeHolder	LOG_ERROR;

class CLogger : public CSingleton
{
public:
	virtual ~CLogger(void);

	virtual void destroy();
	// we depend from nothing
	virtual CSingletonVector getDependencies() { return CSingletonVector(); }

	static CLogger* getInstance() {
		static CLogger	logger;

		return &logger;
	}

	virtual CLogger& operator << (const ModeHolder modeHolder);

	virtual CLogger& operator << (const tstring& message);
	virtual CLogger& operator << (LPCTSTR message);
	virtual CLogger& operator << (int value);
	virtual CLogger& operator << (unsigned short value);

	virtual CLogger& operator << (const CAddress& address);

	virtual void flush();

	virtual bool openLogFile(LPCTSTR logFileName);

	virtual void rememberMode();
	virtual void restoreMode();

	// formats date and time to be represented in log
	static void formatDateTime(const SYSTEMTIME& someTime, LPTSTR timeString);

protected:
	LogMode				m_mode;
	bool				m_startOfLine;
	// used to remember/restore mode
	std::stack<LogMode>	m_modeHistory;

	std::ofstream	m_logFile;

	CLogger(void);
	virtual void logToStdOut(LPCTSTR message);
	virtual void logToStdErr(LPCTSTR message);
	virtual void logToFile(LPCTSTR message);

	virtual void logSingleLinePart(LPCTSTR part);
};

// smart pointer ;)
class LoggerPtr {
public:
	LoggerPtr() { 
		m_loggerPtr = CLogger::getInstance();
		m_loggerPtr->rememberMode();
	}
	~LoggerPtr() {
		m_loggerPtr->restoreMode();
	}

	void rememberMode() { m_loggerPtr->rememberMode(); }
	void restoreMode()  { m_loggerPtr->restoreMode();  }

	void flush() { m_loggerPtr->flush(); }

	CLogger& operator << (const ModeHolder modeHolder) { return ((*m_loggerPtr) << modeHolder); }

	CLogger& operator << (const tstring& message) {
		return ((*m_loggerPtr) << message);
	}

	CLogger& operator << (LPCTSTR message) {
		return ((*m_loggerPtr) << message);
	}

	CLogger& operator << (int value) {
		return ((*m_loggerPtr) << value);
	}

	CLogger& operator << (unsigned short value) {
		return ((*m_loggerPtr) << value);
	}

	CLogger& operator << (const CAddress& address) {
		return ((*m_loggerPtr) << address);
	}
protected:
	CLogger*	m_loggerPtr;
};

