#pragma once

#include "MiniSAX.h"

namespace miniSAX {

	class CMiniSAXParserImpl : public CMiniSAXParser {
	public:
		CMiniSAXParserImpl();
		virtual ~CMiniSAXParserImpl();

		// Sets encoding that will be used to convert data before submitting it to handlers.
		// Currently supported encoding is "cp866".
		// Exception is thrown when caller tries to set unsupported encoding
		virtual void setHandlerEncoding(const char* encoding) throw(CMiniSAXException);
		// Gets original encoding of the document.
		// Returns NULL if encoding was not specified.
		virtual const char* getDocumentEncoding() const;

		virtual void setDocumentHandler(CDocumentHandler* aHandler) { m_documentHandler = aHandler; }
		virtual void setContentHandler(CContentHandler* aHandler) { m_contentHandler = aHandler; }
		virtual void setErrorHandler(CErrorHandler* aHandler) { m_errorHandler = aHandler; }
		virtual void setEntityResolver(CEntityResolver* aResolver) { m_entityResolver = aResolver; }

		virtual void parse(CDataSource* dataSource) throw(CMiniSAXException);

		virtual int getCurrentRow()    const { return m_currentRow;    }
		virtual int getCurrentColumn() const { return m_currentColumn; }

		virtual void raiseEx(const char* msg) const throw(CMiniSAXException);

		virtual void checkUnexpected(CAttributeList& list) const throw(CMiniSAXException);

		static CMiniSAXParser* create() { return new CMiniSAXParserImpl(); }
	protected:
		CDocumentHandler*	m_documentHandler;
		CContentHandler*	m_contentHandler;
		CErrorHandler*		m_errorHandler;
		CEntityResolver*	m_entityResolver;

		std::string		m_handlerEncoding;
		std::string		m_documentEncoding;

		virtual bool processXMLDeclaration(CDataSource* dataSource);

		virtual bool readStartOfElement(CDataSource* dataSource);
		virtual bool readEndOfElement(CDataSource* dataSource);

		virtual bool isGoodBeginning(CDataSource* dataSource) throw (CMiniSAXException);

		virtual bool readName(CDataSource* dataSource, char*& name);
		virtual bool readValue(CDataSource* dataSource, char*& value);

		virtual bool skipComment(CDataSource* dataSource);

		virtual void startDocument() throw (CMiniSAXException);
		virtual void endDocument()   throw (CMiniSAXException);

		virtual void startElement(const char* name, CAttributeList& attributes)
			throw (CMiniSAXException);
		virtual void endElement(const char* name)
			throw (CMiniSAXException);
		virtual void handleCharacters(const char* buffer, size_t len)
			throw (CMiniSAXException);

		virtual void handleError(const CMiniSAXException& e);

		virtual void resolveEntities(char* data, char* storage);

		/**
		 * These fields are accessible via corresponding get... methods.
		 * They are being updated each time before we transmit execution to the handler.
		 * This way handler can throw correct CMiniSAXException ;)
		 */
		int	m_currentRow, m_currentColumn;

		CCharSetConvertor*	m_charSetConvertor;
	};
}
