#include "pch.h"

#include "Mem.hpp"
#include "Common.h"
#include "CfgParse.hpp"
#include "Misc.hpp"
#include "Links.h"
#include "Areas.h"
#include "Msg.hpp"
#include "Tic.h"
#include "crc32.hpp"
#include "Temp.hpp"
#include "File.h"

#include "CfgLexems.hpp"
#include "WELexems.h"

#include "MsgAttrs.h"

#include "Adrenalin.h"

#include "LinksManager.h"
#include "AreasManager.h"

// TODO: not all info printed (see links, areas)

/**
 * Implementation of void CAdrenalin::PrintDebugInfo()
 */

void CAdrenalin::PrintDebugInfo() {
	int		i;

	CommonDataPtr	commonData = CCommonData::getInstance();

	std::cout << _T("Debug information\n");

	std::cout << _T("System: ") << commonData->getSysopAddress().toString().c_str() << _T("\n");
	for (unsigned int index = 0; index < commonData->getSysopAKACount(); index++) {
		std::cout << _T("   AKA: ") << commonData->getSysopAKA(index).toString().c_str() << _T("\n");
	}
	std::cout << _T(" SysOp: ") << commonData->getSysopName().c_str() << _T("\n");

	std::cout << _T("Directories:\n");
	std::cout << _T("       Inbound: ") << commonData->getInboundPath().c_str() << TEXT_EOL
			  << _T("  BinkOutbound: ") << commonData->getBinkOutboundPath().c_str() << TEXT_EOL;

	// file boxes
	std::cout << TEXT("     FileBoxes: ");
	if (commonData->areLongFileBoxes() || commonData->getFileBoxesPath().size() == 0) {
		std::cout << TEXT("<undefined>");
	} else {
		std::cout << commonData->getFileBoxesPath().c_str();
	}
	std::cout << TEXT_EOL;

	// long file boxes
	std::cout << TEXT(" LongFileBoxes: ");
	if (!commonData->areLongFileBoxes() || commonData->getFileBoxesPath().size() == 0) {
		std::cout << TEXT("<undefined>");
	} else {
		std::cout << commonData->getFileBoxesPath().c_str();
	}
	std::cout << TEXT_EOL;

	std::cout << _T("       Netmail: ") << commonData->getNetmailPath().c_str() << TEXT_EOL
		      << _T("       BadTics: ") << commonData->getBadTicsPath().c_str() << TEXT_EOL
			  << _T("       NewTics: ") << commonData->getNewTicsPath().c_str() << TEXT_EOL
			  << _T(" ProcessedTics: ");
	if (commonData->getProcessedTicsPath().empty()) {
		std::cout << TEXT("undefined (TICs will be removed after processing)\n");
	}
	else {
		std::cout << commonData->getProcessedTicsPath().c_str() << TEXT_EOL;
	}
    
	std::cout << TEXT("      Filelist: ") << commonData->getFileListName().c_str() << TEXT_EOL;
	std::cout << TEXT("FilelistFormat: ") << commonData->getFileListFormat().getFormat().c_str() << TEXT_EOL;

    // requests log file
    std::cout << TEXT("Requests log: ")
			  << (commonData->getRequestsLogFileName().empty() ? tstring(TEXT("undefined"))
                                                     : commonData->getRequestsLogFileName())
              << TEXT_EOL;

	std::cout << _T("Flags: ");
	if (commonData->getEventFileFlagCount() == 0) {
		std::cout << TEXT("not defined");
	} else {
		std::cout << std::endl;
		for (size_t	flagIndex = 0; flagIndex < commonData->getEventFileFlagCount(); flagIndex++) {
			const CEventFileFlag&	flag = commonData->getEventFileFlag( flagIndex );
			switch (flag.getEvent()) {
				case EE_FIX:
					std::cout << _T("    after fix: ");
					break;
				case EE_TOSS:
					std::cout << _T("   after toss: ");
					break;
				case EE_HATCH:
					std::cout << _T("  after hatch: ");
					break;
			}
			std::cout << flag.getFileName().c_str() << std::endl;
		}
	}
    
	int	aliasCount = commonData->getAliasCount();
	if (aliasCount > 0) {
		std::cout << _T("Alias(es):\n");
		for (i = 0; i < commonData->getAliasCount(); i++) {
			std::cout << _T("  ") << commonData->getAlias(i).c_str() << _T("\n");
		}
	} else {
		std::cout << TEXT("No aliases defined\n");
	}

	// attributes
	std::cout << TEXT("Reply message attributes: ") 
		     << MsgAttrsToString( CCommonData::getInstance()->getReplyMessageAttributes(),
								  CCommonData::getInstance()->getReplyMessageFlags() )
		     << TEXT_EOL;

	// warn sysop
	std::cout << TEXT("Warn sysop by mail about:");
	if (!CCommonData::getInstance()->doNotifyOnErrors() && !CCommonData::getInstance()->doNotifyOnWarnings()) {
		std::cout << TEXT( " nothing" );
	}
	else {
		if (CCommonData::getInstance()->doNotifyOnWarnings()) {
			std::cout << TEXT( " warnings" );
		}
		if (CCommonData::getInstance()->doNotifyOnErrors()) {
			std::cout << TEXT( " errors" );
		}
	}
	std::cout << TEXT_EOL;

	// no mail notify
	std::cout << TEXT("No mail notifications for: ");
	if (commonData->isNoMailNotifyForBasePurge()) {
		std::cout << "BasePurge";
	}
	else {
		std::cout << "nothing";
	}
	std::cout << TEXT_EOL;

	// links
	std::cout << _T("Link(s):\n");
	for (CLinksManager::CIterator itLink = CLinksManager::getInstance()->begin();
		 itLink != CLinksManager::getInstance()->end(); ++itLink)
	{
		PLink	link = (*itLink);
		std::cout << _T("  ---------------\n")
				  << _T("         address: ") << link->getAddress().toString().c_str() << _T("\n")
				  << _T("  ---------------\n")
				  << TEXT("            name: ") << link->getName().c_str() << TEXT("\n")
				  << TEXT("      robot name: ") << link->getRobotName().c_str() << TEXT("\n")
				  << TEXT("        password: ") << link->getPassword().c_str() << TEXT_EOL;

		// outbound type
		std::cout << TEXT("   outbound type: ");
		switch (link->getOutboundType()) {
			case OT_BINKLEY:
				std::cout << TEXT("binkley");
				break;
			case OT_FILE_BOXES:
				std::cout << TEXT("fileBoxes");
				break;
		}
		std::cout << TEXT_EOL;

		// flavour type
		std::cout << _T("         flavour: ");
		switch (link->getFlavourType()) {
		case FT_NORMAL:
			std::cout << _T("normal");
			break;
		case FT_CRASH:
			std::cout << _T("crash");
			break;
		case FT_DIRECT:
			std::cout << _T("direct");
			break;
		case FT_HOLD:
			std::cout << _T("hold");
			break;
		case FT_IMMEDIATE:
			std::cout << _T("immediate");
			break;
		}
		std::cout << TEXT("\n");

		std::cout << TEXT("   send priority: ") << link->getSendPriority() << TEXT_EOL
					 << TEXT("receive priority: ") << link->getReceivePriority() << TEXT_EOL;

		int groupsCount = link->getGroupsCount();
		if (groupsCount > 0) {
			std::cout << TEXT("  allowed groups: ");
			for (i = 0; i < groupsCount; i++) {
				std::cout << link->getGroup(i).c_str() << TEXT_EOL;
				if (i < groupsCount-1) {
					// it's not last group
					std::cout << TEXT("                  ");
				}
			}
		} else {
			std::cout << TEXT("no allowed groups\n");
		}

		std::cout << _T("     our address: ") << link->getOurAddress().toString().c_str() << TEXT_EOL;

		// available areas file name
		std::cout << TEXT("      AvailAreas: ");
		if (link->getAvailAreasFileName().size() > 0) {
			std::cout << link->getAvailAreasFileName().c_str();
		} else {
			std::cout << TEXT("<undefined>");
		}
		std::cout << TEXT_EOL;

		// area creation

		std::cout << TEXT("\n   area creation: ");
		if (link->canCreateAreas()) {
			std::cout << TEXT("allowed\n");
			std::cout << TEXT("      default group: ") << link->getDefaultGroup().c_str() << TEXT_EOL;
			std::cout << TEXT("      send priority: ") << link->getAutocreatedSendPriority()    << TEXT_EOL;
			std::cout << TEXT("   receive priority: ") << link->getAutocreatedReceivePriority() << TEXT_EOL;
			std::cout << TEXT("        passthrough: ")
						 << (link->isAutocreatedPassthrough() ? TEXT("yes") : TEXT("no")) << TEXT_EOL;
			std::cout << TEXT("    open for resend: ")
						 << (link->isAutocreatedOpenForResend() ? TEXT("yes") : TEXT("no")) << TEXT_EOL;
		}
		else {
			std::cout << TEXT("not allowed\n");
		}

		std::cout << TEXT_EOL;
	}

	// Areas
	std::cout << TEXT("Area(s):\n");
	int	areasCount = CAreasManager::getInstance()->getAreasCount();
	for (i = 0; i < areasCount; i++) {
		PArea	area = CAreasManager::getInstance()->getArea( i );
		std::cout << _T("---\n");
		std::cout << _T("    name: ") << area->getName().c_str() << _T("\n")
				 << _T("    path: ") << area->getPath().c_str() << _T("\n")
				 << _T("   wpath: ") << area->getWorkingPath().c_str() << _T("\n")
				 << _T("    desc: ") << area->getDescription().c_str() << _T("\n")
				 << _T("  send access: ") << area->getSendPriority() << _T("\n")
				 << _T("  receive access: ") << area->getReceivePriority() << _T("\n")
				 << _T("   group: ") << area->getGroup()->getName().c_str() << _T("\n")
				 << _T("  passthrough: ");
		if (area->hasFlag( CArea::FLAG_PASSTHROUGH ))
			std::cout << _T("yes\n");
		else
			std::cout << _T("no\n");
		std::cout << _T("  veto manual purge: ");
		if (area->hasFlag( CArea::FLAG_VETO_MANUAL_PURGE ))
			std::cout << _T("yes\n");
		else
			std::cout << _T("no\n");
		std::cout << TEXT("  open for resend: ")
			         << (area->hasFlag( CArea::FLAG_OPEN_FOR_RESEND ) ? TEXT("yes") : TEXT("no")) << TEXT_EOL;

		// outbound
		std::cout << TEXT("         outbound: ");
		if (area->isOutboundDefined()) {
			switch (area->getOutboundType()) {
				case OT_BINKLEY:
					std::cout << TEXT("binkley");
					break;
				case OT_FILE_BOXES:
					std::cout << TEXT("fileBoxes");
					break;
			}
		} else {
			std::cout << "<undefined>";
		}
		std::cout << TEXT_EOL;

		int	tiesCount = area->getTiesCount();
		if (tiesCount == 0) {
			std::cout << _T("  no links linked\n");
		}
		else {
			std::cout << _T("  linked:\n");
			for (int j = 0; j < tiesCount; j++) {
				ConstTiePtr	tie = area->getTie( j );
				std::cout << _T("  ") << tie->getLink()->getAddress().toString().c_str() << _T(", ")
						 << tie->getRouteType() << _T("\n");
			}
		}
	}
}
